import re
import sys

from ..model import wtypes

from ..model.errors import AssertionVMError
from ..model.memory import Ref
from ..model.wtypes import BASE, CONST, VAR
from ..model.ops import Dynamic


__all__ = ['builtins']


def builtin_assert(captures, args):
    '''
    Implements the Welly `assert` built-in. This is a function obeying the
    Welly-in-Python calling convention. Its Welly type is
    `fn(bool, str): void`. Therefore `args` is a const Ref to a 2-word struct
    containing a `bool` and a `str`, and we have to return `('return', r)`
    where `r` is a const Ref to an empty struct.
    '''
    assert captures == []
    condition, message = args.read()
    if not condition:
        raise AssertionVMError(message)
    return ('return', Ref.new(()))


def builtin_f2i(captures, args):
    '''
    Implements the Welly `f2i` built-in. This is a function obeying the
    Welly-in-Python calling convention. Its Welly type is `fn(float): int`.
    Therefore `args` is a const Ref to a 1-word struct containing a `float`,
    and we have to return `('return', r)` where `r` is a const Ref to an
    int.
    '''
    assert captures == []
    (f,) = args.read()
    assert type(f) is float
    try:
        i = int(f)
    except ValueError:
        raise AssertionVMError('cannot convert the float value to type int')
    return ('return', Ref.new((i,)))


def builtin_f2s(captures, args):
    '''
    Implements the Welly `f2s` built-in. This is a function obeying the
    Welly-in-Python calling convention. Its Welly type is `fn(float): str`.
    Therefore `args` is a const Ref to a 1-word struct containing a `float`,
    and we have to return `('return', r)` where `r` is a const Ref to a str.
    '''
    assert captures == []
    (f,) = args.read()
    assert type(f) is float
    try:
        s = repr(f)
    except ValueError:
        raise AssertionVMError('cannot convert the float value to type str')
    return ('return', Ref.new((s,)))


def builtin_i2f(captures, args):
    '''
    Implements the Welly `i2f` built-in. This is a function obeying the
    Welly-in-Python calling convention. Its Welly type is `fn(int): float`.
    Therefore `args` is a const Ref to a 1-word struct containing an `int`,
    and we have to return `('return', r)` where `r` is a const Ref to a float.
    '''
    assert captures == []
    (i,) = args.read()
    assert type(i) in (int, long)
    try:
        f = float(i)
    except ValueError:
        raise AssertionVMError('cannot convert the int value to type float')
    return ('return', Ref.new((f,)))


def builtin_i2s(captures, args):
    '''
    Implements the Welly `i2s` built-in. This is a function obeying the
    Welly-in-Python calling convention. Its Welly type is `fn(int): str`.
    Therefore `args` is a const Ref to a 1-word struct containing an `int`,
    and we have to return `('return', r)` where `r` is a const Ref to a str.
    '''
    assert captures == []
    (i,) = args.read()
    assert type(i) in (int, long)
    try:
        s = str(i)
    except ValueError:
        raise AssertionVMError('cannot convert the int value to type str')
    return ('return', Ref.new((s,)))


def builtin_s2f(captures, args):
    '''
    Implements the Welly `s2f` built-in. This is a function obeying the
    Welly-in-Python calling convention. Its Welly type is `fn(str): float`.
    Therefore `args` is a const Ref to a 1-word struct containing a `str`,
    and we have to return `('return', r)` where `r` is a const Ref to a float.
    '''
    assert captures == []
    (s,) = args.read()
    assert type(s) is str
    try:
        f = float(s)
    except ValueError:
        raise AssertionVMError('cannot convert the str value to type float')
    return ('return', Ref.new((f,)))


def builtin_s2i(captures, args):
    '''
    Implements the Welly `s2i` built-in. This is a function obeying the
    Welly-in-Python calling convention. Its Welly type is `fn(str): int`.
    Therefore `args` is a const Ref to a 1-word struct containing a `str`,
    and we have to return `('return', r)` where `r` is a const Ref to a int.
    '''
    assert captures == []
    (s,) = args.read()
    assert type(s) is str
    m = re.match(r'-?\d+\Z', s)
    if m:
        i = int(m.group())
    else:
        raise AssertionVMError('cannot convert the str value to type int')
    return ('return', Ref.new((i,)))


def builtin_print(captures, args):
    '''
    Implements the Welly `print` built-in. This is a function obeying the
    Welly-in-Python calling convention. Its Welly type is `fn(str): void`.
    Therefore `args` is a const Ref to a 1-word struct containing a `str`,
    and we have to return `('return', r)` where `r` is a const Ref to an
    empty struct.
    '''
    assert captures == []
    (line,) = args.read()
    assert type(line) is str
    sys.stdout.write(line)
    return ('return', Ref.new(()))


# Map from identifier to dynamic.
builtins = {
    'array0': Dynamic(
        wtypes.type_,
        wtypes.array0,
    ),
    'assert': Dynamic.new(
        wtypes.function([wtypes.bool_, wtypes.str_], wtypes.void),
        (builtin_assert, []),
    ),
    'bool': Dynamic(
        wtypes.type_,
        wtypes.bool_,
    ),
    'dynamic': Dynamic(
        wtypes.type_,
        wtypes.dynamic,
    ),
    'f2i': Dynamic.new(
        wtypes.function([wtypes.float_], wtypes.int_),
        (builtin_f2i, []),
    ),
    'f2s': Dynamic.new(
        wtypes.function([wtypes.float_], wtypes.str_),
        (builtin_f2s, []),
    ),
    'false': Dynamic.new(
        wtypes.bool_,
        (False,),
    ),
    'float': Dynamic(
        wtypes.type_,
        wtypes.float_,
    ),
    'i2f': Dynamic.new(
        wtypes.function([wtypes.int_], wtypes.float_),
        (builtin_i2f, []),
    ),
    'i2s': Dynamic.new(
        wtypes.function([wtypes.int_], wtypes.str_),
        (builtin_i2s, []),
    ),
    'inf': Dynamic.new(
        wtypes.float_,
        (float('inf'),),
    ),
    'int': Dynamic(
        wtypes.type_,
        wtypes.int_,
    ),
    # No 'module' because it's a keyword.
    'nan': Dynamic.new(
        wtypes.float_,
        (float('nan'),),
    ),
    'never': Dynamic(
        wtypes.type_,
        wtypes.never,
    ),
    'print': Dynamic.new(
        wtypes.function([wtypes.str_], wtypes.void),
        (builtin_print, []),
    ),
    's2f': Dynamic.new(
        wtypes.function([wtypes.str_], wtypes.float_),
        (builtin_s2f, []),
    ),
    's2i': Dynamic.new(
        wtypes.function([wtypes.str_], wtypes.int_),
        (builtin_s2i, []),
    ),
    'str': Dynamic(
        wtypes.type_,
        wtypes.str_,
    ),
    'true': Dynamic.new(
        wtypes.bool_,
        (True,),
    ),
    'type': Dynamic(
        wtypes.type_,
        wtypes.type_,
    ),
    'version': Dynamic.new(
        wtypes.str_,
        ("0.0.0",),
    ),
    'void': Dynamic(
        wtypes.type_,
        wtypes.void,
    ),
}
