from ..printer import key_value
from ..model.wtypes import BASE, CONST, VAR


'''
Data structures and utilities for generating a low-level representation of
source code. This representation has two main purposes:

 - Sometimes the same source code is compiled multiple times, with different
values and types for the external variables. In such a situation, you want to
do as much work as possible once up front, to minimise the work that is
repeated. Flat code represents the work that must be repeated.

 - There are many different source code constructs that mean the same thing.
Later stages of the compiler should not depend directly on the original source
code, but only on its meaning. Translating into flat code avoids bugs by
removing some of the irrelevant information.

For further motivation see "doc/compiler/compiler-levels.txt" and
"doc/compiler/il-levels.txt".

Flat code is organised into Blocks. A Block contains a list of instructions
which are to be executed in sequence. Blocks are mutable: you can build a
Block gradually by appending instructions to it.

The control flow instructions 'if' and 'loop' contain nested lists of
instructions. The instruction 'closure' for constructing function values also
contains a nested list of instructions. Instructions are thereby organised
into a tree. These lists of instructions are passed as Blocks.

Instructions notionally pass values to one another in boxed form, i.e. each
value is in its own heap block. Heap blocks are allocated for every declared
variable, and also for temporary values created while evaluating expressions.
The boxing is inefficient, but in common cases the inefficiency can be
mitigated by a later optimisation stage of the compiler. The pointers to the
values are read-only values.

Pointers to values can be stored on a data stack. Most instructions,
including all arithmetic instructions, consume and/or produce pointers on the
stack. The stack is private to a Block, and cannot be accessed from nested
Blocks. The size of the stack must be static: all paths to a given point in
the code must leave the same number of items on the stack. The stack must be
empty at key control-flow points, such as on entry and exit from loops and
functions.

A pointer to a value can be popped from the stack and stored in a local
variable using an 'init' instruction. The pointer can later be retrieved and
pushed onto the stack by a 'local' instruction. For each local variable,
there must be exactly one 'init' instruction, which must precede all 'local'
instructions on every possible control-flow path.

Functions take zero or more arguments. When in a function, the arguments
become the first few local variables (their 'init' instructions are implicit)
and can therefore be accessed using 'local' instructions.

A 'capture' instruction makes a pointer to a value captured when the function
was defined. This instructions pushes the pointer onto the stack.

Locals and captured values are numbered sequentially from zero. The numbering
is private to a function, i.e. it restarts for every nested function. Locals
which are not simultaneously in scope, e.g. declared in different branches of
an 'if', can reuse the same number.

Metadata about locals, other than function parameters, is attached to their
corresponding 'init' instructions. Metadata about a function and its
parameters is attached to its 'closure' instruction. The captures are
notionally passed to the 'closure' instruction on the stack.

External variables are those which are mentioned but not declared within the
compilation unit. These are treated as captures of the outermost scope, i.e.
they are accessed using the 'capture' instruction. Flat code has no way of
recording the names of variables. The names of external variables are needed
for linking purposes. The caller must therefore take steps to record them, in
addition to generating the flat code.
'''


class FlatCodeError(Exception):
    pass


class Block(object):
    '''
    Accumulates a sequence of instructions corresponding to a high-level
    lexical block.

    Public fields:

     - instructions - A list of instructions.
     - is_end_reachable - Is there a path to the end of the Block?
     - is_break_reachable - Is there a path to a 'break' instruction?
     - is_continue_reachable - Is there a path to a 'continue' instruction?
     - stack_used - The number of stack items at the end of the Block.

    Instructions are tuples, with the first element being the op-code (a str),
    the last element being a source code location (an opaque value), and
    other elements being opcode-dependent. Non-constant operands are always
    popped from the stack, and results are pushed to the stack, so these do
    not appear in the tuple.

    All the public methods of this class append instructions. Instructions
    should only be appended using the methods, as they do bookkeeping.
    '''

    def __init__(self):
        self.instructions = []
        self.is_end_reachable = True
        self.is_break_reachable = False
        self.is_continue_reachable = False
        self.stack_used = 0

    def __pretty__(self):
        return (
            u'Block {',
            u', ',
            u'}',
            [
                key_value('instructions', self.instructions),
                key_value('is_end_reachable', self.is_end_reachable),
                key_value('is_break_reachable', self.is_break_reachable),
                key_value('is_continue_reachable', self.is_continue_reachable),
                key_value('stack_used', self.stack_used),
            ]
        )

    def _append(self, num_consumed, num_produced, instruction, loc=None):
        '''
        Appends `instruction` to this Block. It is an error to append an
        instruction unless `is_end_reachable`. `stack_used` will be
        decremented by `num_consumed`, checking for underflow. It will then be
        incremented by `num_produced`.

         - loc - An opaque value representing a source code location to be used
           in error messages pertaining to `instruction`.
        '''
        if not self.is_end_reachable:
            raise FlatCodeError('Code is not reachable')
        self.stack_used -= num_consumed
        if self.stack_used < 0:
            raise FlatCodeError('Stack underflow')
        self.stack_used += num_produced
        self.instructions.append(instruction + (loc,))

    def _check_empty(self):
        '''Raises an exception if the stack is not empty.'''
        if self.stack_used != 0:
            raise FlatCodeError(
                'Stack is not empty after %r' % self.instructions)

    def get_code(self, is_break_allowed):
        '''
        Returns a fresh copy of self.instructions.

         - is_break_allowed - `True` if "break" and "continue" are allowed.
           If "break" is not allowed, but is reachable in `self`, then this
           method raises a FlatCodeError.
        '''
        if self.is_break_reachable and not is_break_allowed:
            raise FlatCodeError('Illegal "break" instruction')
        if self.is_continue_reachable and not is_break_allowed:
            raise FlatCodeError('Illegal "continue" instruction')
        return list(self.instructions)

    def address_of(self, loc):
        '''..., value -> ..., reference'''
        self._append(1, 1, ('address_of',), loc=loc)

    def array(self, n, loc):
        '''..., <n values> -> ..., array_value'''
        self._append(n, 1, ('array', n), loc=loc)

    def assign(self, loc):
        '''..., rvalue, lvalue -> ...'''
        self._append(2, 0, ('assign',), loc=loc)

    def assign_op(self, op, loc):
        '''
        ..., rvalue, lvalue -> ...

        A CISC-like instruction equivalent to the following sequence:
        bury(1) pick(1) bury(1) binary(op) bury(1) assign. This sequence is
        used to implement in-place modification, e.g. "x += 1".
        '''
        # TODO: Give it its own op-code.
        self.bury(1, loc)
        self.pick(1, loc)
        self.bury(1, loc)
        self.binary(op, loc)
        self.bury(1, loc)
        self.assign(loc)

    def binary(self, op, loc):
        '''..., value, value -> ..., value'''
        self._append(2, 1, ('binary', op), loc=loc)

    def box(self, loc):
        '''..., value -> ..., dynamic_value'''
        self._append(1, 1, ('box',), loc=loc)

    def break_(self, loc):
        '''Stack must be empty.'''
        self._check_empty()
        self.is_break_reachable = self.is_end_reachable
        self._append(0, 0, ('break',), loc=loc)
        self.is_end_reachable = False

    def bury(self, n, loc):
        '''
        ..., <n items>, value -> ..., value, <n items>
        bury(0) is a no-op.
        bury(1) is also called 'swap' and 'zy'.
        bury(2) is also called 'zxy'.
        '''
        self._append(n+1, n+1, ('bury', n), loc=loc)

    def call(self, num_args, loc):
        '''..., func, arg1, ... -> ..., result'''
        self._append(1+num_args, 1, ('call', num_args), loc=loc)

    def capture(self, index, loc):
        '''... -> ..., value'''
        self._append(0, 1, ('capture', index), loc=loc)

    def cast(self, loc):
        '''..., value, type -> ..., value'''
        self._append(2, 1, ('cast',), loc)

    def cast_to_type(self, noun, require_static, loc):
        '''
        ..., type -> ..., type

        Cast the value on top of the stack to a type, with helpful error
        messages. The cast implicitly copies the value.

         - noun - Used to construct error messages.
         - require_static - If `True`, additionally checks that the value is
           static and const.
        '''
        self._append(1, 1, ('cast_to_type', noun, require_static), loc=loc)

    def closure(self, body_block, n, loc):
        '''
        ..., param_types_array, <n captures> -> ..., function

         - body_block - The Block representing the body of the function.
         - n - The number of stack slots holding captured values.
        '''
        if body_block.is_end_reachable:
            raise FlatCodeError('Missing "return" instruction')
        self._append(
            1 + n, 1,
            (
                'closure',
                body_block.get_code(is_break_allowed=False),
                n
            ),
            loc=loc,
        )

    def const_char(self, x, loc):
        '''... -> ..., value'''
        if not 0 <= x < 0x110000:
            raise FlatCodeError('Character constant too large')
        self._append(0, 1, ('const_char', x), loc=loc)

    def const_float(self, x, loc):
        '''... -> ..., value'''
        self._append(0, 1, ('const_float', x), loc=loc)

    def const_int(self, x, loc):
        '''... -> ..., value'''
        if not (-1 << 63) <= x < (1 << 63):
            raise FlatCodeError('Integer constant too large')
        self._append(0, 1, ('const_int', x), loc=loc)

    def const_str(self, x, loc):
        '''... -> ..., value'''
        self._append(0, 1, ('const_str', x), loc=loc)

    def continue_(self, loc):
        '''Stack must be empty.'''
        self._check_empty()
        self.is_continue_reachable = self.is_end_reachable
        self._append(0, 0, ('continue',), loc=loc)
        self.is_end_reachable = False

    def defer(self, loc):
        '''..., type -> ..., value'''
        self._append(1, 1, ('defer',), loc=loc)

    def defer_function(self, loc):
        '''
        ..., param_types_array, return_type ->
            ..., defer_cell, param_types_array, return_type

        A CISC-like instruction equivalent to the following sequence: yzy yzy
        union('FUNCTION', 2) defer zxy. This sequence is used when constructing
        a closure for a recursive function.
        '''
        self.yzy(loc)
        self.yzy(loc)
        self.union('FUNCTION', 2, loc)
        self.defer(loc)
        self.zxy(loc)

    def defer_type(self, loc):
        '''
        ... -> ..., defer_cell

        Constructs a defer cell of type `type`.
        '''
        self._append(0, 1, ('defer_type',), loc=loc)

    def dereference(self, loc):
        '''..., reference -> ..., value'''
        self._append(1, 1, ('dereference',), loc=loc)

    def drop(self, loc):
        '''..., z -> ...'''
        self._append(1, 0, ('drop',), loc=loc)

    def dup(self, loc):
        '''
        ..., z -> ..., z, z

        This is actually implemented as: pick(0).
        '''
        self.pick(0, loc)

    def empty_module(self, loc):
        '''... -> ..., module'''
        self._append(0, 1, ('empty_module',), loc=loc)

    def expr(self, loc):
        '''
        value ->

        Treat an expression as a statement. Stack must become empty.
        '''
        self._append(1, 0, ('expr',), loc=loc)
        self._check_empty()

    def field(self, name, loc):
        '''
        ..., container -> ..., member

        Implements the ".field" notation. `container` can be anything with
        named members, e.g. a struct or a module.
        '''
        self._append(1, 1, ('field', name), loc=loc)

    def fulfil(self, loc):
        '''..., rvalue, lvalue -> ...'''
        self._append(2, 0, ('fulfil',), loc=loc)

    def fulfil_closure(self, body_block, num_captures, loc):
        '''
        ..., defer_cell, param_types_array, capture0, ... -> ..., function

        A CISC-like instruction equivalent to the following sequence:
        closure(body_block) yzy fulfil. This sequence is used when
        constructing a closure for a recursive function.
        '''
        # TODO: Give it its own op-code.
        self.closure(body_block, num_captures, loc)
        self.yzy(loc)
        self.fulfil(loc)

    def get(self, name, loc):
        '''
        ..., module -> ..., dynamic_value

        Looks up `name` in `module`, which need not be static. In contrast to
        'field', the result is of type `dynamic`, and `module` is always a
        module.
        '''
        # TODO: Reimplement as a special case of 'subscript'?
        self._append(1, 1, ('get', name), loc=loc)

    def if_(self, n, then_block, else_block, loc):
        '''
        ..., condition -> ..., <n items>

        The `n` items left on the stack by `then_block` or `else_block` are
        appended to this Block's stack.
        '''
        if n != then_block.stack_used:
            raise FlatCodeError(
                '"Then" branch creates the wrong number of values')
        if n != else_block.stack_used:
            raise FlatCodeError(
                '"Else" branch creates the wrong number of values')
        self._append(
            1, n,
            ('if',
                n,
                then_block.get_code(is_break_allowed=True),
                else_block.get_code(is_break_allowed=True),
            ),
            loc=loc,
        )
        if (then_block.is_break_reachable or
            else_block.is_break_reachable
        ):
            self.is_break_reachable = True
        if (then_block.is_continue_reachable or
            else_block.is_continue_reachable
        ):
            self.is_continue_reachable = True
        if (not then_block.is_end_reachable and
            not else_block.is_end_reachable
        ):
            self.is_end_reachable = False

    def if_reachable(self, block, loc):
        '''
        ... -> ...

        If this instruction is reachable then it behaves like 'inline',
        otherwise it is ignored. The only other difference is that `block`
        may not leave any values on the stack.
        '''
        if block.stack_used != 0:
            raise FlatCodeError('"If-Reachable" code creates values')
        if not self.is_end_reachable:
            return
        self._append(0, 0,
            ('if_reachable', block.get_code(is_break_allowed=True)),
            loc=loc,
        )
        if block.is_break_reachable:
            self.is_break_reachable = True
        if block.is_continue_reachable:
            self.is_continue_reachable = True
        if not block.is_end_reachable:
            self.is_end_reachable = False

    def init(self, is_var, loc):
        '''..., value -> ...'''
        self._append(1, 0, ('init', is_var), loc=loc)

    def inline(self, block, loc):
        '''
        ... -> ..., <n items>

        `block` has access to all the captures and locals but not to the stack
        of this Block. If `block` declares new locals, they are private, and
        this Block will not see them. Values left on the stack by `block` are
        appended to this Block's stack. If any 'break', 'continue', 'return'
        or and 'throw' instruction is reachable in `block` then it is
        reachable in this Block.
        '''
        n = block.stack_used
        self._append(0, n,
            ('inline', block.get_code(is_break_allowed=True)),
            loc=loc,
        )
        if block.is_break_reachable:
            self.is_break_reachable = True
        if block.is_continue_reachable:
            self.is_continue_reachable = True
        if not block.is_end_reachable:
            self.is_end_reachable = False

    def local(self, index, loc):
        '''... -> ..., value'''
        self._append(0, 1, ('local', index), loc=loc)

    def loop(self, loop_block, loc):
        '''Stack must be empty.'''
        self._check_empty()
        loop_block._check_empty()
        self._append(0, 0,
            ('loop', loop_block.get_code(is_break_allowed=True)),
            loc=loc,
        )
        if not loop_block.is_break_reachable:
            self.is_end_reachable = False

    def module(self, block, names, loc):
        '''
        ..., module, captures -> ..., new_module

        Invokes the type-checker. If everything is okay, immediately runs the
        code. Any values left on the stack become new global variables.
        Specifically, they are zipped with `names` and used to construct
        `new_module`, containing all the bindings from `module` plus the new
        bindings. `module` is not modified.

        `module` is a value of type module.
        `captures` is an array of values of type dynamic.

         - block - a Block representing the body of the module.
         - names - a list of str.
        '''
        assert block.stack_used == len(names)
        self._append(2, 1,
            ('module', block.get_code(is_break_allowed=False), names),
            loc=loc,
        )

    def new(self, loc):
        '''..., value -> ..., ptr'''
        self._append(1, 1, ('new',), loc=loc)

    def pick(self, pos, loc):
        '''
        x_pos, ..., x_0 -> x_pos, ..., x_0, x_pos
        pick(0) does the same as 'zz', also called 'dup'.
        pick(1) does the same as 'yzy'.
        pick(2) does the same as 'xyzx'.
        '''
        self._append(pos+1, pos+2, ('pick', pos), loc=loc)

    def return_(self, loc):
        '''
        value ->

        Stack must become empty.
        '''
        self._append(1, 0, ('return',), loc=loc)
        self._check_empty()
        self.is_end_reachable = False

    def return_void(self, loc):
        '''
        return_type ->

        Equivalent to the following sequence: struct(0) swap cast return.
        This is special-cased to improve error messages.
        '''
        self._append(1, 0, ('return_void',), loc=loc)
        self._check_empty()
        self.is_end_reachable = False

    def reverse(self, n, loc):
        '''..., item_1, ..., item_n -> ..., item_n, ..., item_1'''
        self._append(n, n, ('reverse', n), loc=loc)

    def struct(self, n, loc):
        '''
        ..., <n items> -> ..., struct_value
        '''
        self._append(n, 1, ('struct', n), loc=loc)

    def subscript(self, loc):
        '''..., collection, index -> ..., value'''
        self._append(2, 1, ('subscript',), loc=loc)

    def swap(self, loc):
        '''
        ..., y, z -> ..., z, y
        This is actually implemented as `bury(1)`.
        '''
        self.bury(1, loc)

    def switch(self, case_quads, else_pair, loc):
        '''
        ..., condition -> ...

         - case_quads - a list of quads, one for each "case" of the
           "switch" statement. Each quad is of the form `(tag, num_params,
           block, case_loc)`, where:
            - tag - the tag of the case.
            - num_params - the number of locals bound by the case.
            - block - a Block containing the code to be executed when the case
              matches.
            - case_loc - a location for use in error messages.
         - else_pair - a (Block, loc) pair or `None`.
        '''
        case_code_quads = []
        for tag, num_params, block, case_loc in case_quads:
            assert type(tag) is str
            assert type(num_params) is int
            assert type(block) is Block
            code = block.get_code(is_break_allowed=True)
            case_code_quads.append((tag, num_params, code, case_loc))
        if else_pair is None:
            else_code_pair = None
        else:
            else_block, else_loc = else_pair
            else_code_pair = (
                else_block.get_code(is_break_allowed=True),
                else_loc,
            )
        self._append(1, 0,
            ('switch', case_code_quads, else_code_pair),
            loc=loc,
        )
        if (any(b.is_break_reachable for _, _, b, _ in case_quads) or
            (else_pair is not None and else_block.is_break_reachable)
        ):
            self.is_break_reachable = True
        if (any(b.is_continue_reachable for _, _, b, _ in case_quads) or
            (else_pair is not None and else_block.is_continue_reachable)
        ):
            self.is_continue_reachable = True
        if (not any(b.is_end_reachable for _, _, b, _ in case_quads) and
            not (else_pair is not None and else_block.is_end_reachable)
        ):
            self.is_end_reachable = False

    def typeof(self, loc):
        '''..., value -> ..., value'''
        self._append(1, 1, ('typeof',), loc=loc)

    def unary(self, op, loc):
        '''..., value -> ..., value'''
        self._append(1, 1, ('unary', op), loc=loc)

    def union(self, tag, n, loc):
        '''..., <n items> -> union_value'''
        self._append(n, 1, ('union', tag, n), loc=loc)

    def unpack_struct(self, n, loc):
        '''
        ..., struct_value -> ..., field_1, ..., field_n

        Checks that `struct_value` is a struct with `n` fields. Leaves the
        field values on the stack.
        '''
        self._append(1, n, ('unpack_struct', n), loc=loc)

    # Permutations of the top of the stack.
    # For each stack depth in {0, 1, 2}, we provide:
    #  - an instruction for copying a value up from that depth.
    #  - an instruction for copying a value down to that depth.
    #  - an instruction for moving a value down to that depth.
    # Except:
    #  - for depth 0, moving is a no-op, so not provided.
    #  - for depth 0, copying up and copying down are equivalent, and
    #    called 'dup'.
    #  - for depth 1, moving down is called 'swap'.
    # For all depths:
    #  - copying up is implemented using 'pick'.
    #  - moving down is implemented using 'bury'.
    #  - copying down is implemented is 'bury' followed by 'pick'.
    # 'pick' and 'bury' can be used directly to access arbitrary depths in
    # the stack.

    def xyzx(self, loc):
        '''
        ..., x, y, z -> ..., x, y, z, x
        This is actually implemented as: pick(2).
        '''
        self.pick(2, loc)

    def yzy(self, loc):
        '''
        ..., y, z -> ..., y, z, y
        This is actually implemented as: pick(1).
        '''
        self.pick(1, loc)

    def zxy(self, loc):
        '''
        ..., x, y, z -> ..., z, x, y
        This is actually implemented as: bury(2).
        '''
        self.bury(2, loc)

    def zxyz(self, loc):
        '''
        ..., x, y, z -> ..., z, x, y, z
        This is actually implemented as: bury(2) pick(2).
        '''
        self.bury(2, loc)
        self.pick(2, loc)

    zy = swap

    def zyz(self, loc):
        '''
        ..., y, z -> ..., z, y, z
        This is actually implemented as: bury(1) pick(1).
        '''
        self.bury(1, loc)
        self.pick(1, loc)

    zz = dup
