from ..model import pack, wtypes

from ..model.memory import Ref, Array
from ..model.ops import Dynamic, array_element_type
from .option import option_type, some, none


def next_type(t):
    '''
    Returns the type of an iterator that returns values of type `t`.
    '''
    return wtypes.function([], option_type(t))


def constructor_type(t):
    '''
    Returns the type of a constructor function that takes no arguments and
    returns an iterator that returns values of type `t`.
    '''
    return wtypes.function([], next_type(t))


def field_type(t, u):
    '''
    Returns the type of a function that takes a value of type `u` and returns
    a constructor of iterators that return values of type `t`.
    '''
    return wtypes.function([u], constructor_type(t))


# Iterators for integers.

# fn int_field(x: int) {
#     fn int_constructor() {
#         var counter = 0
#         fn int_next() {
#             if counter < x {
#                 const ret = counter
#                 counter += 1
#                 return SOME(ret)
#             } else {
#                 return NONE
#             }
#         }
#         return int_next
#     }
#     return int_constructor
# }


def _int_next(captures, args):
    '''
    An int iterator that returns and increments a integer counter if it has not
    reached a stopping value.

    This is intended to be called from Welly code, and it obeys the calling
    convention specified in "doc/python_model.txt".

     - captures - a Python list containing two integers: the number to return
       next and the number at which to stop (exclusive).
     - args - a Ref to no words.
    '''
    counter, end = captures
    assert type(counter) in (int, long)
    assert type(end) in (int, long)
    if counter < end:
        captures[0] += 1
        ret = some(counter)
    else:
        ret = none
    return ('return', ret)


def _int_constructor(captures, args):
    '''
    A function that takes no arguments and returns an int iterator.

    This is intended to be called from Welly code, and it obeys the calling
    convention specified in "doc/python_model.txt".

     - captures - a Python list containing the number at which to stop
       (exclusive).
     - args - a Ref to no words.
    '''
    (end,) = captures
    assert type(end) in (int, long)
    ret = Ref.new((
        _int_next,
        [0, end],
    ))
    return ('return', ret)


def _int_field(captures, args):
    '''
    A function that maps an int to a constructor of int iterators.

    This is intended to be called from Welly code, and it obeys the calling
    convention specified in "doc/python_model.txt".

     - captures - an empty Python list.
     - args - a Ref to the number at which to stop (exclusive).
    '''
    (end,) = args.read()
    assert type(end) in (int, long)
    ret = Ref.new((
        _int_constructor,
        [end],
    ))
    return ('return', ret)


# A Welly function value that takes an int and returns a constructor of int
# iterators.
int_field = Dynamic.new(
    field_type(wtypes.int_, wtypes.int_),
    (_int_field, []),
)


# Iterators for arrays.

# For each type `t`:
# fn array_field(a: array t) {
#     fn array_constructor() {
#         var counter = 0
#         fn array_next() {
#             if counter < a.length {
#                 const ret = a[counter]
#                 counter += 1
#                 return SOME(ret)
#             } else {
#                 return NONE
#             }
#         }
#         return array_next
#     }
#     return array_constructor
# }


def _array_next(captures, args):
    '''
    An iterator that returns (copies of) the elements of an array in order.

    This is intended to be called from Welly code, and it obeys the calling
    convention specified in "doc/python_model.txt".

     - captures - a Python list containing two items: the index of the element
       to return next and the Array.
     - args - a Ref to no words.
    '''
    counter, array = captures
    assert type(counter) in (int, long)
    assert type(array) is Array
    if counter < len(array):
        captures[0] += 1
        ret = some(array[counter].read())
    else:
        ret = none
    return ('return', ret)


def _array_constructor(captures, args):
    '''
    A function that takes no arguments and returns an int iterator.

    This is intended to be called from Welly code, and it obeys the calling
    convention specified in "doc/python_model.txt".

     - captures - a Python list containing the Array.
     - args - a Ref to no words.
    '''
    (array,) = captures
    assert type(array) is Array
    assert args.size == 0
    ret = Ref.new((
        _array_next,
        [0, array],
    ))
    return ('return', ret)


def array_field(array_t):
    '''
    Returns a Dynamic representing a Welly function that takes an `array_t`
    and returns a constructor of `elem_t` iterators, where `elem_t` is the
    element type of `array_t`.

     - array_t - the Welly type of the array.
    '''
    elem_t = array_element_type(array_t)

    # In real Welly, we'd recompile this code for each call site. Here, the
    # closest equivalent is a Python capture of `elem_t`.
    def _array_field(captures, args):
        '''
        This is intended to be called from Welly code, and it obeys the
        calling convention specified in "doc/python_model.txt".

         - captures - an empty Python list.
         - args - a Ref to the array.
        '''
        array = Array.from_welly(args.read(), wtypes.sizeof(elem_t))
        ret = Ref.new((
            _array_constructor,
            [array],
        ))
        return ('return', ret)

    return Dynamic.new(
        field_type(elem_t, array_t),
        (_array_field, []),
    )
