import pytest

from ..printer import pretty
from ..parser import WellyParser

from .flattener import LexicalScope, Module, FlattenerError


def check(source):
    '''
    Flattens `source` to see if it crashes.
    '''
#    print
#    print source
    scope = LexicalScope(Module())

    def do_stmt(ast):
        scope.flatten_stmt(ast, 'the top-level scope')

    p = WellyParser(do_stmt)
    p.feed(source)
    p.flush()
    p.do_pending_stmts()
#    print
#    print pretty(scope)
#    print
#    print '---'


def test_arithmetic():
    check('"Hello, "+who+"!"')


def test_assign():
    check('(x, y) = 1')
    check('(x, y) += 1')


def test_break():
    with pytest.raises(FlattenerError):
        check('break')
    check('while true { break }')
    with pytest.raises(FlattenerError):
        check('while true { break; break }')
    with pytest.raises(FlattenerError):
        check('while true { break; const x = 1 }')
    with pytest.raises(FlattenerError):
        check('while true { break; x += 1 }')


def test_call():
    check('print("Hello, world!")')


def test_closure():
    check('fn dec() { return n - 1 }')
    check('''
      fn factorial(n: int): int {
        fn dec() { return n - 1; }
        if (n == 0) {
          return 1;
        } else {
          return n * factorial(dec());
        }
      }
    ''')


def test_continue():
    with pytest.raises(FlattenerError):
        check('continue')
    check('while true { continue }')


def test_decl():
    check('var who = "world"')


def test_defer():
    check('''
      var j = new defer fn (int): int;
      fulfil *j = (foo: module).f;
    ''')


def test_for():
    check('for x in [1, 2, 3] { x }')


def test_loop():
    check('''
      while x>3 {
        x -= 1;
        break;
      }
    ''')
    check('''
        fn fib(x: int): int {
            while x <= 1 { return x } else { return fib(x-1) + fib(x-2) }
        }
    ''')


def test_module():
    check('''
      module foo {
        const factorial_type = like_to_like(typeof 1)
        // Compilation unit boundary needed here.
        var f = factorial: factorial_type
        const f = f; // "const" version shadows the "var" version.
        const (f, f) = (f, f); // Introduces an anonymous (shadowed) local.
      }
    ''')


def test_switch():
    check('''
      union int_list { NIL(); CONS(int, int_list) }
      const list_of_one: int_list = CONS(1, NIL())

      switch list_of_one {
        case CONS(head, tail) { print("Non-empty.") }
        case NIL { print("Empty.") }
        else { print("Impossible.") }
      }
    ''')


def type_type_array():
    check('''array int''')
    check('''array const int''')
    check('''array var int''')


def test_type_function():
    check('fn like_to_like(t: type) { return fn(t): t }')


def test_type_ref():
    check('''ref int''')
    check('''ref const int''')
    check('''ref var int''')


def test_type_struct():
    check('''
      struct non_empty_int_list (
        int,
        union {
          CONS(non_empty_int_list);
          NIL();
        },
      )
      const non_empty_list_of_one: non_empty_int_list = (1, NIL())
    ''')


def test_type_union():
    check('''
      union int_list { NIL(); CONS(int, int_list) }
      const list_of_one: int_list = CONS(1, NIL())
    ''')


def test_unpack():
    check('''
      var ((who, (x: int)), j) = (
        ("world", -7),
        3.14,
      );
    ''')
