import contextlib
import pytest

from ..model import pack
from ..model.wtypes import *
from ..model.const import BASE, CONST, VAR
from ..printer import pretty

from linker import *
from typed_code import Block, TypedCodeError


# We need to put dummy location information into every instruction.
# This is the value we will use.
loc = None


BIG_FLAT_CODE = [
    ('const_str', u'world', loc),
    ('const_int', 7, loc),
    ('unary', u'-', loc),
    ('struct', 2, loc),
    ('capture', 0, loc),
    ('cast_to_type', 'the parameter type', False, loc),
    ('array', 1, loc),
    ('capture', 0, loc),
    ('cast_to_type', 'the return type', False, loc),
    ('union', 'FUNCTION', 2, loc),
    ('cast_to_type', 'SHOULD NOT HAPPEN', False, loc),
    ('defer', loc),
    ('new', loc),
    ('struct', 2, loc),
    ('unpack_struct', 2, loc),
    ('reverse', 2, loc),
    ('unpack_struct', 2, loc),
    ('reverse', 2, loc),
    ('init', True, loc),
    ('capture', 0, loc),
    ('cast', loc),
    ('init', True, loc),
    ('init', True, loc),
    (
        'loop',
        [
            ('local', 1, loc),
            ('const_int', 3, loc),
            ('binary', u'>', loc),
            (
                'if',
                0,
                [
                    ('capture', 1, loc),
                    ('const_str', u'Hello, ', loc),
                    ('local', 0, loc),
                    ('binary', u'+', loc),
                    ('const_str', u'!', loc),
                    ('binary', u'+', loc),
                    ('call', 1, loc),
                    ('expr', loc),
                    ('const_int', 1, loc),
                    ('local', 1, loc),
                    ('bury', 1, loc),
                    ('pick', 1, loc),
                    ('bury', 1, loc),
                    ('binary', u'-', loc),
                    ('bury', 1, loc),
                    ('assign', loc),
                    ('break', loc),
                ],
                [('break', loc)],
                loc,
            ),
        ],
        loc,
    ),
    ('capture', 0, loc),
    ('cast_to_type', 'the parameter type', True, loc),
    ('array', 1, loc),
    ('capture', 0, loc),
    ('cast_to_type', 'the return type', True, loc),
    ('pick', 1, loc),
    ('pick', 1, loc),
    ('union', 'FUNCTION', 2, loc),
    ('defer', loc),
    ('bury', 2, loc),
    ('pick', 2, loc),
    (
        'closure',
        [
            ('array', 0, loc),
            ('local', 0, loc),
            (
                'closure',
                [
                    ('capture', 0, loc),
                    ('const_int', 1, loc),
                    ('binary', u'-', loc),
                    ('return', loc),
                ],
                1,
                loc,
            ),
            ('init', False, loc),
            ('local', 0, loc),
            ('const_int', 0, loc),
            ('binary', u'==', loc),
            (
                'if',
                0,
                [
                    ('const_int', 1, loc),
                    ('capture', 0, loc),
                    ('cast', loc),
                    ('return', loc),
                ],
                [
                    ('local', 0, loc),
                    ('capture', 1, loc),
                    ('local', 1, loc),
                    ('call', 0, loc),
                    ('call', 1, loc),
                    ('binary', u'*', loc),
                    ('capture', 0, loc),
                    ('cast', loc),
                    ('return', loc),
                ],
                loc,
            ),
        ],
        2,
        loc,
    ),
    ('pick', 1, loc),
    ('fulfil', loc),
    ('init', False, loc),
    ('capture', 2, loc),
    ('cast_to_type', 'the parameter type', True, loc),
    ('array', 1, loc),
    (
        'closure',
        [
            ('local', 0, loc),
            ('cast_to_type', 'the parameter type', False, loc),
            ('array', 1, loc),
            ('local', 0, loc),
            ('cast_to_type', 'the return type', False, loc),
            ('union', 'FUNCTION', 2, loc),
            ('cast_to_type', 'SHOULD NOT HAPPEN', False, loc),
            ('return', loc),
        ],
        0,
        loc,
    ),
    ('init', False, loc),
    ('empty_module', loc),
    ('local', 4, loc),
    ('box', loc),
    ('array', 1, loc),
    (
        'module',
        [
            ('capture', 0, loc),
            ('const_int', 1, loc),
            ('typeof', loc),
            ('call', 1, loc),
            ('init', False, loc),
            ('local', 0, loc),
        ],
        ['factorial_type'],
        loc,
    ),
    ('local', 3, loc),
    ('box', loc),
    ('pick', 1, loc),
    ('get', 'factorial_type', loc),
    ('array', 2, loc),
    (
        'module',
        [
            ('capture', 0, loc),
            ('capture', 1, loc),
            ('cast', loc),
            ('init', True, loc),
            ('local', 0, loc),
        ],
        ['f'],
        loc,
    ),
    ('pick', 0, loc),
    ('get', 'f', loc),
    ('array', 1, loc),
    (
        'module',
        [
            ('capture', 0, loc),
            ('init', False, loc),
            ('local', 0, loc),
        ],
        ['f'],
        loc,
    ),
    ('pick', 0, loc),
    ('get', 'f', loc),
    ('array', 1, loc),
    (
        'module',
        [
            ('capture', 0, loc),
            ('capture', 0, loc),
            ('struct', 2, loc),
            ('unpack_struct', 2, loc),
            ('reverse', 2, loc),
            ('init', False, loc),
            ('init', False, loc),
            ('local', 1, loc),
        ],
        ['f'],
        loc,
    ),
    ('init', False, loc),
    ('local', 5, loc),
    ('union', 'MODULE', 0, loc),
    ('cast_to_type', 'SHOULD NOT HAPPEN', False, loc),
    ('cast', loc),
    # The following is a hack to work around incorrectly treating the source
    # file as a single compilation unit.
    ('drop', loc), ('local', 3, loc), # Was: ('field', 'f')
    ('local', 2, loc),
    ('dereference', loc),
    ('fulfil', loc),
    (
        'inline',
        [
            ('defer_type', loc),
            ('init', False, loc),
            ('const_str', 'NIL', loc),
            ('array', 0, loc),
            ('struct', 2, loc),
            ('const_str', 'CONS', loc),
            ('capture', 0, loc),
            ('cast_to_type', 'the case field type', False, loc),
            ('local', 6, loc),
            ('cast_to_type', 'the case field type', False, loc),
            ('array', 2, loc),
            ('struct', 2, loc),
            ('array', 2, loc),
            ('union', 'UNION', 1, loc),
            ('cast_to_type', 'SHOULD NOT HAPPEN', False, loc),
            ('local', 6, loc),
            ('fulfil', loc),
            ('local', 6, loc),
        ],
        loc,
    ),
    ('init', False, loc),
    ('const_int', 1, loc),
    ('union', 'NIL', 0, loc),
    ('union', 'CONS', 2, loc),
    # Following cast omitted to work around incorrectly treating the source
    # file as a single compilation unit.
    # ('local', 6, loc),
    # ('cast', loc),
    ('init', False, loc),
    (
        'inline',
        [
            ('defer_type', loc),
            ('init', False, loc),
            ('capture', 0, loc),
            ('cast_to_type', 'the field type', False, loc),
            ('const_str', 'CONS', loc),
            ('local', 8, loc),
            ('cast_to_type', 'the case field type', False, loc),
            ('array', 1, loc),
            ('struct', 2, loc),
            ('const_str', 'NIL', loc),
            ('array', 0, loc),
            ('struct', 2, loc),
            ('array', 2, loc),
            ('union', 'UNION', 1, loc),
            ('cast_to_type', 'SHOULD NOT HAPPEN', False, loc),
            ('cast_to_type', 'the field type', False, loc),
            ('array', 2, loc),
            ('union', 'STRUCT', 1, loc),
            ('cast_to_type', 'SHOULD NOT HAPPEN', False, loc),
            ('local', 8, loc),
            ('fulfil', loc),
            ('local', 8, loc),
        ],
        loc,
    ),
    ('init', False, loc),
    ('const_int', 1, loc),
    ('union', 'NIL', 0, loc),
    ('struct', 2, loc),
    # Following cast omitted to work around incorrectly treating the source
    # file as a single compilation unit.
    # ('local', 8, loc),
    # ('cast', loc),
    ('init', False, loc),
    ('local', 7, loc),
    (
        'switch',
        [
            (
                'CONS',
                2,
                [
                    ('capture', 1, loc),
                    ('const_str', u'Non-empty.', loc),
                    ('call', 1, loc),
                    ('expr', loc),
                ],
                loc,
            ),
            (
                'NIL',
                0,
                [
                    ('capture', 1, loc),
                    ('const_str', u'Empty.', loc),
                    ('call', 1, loc),
                    ('expr', loc),
                ],
                loc,
            ),
        ],
        [
            ('capture', 1, loc),
            ('const_str', u'Impossible.', loc),
            ('call', 1, loc),
            ('expr', loc),
        ],
        loc,
    ),
]


@contextlib.contextmanager
def raises(exc_class, message=None):
    with pytest.raises(exc_class) as exc_info:
        yield exc_info
    if message is not None:
        assert exc_info.value.message == message


def dummy_vm_function(typed_code, param_sizes):
    def closure(arg_struct):
        raise NotImplementedError
    return closure


def check(flat_code, stack_len=0):
#    print
#    print "Entering `check()`"
#    print "Flat code:"
#    print pretty(flat_code)
    captures = [
        builtins['int'],
        builtins['print'],
        builtins['type'],
        Dynamic.new(int_, (3,), constness=VAR),
        Dynamic(union([('FOO', [])]), pack.tag('FOO'), constness=VAR),
        builtins['array0'],
        Dynamic.new(module, ({'int': builtins['int']},)),
    ]
    typed_code, constnesses_and_types = link_module_statement(
        flat_code, captures, dummy_vm_function)
#    print "Compiled to:"
#    print pretty(typed_code)
#    print pretty(exit_)
    assert len(constnesses_and_types) == stack_len


# TODO: Break into smaller bits.
def test_big():
    check(BIG_FLAT_CODE)


def test_array():
    # [int]
    check([
        ('capture', 0, loc),
        ('array', 1, loc),
        ('expr', loc),
    ])
    # [FOO]
    check([
        ('union', 'FOO', 0, loc),
        ('array', 1, loc),
        ('expr', loc),
    ])
    # [FOO, BAR]
    check([
        ('union', 'FOO', 0, loc),
        ('union', 'BAR', 0, loc),
        ('array', 2, loc),
        ('expr', loc),
    ])
    # [FOO, FOO(3)]
    with raises(TypeInferenceVMError):
        check([
            ('union', 'FOO', 0, loc),
            ('const_int', 3, loc),
            ('union', 'FOO', 1, loc),
            ('array', 2, loc),
            ('expr', loc),
        ])
    # [FOO, 3]
    with raises(TypeInferenceVMError):
        check([
            ('union', 'FOO', 0, loc),
            ('const_int', 3, loc),
            ('array', 2, loc),
            ('expr', loc),
        ])
    # [FOO, three]
    with raises(TypeInferenceVMError):
        check([
            ('union', 'FOO', 0, loc),
            ('capture', 3, loc),
            ('array', 2, loc),
            ('expr', loc),
        ])
    # [FOO: union{FOO; BAR}, BAR]
    check([
        ('union', 'FOO', 0, loc),
        ('const_str', 'FOO', loc),
        ('array', 0, loc),
        ('struct', 2, loc),
        ('const_str', 'BAR', loc),
        ('array', 0, loc),
        ('struct', 2, loc),
        ('array', 2, loc),
        ('union', 'UNION', 1, loc),
        ('cast_to_type', 'SHOULD NOT HAPPEN', False, loc),
        ('cast', loc),
        ('union', 'BAR', 0, loc),
        ('array', 2, loc),
        ('expr', loc),
    ])


def test_array_concat():
    # [1, 2, 3] + [4, 5]
    check([
        ('const_int', 1, loc),
        ('const_int', 2, loc),
        ('const_int', 3, loc),
        ('array', 3, loc),
        ('const_int', 4, loc),
        ('const_int', 5, loc),
        ('array', 2, loc),
        ('binary', '+', loc),
        ('expr', loc),
    ])
    # [] + [4, 5]
    with raises(
        StaticVMError,
        '"x + y" is not defined when x is an array0; '
        'cast x to the desired array type',
    ):
        check([
            ('array', 0, loc),
            ('const_int', 4, loc),
            ('const_int', 5, loc),
            ('array', 2, loc),
            ('binary', '+', loc),
            ('expr', loc),
        ])
    # [1, 2, 3] + []
    check([
        ('const_int', 1, loc),
        ('const_int', 2, loc),
        ('const_int', 3, loc),
        ('array', 3, loc),
        ('array', 0, loc),
        ('binary', '+', loc),
        ('expr', loc),
    ])
    # [] + []
    with raises(
        StaticVMError,
        '"x + y" is not defined when x is an array0; '
        'cast x to the desired array type',
    ):
        check([
            ('array', 0, loc),
            ('array', 0, loc),
            ('binary', '+', loc),
            ('expr', loc),
        ])
    # ([]: array const int) + [3]
    check([
        ('array', 0, loc),
        ('union', 'CONST', 0, loc),
        ('capture', 0, loc),
        ('union', 'ARRAY', 2, loc),
        ('cast_to_type', 'SHOULD NOT HAPPEN', False, loc),
        ('cast', loc),
        ('const_int', 3, loc),
        ('array', 1, loc),
        ('binary', '+', loc),
        ('expr', loc),
    ])
    # ([]: array var int) + [3]
    # Cannot be constant folded.
    check([
        ('array', 0, loc),
        ('union', 'VAR', 0, loc),
        ('capture', 0, loc),
        ('union', 'ARRAY', 2, loc),
        ('cast_to_type', 'SHOULD NOT HAPPEN', False, loc),
        ('cast', loc),
        ('const_int', 3, loc),
        ('array', 1, loc),
        ('binary', '+', loc),
        ('expr', loc),
    ])
    # [int, type] + [array0]
    # Array elements occupying more than one word.
    check([
        ('capture', 0, loc),
        ('capture', 2, loc),
        ('array', 2, loc),
        ('capture', 5, loc),
        ('array', 1, loc),
        ('binary', '+', loc),
        ('expr', loc),
    ])
    # [1] + ["a"]
    with raises(TypeVMError):
        check([
            ('const_int', 1, loc),
            ('array', 1, loc),
            ('const_str', "a", loc),
            ('array', 1, loc),
            ('binary', '+', loc),
            ('expr', loc),
        ])
    # [FOO] + [BAR]
    with raises(TypeVMError):
        check([
            ('union', 'FOO', 0, loc),
            ('array', 1, loc),
            ('union', 'BAR', 0, loc),
            ('array', 1, loc),
            ('binary', '+', loc),
            ('expr', loc),
        ])
    # [FOO: union {FOO; BAR}] + [BAR]
    check([
        ('union', 'FOO', 0, loc),
        ('const_str', 'FOO', loc),
        ('array', 0, loc),
        ('struct', 2, loc),
        ('const_str', 'BAR', loc),
        ('array', 0, loc),
        ('struct', 2, loc),
        ('array', 2, loc),
        ('union', 'UNION', 1, loc),
        ('cast_to_type', 'SHOULD NOT HAPPEN', False, loc),
        ('cast', loc),
        ('array', 1, loc),
        ('union', 'BAR', 0, loc),
        ('array', 1, loc),
        ('binary', '+', loc),
        ('expr', loc),
    ])
    # ([FOO]: array const union {FOO; BAR}) + [BAR]
    check([
        ('union', 'FOO', 0, loc),
        ('array', 1, loc),
        ('union', 'CONST', 0, loc),
        ('const_str', 'FOO', loc),
        ('array', 0, loc),
        ('struct', 2, loc),
        ('const_str', 'BAR', loc),
        ('array', 0, loc),
        ('struct', 2, loc),
        ('array', 2, loc),
        ('union', 'UNION', 1, loc),
        ('cast_to_type', 'SHOULD NOT HAPPEN', False, loc),
        ('union', 'ARRAY', 2, loc),
        ('cast_to_type', 'SHOULD NOT HAPPEN', False, loc),
        ('cast', loc),
        ('union', 'BAR', 0, loc),
        ('array', 1, loc),
        ('binary', '+', loc),
        ('expr', loc),
    ])
    # var a = [1, 2, 3]
    # a + [4]
    check([
        ('const_int', 1, loc),
        ('const_int', 2, loc),
        ('const_int', 3, loc),
        ('array', 3, loc),
        ('init', True, loc),
        ('local', 0, loc),
        ('const_int', 4, loc),
        ('array', 1, loc),
        ('binary', '+', loc),
        ('expr', loc),
    ])
    # var x: array const int = []
    # x = [6]
    # x += [7]
    check([
        ('array', 0, loc),
        ('union', 'CONST', 0, loc),
        ('capture', 0, loc),
        ('union', 'ARRAY', 2, loc),
        ('cast_to_type', 'SHOULD NOT HAPPEN', False, loc),
        ('cast', loc),
        ('init', True, loc),
        ('const_int', 6, loc),
        ('array', 1, loc),
        ('local', 0, loc),
        ('assign', loc),
        ('const_int', 7, loc),
        ('array', 1, loc),
        ('local', 0, loc),
        ('bury', 1, loc),
        ('pick', 1, loc),
        ('bury', 1, loc),
        ('binary', '+', loc),
        ('bury', 1, loc),
        ('assign', loc),
    ])
    # const a = ([]: array var int) + [1, 2, 3]
    # a[2] = 88
    check([
        ('array', 0, loc),
        ('union', 'VAR', 0, loc),
        ('capture', 0, loc),
        ('union', 'ARRAY', 2, loc),
        ('cast_to_type', 'SHOULD NOT HAPPEN', False, loc),
        ('cast', loc),
        ('const_int', 1, loc),
        ('const_int', 2, loc),
        ('const_int', 3, loc),
        ('array', 3, loc),
        ('binary', '+', loc),
        ('init', False, loc),
        ('const_int', 88, loc),
        ('local', 0, loc),
        ('const_int', 2, loc),
        ('subscript', loc),
        ('assign', loc),
    ])


def test_array_repeat():
    # [3.14] * 7 // static, static.
    check([
        ('const_float', 3.14, loc),
        ('array', 1, loc),
        ('const_int', 7, loc),
        ('binary', '*', loc),
        ('expr', loc),
    ])
    # [3.14] * -7 // static, static.
    with raises(ArrayRepeatVMError):
        check([
            ('const_float', 3.14, loc),
            ('array', 1, loc),
            ('const_int', -7, loc),
            ('binary', '*', loc),
            ('expr', loc),
        ])
    # [3.14] * 7 // static, runtime.
    check([
        ('const_float', 3.14, loc),
        ('array', 1, loc),
        ('const_int', 7, loc),
        ('init', False, loc),
        ('local', 0, loc),
        ('binary', '*', loc),
        ('expr', loc),
    ])
    # [3.14] * 7 // runtime, static.
    check([
        ('const_float', 3.14, loc),
        ('array', 1, loc),
        ('init', False, loc),
        ('local', 0, loc),
        ('const_int', 7, loc),
        ('binary', '*', loc),
        ('expr', loc),
    ])
    # [3.14] * -7 // runtime, static.
    with raises(ArrayRepeatVMError):
        check([
            ('const_float', 3.14, loc),
            ('array', 1, loc),
            ('init', False, loc),
            ('local', 0, loc),
            ('const_int', -7, loc),
            ('binary', '*', loc),
            ('expr', loc),
        ])
    # [3.14] * 7 // runtime, runtime.
    check([
        ('const_float', 3.14, loc),
        ('array', 1, loc),
        ('init', False, loc),
        ('local', 0, loc),
        ('const_int', 7, loc),
        ('init', False, loc),
        ('local', 1, loc),
        ('binary', '*', loc),
        ('expr', loc),
    ])
    # 7 * [3.14] // static, static.
    check([
        ('const_int', 7, loc),
        ('const_float', 3.14, loc),
        ('array', 1, loc),
        ('binary', '*', loc),
        ('expr', loc),
    ])
    # -7 * [3.14] // static, static.
    with raises(ArrayRepeatVMError):
        check([
            ('const_int', -7, loc),
            ('const_float', 3.14, loc),
            ('array', 1, loc),
            ('binary', '*', loc),
            ('expr', loc),
        ])
    # 7 * [3.14] // static, runtime.
    check([
        ('const_int', 7, loc),
        ('const_float', 3.14, loc),
        ('array', 1, loc),
        ('init', False, loc),
        ('local', 0, loc),
        ('binary', '*', loc),
        ('expr', loc),
    ])
    # -7 * [3.14] // static, runtime.
    with raises(ArrayRepeatVMError):
        check([
            ('const_int', -7, loc),
            ('const_float', 3.14, loc),
            ('array', 1, loc),
            ('init', False, loc),
            ('local', 0, loc),
            ('binary', '*', loc),
            ('expr', loc),
        ])
    # 7 * [3.14] // runtime, static.
    check([
        ('const_int', 7, loc),
        ('init', False, loc),
        ('local', 0, loc),
        ('const_float', 3.14, loc),
        ('array', 1, loc),
        ('binary', '*', loc),
        ('expr', loc),
    ])
    # 7 * [3.14] // runtime, runtime.
    check([
        ('const_int', 7, loc),
        ('init', False, loc),
        ('local', 0, loc),
        ('const_float', 3.14, loc),
        ('array', 1, loc),
        ('init', False, loc),
        ('local', 1, loc),
        ('binary', '*', loc),
        ('expr', loc),
    ])


def test_array0_repeat():
    # [] * 7 // static, static.
    check([
        ('array', 0, loc),
        ('const_int', 7, loc),
        ('binary', '*', loc),
        ('expr', loc),
    ])
    # [] * -7 // static, static.
    with raises(ArrayRepeatVMError):
        check([
            ('array', 0, loc),
            ('const_int', -7, loc),
            ('binary', '*', loc),
            ('expr', loc),
        ])
    # [] * 7 // static, runtime.
    check([
        ('array', 0, loc),
        ('const_int', 7, loc),
        ('init', False, loc),
        ('local', 0, loc),
        ('binary', '*', loc),
        ('expr', loc),
    ])
    # [] * 7 // runtime, static.
    check([
        ('array', 0, loc),
        ('init', False, loc),
        ('local', 0, loc),
        ('const_int', 7, loc),
        ('binary', '*', loc),
        ('expr', loc),
    ])
    # [] * -7 // runtime, static.
    with raises(ArrayRepeatVMError):
        check([
            ('array', 0, loc),
            ('init', False, loc),
            ('local', 0, loc),
            ('const_int', -7, loc),
            ('binary', '*', loc),
            ('expr', loc),
        ])
    # [] * 7 // runtime, runtime.
    check([
        ('array', 0, loc),
        ('init', False, loc),
        ('local', 0, loc),
        ('const_int', 7, loc),
        ('init', False, loc),
        ('local', 1, loc),
        ('binary', '*', loc),
        ('expr', loc),
    ])
    # 7 * [] // static, static.
    check([
        ('const_int', 7, loc),
        ('array', 0, loc),
        ('binary', '*', loc),
        ('expr', loc),
    ])
    # -7 * [] // static, static.
    with raises(ArrayRepeatVMError):
        check([
            ('const_int', -7, loc),
            ('array', 0, loc),
            ('binary', '*', loc),
            ('expr', loc),
        ])
    # 7 * [] // static, runtime.
    check([
        ('const_int', 7, loc),
        ('array', 0, loc),
        ('init', False, loc),
        ('local', 0, loc),
        ('binary', '*', loc),
        ('expr', loc),
    ])
    # -7 * [] // static, runtime.
    with raises(ArrayRepeatVMError):
        check([
            ('const_int', -7, loc),
            ('array', 0, loc),
            ('init', False, loc),
            ('local', 0, loc),
            ('binary', '*', loc),
            ('expr', loc),
        ])
    # 7 * [] // runtime, static.
    check([
        ('const_int', 7, loc),
        ('init', False, loc),
        ('local', 0, loc),
        ('array', 0, loc),
        ('binary', '*', loc),
        ('expr', loc),
    ])
    # 7 * [] // runtime, runtime.
    check([
        ('const_int', 7, loc),
        ('init', False, loc),
        ('local', 0, loc),
        ('array', 0, loc),
        ('init', False, loc),
        ('local', 1, loc),
        ('binary', '*', loc),
        ('expr', loc),
    ])


def test_binary():
    check([
        ('const_int', 2, loc),
        ('const_int', 2, loc),
        ('binary', '+', loc),
        ('expr', loc),
    ])
    check([
        ('const_int', 2, loc),
        ('init', True, loc),
        ('local', 0, loc),
        ('const_int', 3, loc),
        ('binary', '+', loc),
        ('expr', loc),
    ])
    check([
        ('const_int', 2, loc),
        ('init', True, loc),
        ('const_int', 3, loc),
        ('local', 0, loc),
        ('binary', '+', loc),
        ('expr', loc),
    ])
    check([
        ('capture', 3, loc),
        ('const_int', 2, loc),
        ('binary', '+', loc),
        ('expr', loc),
    ])


def test_cast():
    # 3: int
    check([
        ('const_int', 3, loc),
        ('capture', 0, loc),
        ('cast', loc),
    ], stack_len=1)
    # three: int
    check([
        ('capture', 3, loc),
        ('capture', 0, loc),
        ('cast', loc),
    ], stack_len=1)
    # []: array int
    check([
        ('array', 0, loc),
        ('union', 'BASE', 0, loc),
        ('capture', 0, loc),
        ('union', 'ARRAY', 2, loc),
        ('cast_to_type', 'SHOULD NOT HAPPEN', False, loc),
        ('cast', loc),
        ('expr', loc),
    ])
    # []: array var int
    check([
        ('array', 0, loc),
        ('union', 'VAR', 0, loc),
        ('capture', 0, loc),
        ('union', 'ARRAY', 2, loc),
        ('cast_to_type', 'SHOULD NOT HAPPEN', False, loc),
        ('cast', loc),
        ('expr', loc),
    ])
    # []: array const int
    check([
        ('array', 0, loc),
        ('union', 'CONST', 0, loc),
        ('capture', 0, loc),
        ('union', 'ARRAY', 2, loc),
        ('cast_to_type', 'SHOULD NOT HAPPEN', False, loc),
        ('cast', loc),
        ('expr', loc),
    ])
    # [3]: array0
    with raises(TypeVMError):
        check([
            ('const_int', 3, loc),
            ('array', 1, loc),
            ('capture', 5, loc),
            ('cast', loc),
            ('expr', loc),
        ])
    # [FOO]: array const union { FOO }
    check([
        ('union', 'FOO', 0, loc),
        ('array', 1, loc),
        ('union', 'CONST', 0, loc),
        ('const_str', 'FOO', loc),
        ('array', 0, loc),
        ('struct', 2, loc),
        ('array', 1, loc),
        ('union', 'UNION', 1, loc),
        ('cast_to_type', 'SHOULD NOT HAPPEN', False, loc),
        ('union', 'ARRAY', 2, loc),
        ('cast_to_type', 'SHOULD NOT HAPPEN', False, loc),
        ('cast', loc),
        ('expr', loc),
    ])
    # [FOO]: array const union { FOO, BAR }
    check([
        ('union', 'FOO', 0, loc),
        ('array', 1, loc),
        ('union', 'CONST', 0, loc),
        ('const_str', 'FOO', loc),
        ('array', 0, loc),
        ('struct', 2, loc),
        ('const_str', 'BAR', loc),
        ('array', 0, loc),
        ('struct', 2, loc),
        ('array', 2, loc),
        ('union', 'UNION', 1, loc),
        ('cast_to_type', 'SHOULD NOT HAPPEN', False, loc),
        ('union', 'ARRAY', 2, loc),
        ('cast_to_type', 'SHOULD NOT HAPPEN', False, loc),
        ('cast', loc),
        ('expr', loc),
    ])


def test_closure():
    check([
        ('array', 0, loc),
        ('capture', 0, loc),
        ('closure',
            [
                ('capture', 0, loc),
                ('return', loc),
            ],
            1,
            loc,
        ),
        ('expr', loc),
    ])
    check([
        ('array', 0, loc),
        ('capture', 0, loc),
        ('init', True, loc),
        ('local', 0, loc),
        ('closure',
            [
                ('capture', 0, loc),
                ('return', loc),
            ],
            1,
            loc,
        ),
        ('expr', loc),
    ])


def test_const():
    check([
        ('const_char', ord("1"), loc),
        ('const_float', 1.0, loc),
        ('const_int', 1, loc),
        ('const_str', "1", loc),
    ], stack_len=4)


def test_get():
    check([
        ('const_int', 1, loc),
        ('capture', 6, loc),
        ('get', 'int', loc),
        ('field', 'value', loc),
        ('cast', loc),
        ('expr', loc),
    ])
    check([
        ('empty_module', loc),
        ('get', 'int', loc),
        ('expr', loc),
    ])


def test_inline():
    check([
        ('inline', [], loc),
    ])
    check([
        ('inline', [
            ('const_int', 1, loc),
        ], loc),
        ('expr', loc),
    ])


def test_module():
    check([
        ('empty_module', loc),
        ('capture', 0, loc),
        ('box', loc),
        ('array', 1, loc),
        (
            'module',
            [
                ('capture', 0, loc),
                ('init', False, loc),
                ('local', 0, loc),
            ],
            ['f'],
            loc,
        ),
        ('expr', loc),
    ])


def test_return():
    # "return" outside a function.
    with raises(
        IllegalInstructionVMError,
        '"return" statement is not in a function',
    ):
        check([
            ('struct', 0, loc),
            ('return', loc),
        ])
    # fn() { return }
    check([
        ('array', 0, loc),
        ('closure', [
                ('struct', 0, loc),
                ('return', loc),
            ],
            0,
            loc
        ),
        ('expr', loc),
    ])


def test_switch():
    # Static cases.
    # switch 1 {}
    with raises(SwitchTypeVMError):
        check([
            ('const_int', 1, loc),
            ('switch', [], None, loc),
        ])
    # switch FOO {}
    with raises(InexhaustiveSwitchVMError):
        check([
            ('union', 'FOO', 0, loc),
            ('switch', [], None, loc),
        ])
    # switch FOO(3) { case FOO() {1} }
    with raises(NumberOfFieldsVMError):
        check([
            ('const_int', 3, loc),
            ('union', 'FOO', 1, loc),
            ('switch', [
                ('FOO', 0, [('const_int', 1, loc), ('expr', loc)], loc),
            ], None, loc),
        ])
    # switch FOO(3) { case FOO(x, y) {y} }
    with raises(NumberOfFieldsVMError):
        check([
            ('const_int', 3, loc),
            ('union', 'FOO', 1, loc),
            ('switch', [
                ('FOO', 2, [('local', 1, loc), ('expr', loc)], loc),
            ], None, loc),
        ])
    # switch FOO { case FOO {1} }
    check([
        ('union', 'FOO', 0, loc),
        ('switch', [
            ('FOO', 0, [('const_int', 1, loc), ('expr', loc)], loc),
        ], None, loc),
    ])
    # switch FOO { case BAR {2} }
    with raises(InexhaustiveSwitchVMError):
        check([
            ('union', 'FOO', 0, loc),
            ('switch', [
                ('BAR', 0, [('const_int', 2, loc), ('expr', loc)], loc),
            ], None, loc),
        ])
    # switch FOO { case FOO {1} case BAR {2} }
    check([
        ('union', 'FOO', 0, loc),
        ('switch', [
            ('FOO', 0, [('const_int', 1, loc), ('expr', loc)], loc),
            ('BAR', 0, [('const_int', 2, loc), ('expr', loc)], loc),
        ], None, loc),
    ])
    # switch FOO { else {3} }
    check([
        ('union', 'FOO', 0, loc),
        ('switch',
            [],
            ([('const_int', 3, loc), ('expr', loc)], loc),
            loc,
        ),
    ])
    # switch FOO { case FOO {1} else {3} }
    check([
        ('union', 'FOO', 0, loc),
        ('switch',
            [('FOO', 0, [('const_int', 1, loc), ('expr', loc)], loc)],
            ([('const_int', 3, loc), ('expr', loc)], loc),
            loc,
        ),
    ])
    # switch FOO { case BAR {2} else {3} }
    check([
        ('union', 'FOO', 0, loc),
        ('switch',
            [('BAR', 0, [('const_int', 2, loc), ('expr', loc)], loc)],
            ([('const_int', 3, loc), ('expr', loc)], loc),
            loc,
        ),
    ])
    # switch FOO { case FOO {1} case BAR {2} else {3} }
    check([
        ('union', 'FOO', 0, loc),
        ('switch',
            [
                ('FOO', 0, [('const_int', 1, loc), ('expr', loc)], loc),
                ('BAR', 0, [('const_int', 2, loc), ('expr', loc)], loc),
            ],
            ([('const_int', 3, loc), ('expr', loc)], loc),
            loc,
        ),
    ])
    # switch FOO(int, 2, "3", 4.0) { case FOO(w, x, y, z) {x} }
    check([
        ('capture', 0, loc),
        ('const_int', 2, loc),
        ('const_str', u'3', loc),
        ('const_float', 4.0, loc),
        ('union', 'FOO', 4, loc),
        ('switch', [
            ('FOO', 4, [('local', 1, loc), ('expr', loc)], loc),
        ], None, loc),
    ])

    # Dynamic cases.
    # switch FOO(three) {}
    with raises(InexhaustiveSwitchVMError):
        check([
            ('capture', 3, loc),
            ('union', 'FOO', 1, loc),
            ('switch', [], None, loc),
        ])
    # switch FOO(three) { case FOO() {1} }
    with raises(NumberOfFieldsVMError):
        check([
            ('capture', 3, loc),
            ('union', 'FOO', 1, loc),
            ('switch', [
                ('FOO', 0, [('const_int', 1, loc), ('expr', loc)], loc),
            ], None, loc),
        ])
    # switch FOO(three) { case FOO(x, y) {y} }
    with raises(NumberOfFieldsVMError):
        check([
            ('capture', 3, loc),
            ('union', 'FOO', 1, loc),
            ('switch', [
                ('FOO', 2, [('local', 1, loc), ('expr', loc)], loc),
            ], None, loc),
        ])
    # switch FOO(three) { case FOO(x) {1} }
    check([
        ('capture', 3, loc),
        ('union', 'FOO', 1, loc),
        ('switch', [
            ('FOO', 1, [('const_int', 1, loc), ('expr', loc)], loc),
        ], None, loc),
    ])
    # switch FOO(three) { case BAR {2} }
    with raises(UnreachableCaseVMError):
        check([
            ('capture', 3, loc),
            ('union', 'FOO', 1, loc),
            ('switch', [
                ('BAR', 0, [('const_int', 2, loc), ('expr', loc)], loc),
            ], None, loc),
        ])
    # switch FOO(three) { case FOO(x) {1} case BAR {2} }
    with raises(UnreachableCaseVMError):
        check([
            ('capture', 3, loc),
            ('union', 'FOO', 1, loc),
            ('switch', [
                ('FOO', 1, [('const_int', 1, loc), ('expr', loc)], loc),
                ('BAR', 0, [('const_int', 2, loc), ('expr', loc)], loc),
            ], None, loc),
        ])
    # switch FOO(three) { else {3} }
    check([
        ('capture', 3, loc),
        ('union', 'FOO', 1, loc),
        ('switch',
            [],
            ([('const_int', 3, loc), ('expr', loc)], loc),
            loc,
        ),
    ])
    # switch FOO(three) { case FOO(x) {1} else {3} }
    with raises(UnreachableCaseVMError):
        check([
            ('capture', 3, loc),
            ('union', 'FOO', 1, loc),
            ('switch',
                [('FOO', 1, [('const_int', 1, loc), ('expr', loc)], loc)],
                ([('const_int', 3, loc), ('expr', loc)], loc),
                loc,
            ),
        ])
    # switch FOO(three) { case BAR {2} else {3} }
    with raises(UnreachableCaseVMError):
        check([
            ('capture', 3, loc),
            ('union', 'FOO', 1, loc),
            ('switch',
                [('BAR', 0, [('const_int', 2, loc), ('expr', loc)], loc)],
                ([('const_int', 3, loc), ('expr', loc)], loc),
                loc,
            ),
        ])
    # switch FOO(three) { case FOO(x) {1} case BAR {2} else {3} }
    with raises(UnreachableCaseVMError):
        check([
            ('capture', 3, loc),
            ('union', 'FOO', 1, loc),
            ('switch',
                [
                    ('FOO', 1, [('const_int', 1, loc), ('expr', loc)], loc),
                    ('BAR', 0, [('const_int', 2, loc), ('expr', loc)], loc),
                ],
                ([('const_int', 3, loc), ('expr', loc)], loc),
                loc,
            ),
        ])
    # switch FOO(int, 2, three, 4.0) { case FOO(w, x, y, z) {x} }
    check([
        ('capture', 0, loc),
        ('const_int', 2, loc),
        ('capture', 3, loc),
        ('const_float', 4.0, loc),
        ('union', 'FOO', 4, loc),
        ('switch', [
            ('FOO', 4, [('local', 1, loc), ('expr', loc)], loc),
        ], None, loc),
    ])

    # Regression tests.
    # Check that empty "else" block is treated as present.
    check([
        ('union', 'FOO', 0, loc),
        ('switch', [], ([], loc), loc),
    ])
    # Check that static var condition is reified.
    check([
        ('capture', 4, loc),
        ('switch', [], ([], loc), loc),
    ])


def test_static():
    # We use 'defer' as an example of something that requires a static value.
    # Test 'local' instruction.
    # fn(t: int) { return defer t }
    with raises(
        StaticnessVMError,
        'the type used for "defer" must be static',
    ):
        check([
            ('capture', 0, loc),
            ('cast_to_type', 'the parameter type', True, loc),
            ('array', 1, loc),
            ('closure', [
                ('local', 0, loc),
                ('defer', loc),
                ('return', loc),
            ], 0, loc),
            ('expr', loc),
        ])

    # Test 'capture' instruction.
    # fn(t: int) { return fn() { return defer t } }
    with raises(
        StaticnessVMError,
        'the type used for "defer" must be static',
    ):
        check([
            ('capture', 0, loc),
            ('cast_to_type', 'the parameter type', True, loc),
            ('array', 1, loc),
            ('closure', [
                ('array', 0, loc),
                ('local', 0, loc),
                ('closure', [
                    ('capture', 0, loc),
                    ('defer', loc),
                    ('return', loc),
                ], 1, loc),
                ('return', loc),
            ], 0, loc),
            ('expr', loc),
        ])


def test_subscript_array():
    # Static cases.
    with raises(SubscriptTypeVMError):
        check([
            ('array', 0, loc),
            ('const_int', 0, loc),
            ('subscript', loc),
            ('expr', loc),
        ])
    with raises(SubscriptTypeVMError):
        check([
            ('array', 0, loc),
            ('const_int', -1, loc),
            ('subscript', loc),
            ('expr', loc),
        ])
    with raises(ArrayIndexVMError):
        check([
            ('const_int', 3, loc),
            ('const_int', 4, loc),
            ('array', 2, loc),
            ('const_int', -1, loc),
            ('subscript', loc),
            ('expr', loc),
        ])
    check([
        ('const_int', 3, loc),
        ('const_int', 4, loc),
        ('array', 2, loc),
        ('const_int', 0, loc),
        ('subscript', loc),
        ('expr', loc),
    ])
    check([
        ('const_int', 3, loc),
        ('const_int', 4, loc),
        ('array', 2, loc),
        ('const_int', 1, loc),
        ('subscript', loc),
        ('expr', loc),
    ])
    with raises(ArrayIndexVMError):
        check([
            ('const_int', 3, loc),
            ('const_int', 4, loc),
            ('array', 2, loc),
            ('const_int', 2, loc),
            ('subscript', loc),
            ('expr', loc),
        ])
    # Dynamic cases.
    check([
        ('capture', 3, loc),
        ('const_int', 4, loc),
        ('array', 2, loc),
        ('const_int', -1, loc),
        ('subscript', loc),
        ('expr', loc),
    ])
    check([
        ('capture', 3, loc),
        ('const_int', 4, loc),
        ('array', 2, loc),
        ('const_int', 0, loc),
        ('subscript', loc),
        ('expr', loc),
    ])
    check([
        ('capture', 3, loc),
        ('const_int', 4, loc),
        ('array', 2, loc),
        ('const_int', 1, loc),
        ('subscript', loc),
        ('expr', loc),
    ])
    check([
        ('capture', 3, loc),
        ('const_int', 4, loc),
        ('array', 2, loc),
        ('const_int', 2, loc),
        ('subscript', loc),
        ('expr', loc),
    ])
    # Static array, dynamic index.
    check([
        ('const_int', 3, loc),
        ('const_int', 4, loc),
        ('array', 2, loc),
        ('capture', 3, loc),
        ('subscript', loc),
        ('expr', loc),
    ])
    # Dynamic array, dynamic index.
    check([
        ('capture', 3, loc),
        ('const_int', 4, loc),
        ('array', 2, loc),
        ('capture', 3, loc),
        ('subscript', loc),
        ('expr', loc),
    ])


def test_subscript_struct():
    # Static cases.
    with raises(StructIndexVMError):
        check([
            ('struct', 0, loc),
            ('const_int', 0, loc),
            ('subscript', loc),
            ('expr', loc),
        ])
    with raises(StructIndexVMError):
        check([
            ('struct', 0, loc),
            ('const_int', -1, loc),
            ('subscript', loc),
            ('expr', loc),
        ])
    with raises(StructIndexVMError):
        check([
            ('const_int', 3, loc),
            ('const_int', 4, loc),
            ('struct', 2, loc),
            ('const_int', -1, loc),
            ('subscript', loc),
            ('expr', loc),
        ])
    check([
        ('const_int', 3, loc),
        ('const_int', 4, loc),
        ('struct', 2, loc),
        ('const_int', 0, loc),
        ('subscript', loc),
        ('expr', loc),
    ])
    check([
        ('const_int', 3, loc),
        ('const_int', 4, loc),
        ('struct', 2, loc),
        ('const_int', 1, loc),
        ('subscript', loc),
        ('expr', loc),
    ])
    with raises(StructIndexVMError):
        check([
            ('const_int', 3, loc),
            ('const_int', 4, loc),
            ('struct', 2, loc),
            ('const_int', 2, loc),
            ('subscript', loc),
            ('expr', loc),
        ])
    # Dynamic cases.
    with raises(StructIndexVMError):
        check([
            ('capture', 3, loc),
            ('const_int', 4, loc),
            ('struct', 2, loc),
            ('const_int', -1, loc),
            ('subscript', loc),
            ('expr', loc),
        ])
    check([
        ('capture', 3, loc),
        ('const_int', 4, loc),
        ('struct', 2, loc),
        ('const_int', 0, loc),
        ('subscript', loc),
        ('expr', loc),
    ])
    check([
        ('capture', 3, loc),
        ('const_int', 4, loc),
        ('struct', 2, loc),
        ('const_int', 1, loc),
        ('subscript', loc),
        ('expr', loc),
    ])
    with raises(StructIndexVMError):
        check([
            ('capture', 3, loc),
            ('const_int', 4, loc),
            ('struct', 2, loc),
            ('const_int', 2, loc),
            ('subscript', loc),
            ('expr', loc),
        ])


def test_ternary():
    check([
        ('const_int', 1, loc),
        ('const_int', 2, loc),
        ('binary', '<', loc),
        ('if', 1,
            [('const_int', 3, loc)],
            [('const_int', 4, loc)],
            loc,
        ),
        ('expr', loc),
    ])
    check([
        ('const_int', 1, loc),
        ('const_int', 2, loc),
        ('binary', '>', loc),
        ('if', 1,
            [('const_int', 3, loc)],
            [('const_int', 4, loc)],
            loc,
        ),
        ('expr', loc),
    ])
    check([
        ('const_int', 1, loc),
        ('const_int', 2, loc),
        ('binary', '<', loc),
        ('if', 1,
            [('const_int', 3, loc)],
            [('const_int', 4, loc)],
            loc,
        ),
        ('expr', loc),
    ])
    check([
        ('const_int', 1, loc),
        ('const_int', 2, loc),
        ('binary', '<', loc),
        ('init', True, loc),
        ('local', 0, loc),
        ('if', 1,
            [('const_int', 3, loc), ('init', True, loc), ('local', 1, loc)],
            [('const_int', 4, loc)],
            loc,
        ),
        ('expr', loc),
    ])
    check([
        ('const_int', 1, loc),
        ('const_int', 2, loc),
        ('binary', '<', loc),
        ('init', True, loc),
        ('local', 0, loc),
        ('if', 1,
            [('const_int', 3, loc)],
            [('const_int', 4, loc)],
            loc,
        ),
        ('expr', loc),
    ])
    check([
        ('const_int', 1, loc),
        ('const_int', 2, loc),
        ('binary', '<', loc),
        ('init', True, loc),
        ('local', 0, loc),
        ('if', 1,
            [('const_int', 3, loc), ('init', True, loc), ('local', 1, loc)],
            [('const_int', 4, loc)],
            loc,
        ),
        ('expr', loc),
    ])


def test_union():
    check([
        ('const_int', 3, loc),
        ('union', 'SOME', 1, loc),
        ('expr', loc),
    ])
    check([
        ('capture', 3, loc), # `three`.
        ('union', 'SOME', 1, loc),
        ('expr', loc),
    ])


def test_unpack():
    check([
        ('struct', 0, loc),
        ('unpack_struct', 0, loc),
    ])
    with raises(NumberOfFieldsVMError):
        check([
            ('struct', 0, loc),
            ('unpack_struct', 1, loc),
        ])
    with raises(NumberOfFieldsVMError):
        check([
            ('const_int', 1, loc),
            ('struct', 1, loc),
            ('unpack_struct', 0, loc),
        ])


def test_while():
    with raises(TypedCodeError, 'Illegal "break" instruction'):
        check([
            ('break', loc),
        ])
    with raises(TypedCodeError, 'Illegal "continue" instruction'):
        check([
            ('continue', loc),
        ])
    check([
        ('loop', [('break', loc)], loc),
    ])
    check([
        ('loop', [('continue', loc)], loc),
    ])
