from ..model import const, wtypes, subtypes, ops, defer

from ..model.errors import *
from ..model.memory import Array
from ..model.pack import pack, tag
from ..model.const import BASE, CONST, VAR
from ..model.wtypes import sizeof
from ..model.ops import Dynamic
from ..compiler.typed_code import TypedCodeError, Slot, Block


def array(elem_type, items, loc):
    '''
    ..., item1, ... -> ..., array_value

     - elem_type - Welly type.
     - items - Python list of Dynamic.
    '''
    array = Array.malloc(len(items), sizeof(elem_type))
    for item, elem_ref in zip(items, array):
        ops.assign(Dynamic(elem_type, elem_ref), item, loc)
    array_type = wtypes.array(CONST, elem_type)
    return Dynamic.new(array_type, pack(array))


def array_concat(left_array, right_array, loc):
    '''
    ..., left_array, right_array -> ..., array

    The result has the same element type as `left_array`, which must not be
    of type `array0`. `right_array` can be of type `array0`. Otherwise, its
    element type must be a subtype of the result element type. Both
    `left_array` and `right_array` must have const elements (if any).
    '''
    assert left_array.is_const()
    assert right_array.is_const()
    # Unpack `left_array.type_`.
    left_tag, left_payload = wtypes.switch(left_array.type_)
    if left_tag == 'ARRAY0':
        raise TypedCodeError('left operand is an array0')
    if left_tag != 'ARRAY':
        raise TypedCodeError('left operand is not an array')
    left_constness, left_elem_t = left_payload
    assert const.switch(left_constness) == 'CONST'
    elem_size = sizeof(left_elem_t)
    left_array = Array.from_welly(left_array.read(), elem_size)
    # Unpack `right_array.type_`.
    right_tag, right_payload = wtypes.switch(right_array.type_)
    if right_tag == 'ARRAY0':
        right_array = Array.malloc(0, elem_size)
    elif right_tag == 'ARRAY':
        right_constness, right_elem_t = right_payload
        assert const.switch(right_constness) == 'CONST'
        subtypes.check(right_elem_t, left_elem_t, 'the element type', loc)
        right_array = Array.from_welly(right_array.read(), elem_size)
    else:
        raise TypedCodeError('right operand is not an array')
    # Construct the result and return it.
    result = Array.malloc(len(left_array) + len(right_array), elem_size)
    out = iter(result)
    for elem in left_array.read():
        next(out).write(elem)
    for elem in right_array.read():
        next(out).write(elem)
    for ref in out:
        assert False
    return Dynamic.new(wtypes.array(CONST, left_elem_t), result.to_welly())


def array_length(array):
    '''..., array -> ..., length'''
    assert array.is_const()
    t0, payload = wtypes.switch(array.type_)
    assert t0 == 'ARRAY'
    constness, elem_type = payload
    elem_size = sizeof(elem_type)
    a = Array.from_welly(array.ref.read(), elem_size)
    result = len(a)
    return Dynamic.new(wtypes.int_, (result,))


def array_repeat(array, count, loc):
    '''
    ..., array, count -> ..., array

    The result has the same constness and element type as `array`,
    which must not be of type `array0`. `count` is an integer.
    '''
    assert array.is_const()
    assert count.is_const()
    # Unpack `array.type_`.
    array_tag, array_payload = wtypes.switch(array.type_)
    if array_tag == 'ARRAY0':
        raise TypedCodeError('left operand is an array0')
    if array_tag != 'ARRAY':
        raise TypedCodeError('left operand is not an array')
    constness, elem_t = array_payload
    assert const.switch(constness) == 'CONST'
    elem_size = sizeof(elem_t)
    array = Array.from_welly(array.read(), elem_size)
    # Unpack `count.type_`.
    count.check_is_a(wtypes.int_, 'the repeat count', loc)
    (count,) = count.read()
    if count < 0:
        raise ArrayRepeatVMError(count)
    # Construct the result and return it.
    result = Array.malloc(len(array) * count, elem_size)
    out = iter(result)
    for i in xrange(count):
        for elem in array.read():
            next(out).write(elem)
    for ref in out:
        assert False
    return Dynamic.new(wtypes.array(CONST, elem_t), result.to_welly())


def array_slice(array, start, end, loc):
    '''
    ..., array, start, end -> ..., subarray

    The result has the same constness and element type as `array` and
    shares storage with it.
    '''
    assert array.is_const()
    assert start.is_const()
    assert end.is_const()
    # Unpack `array`.
    a_tag, a_payload = wtypes.switch(array.type_)
    if a.tag == 'ARRAY0':
        raise TypedCodeError('operand is an array0')
    if a.tag != 'ARRAY':
        raise TypedCodeError('operand is not an array')
    constness, elem_t = a_payload
    # Unpack `start` and `end`.
    start.check_is_a(wtypes.int_, 'the start index', loc)
    (start,) = start.read()
    end.check_is_a(wtypes.int_, 'the end index', loc)
    (end,) = end.read()
    # Check range.
    if 0 > start or start > end or end > len(array):
        raise ArraySliceVMError(start, end, len(array))
    # Construct the result and return it.
    result = array.slice(start, end)
    return Dynamic.new(array.type_, result.to_welly())


def array0():
    '''
    ... -> ..., empty_array

    Each use of this instruction constructs a fresh empty array.
    '''
    return Dynamic.new(wtypes.array0, Array.malloc(0, 0).to_welly())


def array0_repeat(array, count, loc):
    '''
    ..., array0, count -> ..., array0

    Does the same as `array_repeat()`, but applies different type checking.
    `array0` is an `array0`. `count` is an integer.
    '''
    assert array.is_const()
    assert count.is_const()
    # Unpack `array.type_`.
    array_tag, _ = wtypes.switch(array.type_)
    if array_tag != 'ARRAY0':
        raise TypedCodeError('left operand is not an array0')
    # Unpack `count.type_`.
    subtypes.check(count.type_, wtypes.int_, 'the repeat count', loc)
    (count,) = count.read()
    if count < 0:
        raise ArrayRepeatVMError(count)
    # Construct the result and return it.
    return array0()


def binary(op, x, y, loc):
    '''
    ..., value, value -> ..., value

     - op - a welly.compiler.operators.Operator.
    '''
    assert x.is_const() and y.is_const()
    x_t, y_t = op.in_types
    x.check_is_a(x_t, None, loc)
    y.check_is_a(y_t, None, loc)
    (x_value,) = x.read()
    (y_value,) = y.read()
    result = op.implementation(x_value, y_value)
    return Dynamic.new(op.result_type, (result,))


def box(value):
    '''
    ..., value -> ..., dynamic_value

     - value - Dynamic.

    Returns a Dynamic containing `dynamic_value`, i.e. two levels of dynamism.
    '''
    return Dynamic.new(wtypes.dynamic, value.to_welly())


def cast(type_, value, loc):
    '''..., value -> ..., value'''
    assert value.is_const()
    return value.copy(loc, type_=type_)


def closure(make_body, param_types, vm_function, loc):
    '''
    ..., capture0, ... -> ..., function

     - make_body - a callback that generates the body of the function.
       The callback is given a Block and populates it by calling its methods.
     - param_types - an Array of Welly types.
     - vm_function - a callback that makes functions. It takes:
        - typed_code - a list of typed code instructions representing the body
          of the function.
        - param_sizes - an Array of the sizes in words of the function
          parameters.
       and returns a function value, obeying "doc/python_model.txt". This
       callback effectively defines which VM implementation will run the
       function when it is called.
     - loc - an opaque source code location for use in error messages.

    There is no `n` parameter because the number of runtime captured values
    is zero.
    '''
    assert type(param_types) is Array
    for t in param_types:
        wtypes.check_valid(t)
    # Generate the typed code for the body.
    param_slots = [Slot(CONST, t) for t in param_types]
    block = Block(locals_=param_slots, captures=[])
    make_body(block)
    if block.is_end_reachable:
        raise TypedCodeError('Missing "return" instruction')
    # Construct the closure value
    closure = vm_function(
        block.get_code(is_break_allowed=False),
        Array.from_list(
            [(sizeof(t),) for t in param_types],
            sizeof(wtypes.int_),
        ),
    )
    # Box the closure and put it on the stack.
    return_type = block.get_return_type(loc)
    if return_type is None:
        return_type = wtypes.never
    func_type = wtypes.function(param_types.to_list(), return_type)
    return Dynamic.new(func_type, (closure, []))


def dereference(x):
    '''
    ..., ref -> ..., value
    '''
    assert x.is_const()
    t0, payload = wtypes.switch(x.type_)
    if t0 != 'REF':
        raise TypedCodeError('Value is not a reference')
    constness, type_ = payload
    (ref,) = x.read()
    return Dynamic(type_, ref, constness=constness)


def dynamic_constness(x):
    '''
    ..., dynamic -> ..., constness
    '''
    assert x.is_const()
    t0, _ = wtypes.switch(x.type_)
    if t0 != 'DYNAMIC':
        raise TypedCodeError('Value is not a dynamic')
    d = Dynamic.from_welly(x.ref)
    return Dynamic(wtypes.constness, d.constness)


def dynamic_type(x):
    '''
    ..., dynamic -> ..., type
    '''
    assert x.is_const()
    t0, _ = wtypes.switch(x.type_)
    if t0 != 'DYNAMIC':
        raise TypedCodeError('Value is not a dynamic')
    d = Dynamic.from_welly(x.ref)
    return Dynamic(wtypes.type_, d.type_)


def get(name, module, loc):
    '''..., module -> ..., dynamic_value'''
    assert module.is_const()
    module.check_is_a(wtypes.module, 'the module used for "get"', loc)
    (m,) = module.read()
    if name not in m:
        raise NameVMError(name)
    result = m[name] # A Welly dynamic value.
    return Dynamic.new(wtypes.dynamic, result.to_welly())


def str_length(str_, loc):
    '''..., str -> ..., length'''
    assert str_.is_const()
    str_.check_is_a(wtypes.str_, None, loc)
    (s,) = str_.ref.read()
    result = len(s)
    return Dynamic.new(wtypes.int_, (result,))


def struct(items):
    '''
    ..., <n items> -> ..., struct_value

     - items - a list of Dynamic of length `n`.
    '''
    assert all(item.is_const() for item in items)
    types = [item.type_ for item in items]
    result = Dynamic.malloc(wtypes.struct(types))
    for field_, item in zip(ops.struct_iter(result), items):
        ops.assign(field_, item, None)
    return result


def subscript_array(array, index, loc):
    '''
    ..., array, index -> ..., value
    '''
    assert array.is_const()
    assert index.is_const()
    t0, payload = wtypes.switch(array.type_)
    if t0 != 'ARRAY':
        raise TypedCodeError('Value is not an array')
    constness, elem_type = payload
    a = Array.from_welly(array.read(), sizeof(elem_type))
    index.check_is_a(wtypes.int_, "the index", loc)
    (i,) = index.read()
    if i < 0 or i >= a.length:
        raise ArrayIndexVMError(i, a.length)
    return Dynamic(elem_type, a[i], constness=constness)


def subscript_struct(index, struct_, loc):
    '''..., struct -> ..., value'''
    assert type(index) in (int, long)
    if wtypes.switch(struct_.type_)[0] != 'STRUCT':
        raise TypedCodeError('Value is not a struct')
    for i, field_ in enumerate(ops.struct_iter(struct_)):
        if i == index:
            return field_
    else:
        raise StructIndexVMError(struct_.type_, index, loc)


def unary(op, x, loc):
    '''
    ..., value -> ..., value

     - op - a welly.compiler.operators.Operator.
    '''
    assert x.is_const()
    (x_t,) = op.in_types
    x.check_is_a(x_t, None, loc)
    (x_value,) = x.read()
    result = op.implementation(x_value)
    return Dynamic.new(op.result_type, (result,))


def union(tag, items):
    '''
    ..., <n items> -> ..., union_value

     - tag - a str.
     - items - a list of Dynamic of length `n`.
    '''
    assert type(tag) is str
    assert all(x.is_const() for x in items)
    types = [item.type_ for item in items]
    # Allocate and populate the payload struct.
    payload = Dynamic.malloc(wtypes.struct(types))
    for ref, item in zip(ops.struct_iter(payload), items):
        ops.assign(ref, item, None)
    return Dynamic.new(
        wtypes.union([(tag, types)]),
        (tag, payload.ref.block)
    )


def unpack_struct(struct, n, loc):
    '''..., struct_value -> ..., field_1, ..., field_n'''
    if wtypes.switch(struct.type_)[0] != 'STRUCT':
        raise TypedCodeError('Value is not a struct')
    if ops.struct_length(struct.type_) != n:
        raise NumberOfFieldsVMError(
            n, ops.struct_length(struct.type_), loc=loc)
    return list(ops.struct_iter(struct))
