import errors

from memory import Ref


__all__ = ['defer', 'fulfil', 'read']


'''
Utility functions for deferring and fulfilling values.

Deferred values are used to construct immutable cyclic values incrementally,
as follows:

 - Use `defer()` to construct one or more deferred values.

 - You can then make copies of them, pass them to and from functions, include
   them in structs and arrays, and so on. Any copy of a deferred value is as
   good as the original.

 - Incorporate these copies into other values.

 - Use `fulfil()` to reveal that these other values are actually the same
   values as the deferred values.

If a value is deferrable, you should always read it using the `read()`
function defined here. The run-time overhead is very small (and will be zero
in the real Welly).

Not all values are deferrable. There is a requirement on the size of the
value, the value must be const, and the value must be distinguishable from a
forwarding pointer. In practice, the deferrable values include Welly
functions, macros, unions and modules.

This module is defined in terms of raw Refs, i.e. below the type-checked
layer.

See "doc/cycles.txt" for motivation and design.
See "doc/python_model.txt" for implementation notes.
'''


def _defer_iter(ref):
    '''
    Yields Refs to all of the values in the forwarding chain starting at
    `ref`. The first reference will be `ref` itself. The last reference will
    be to the fulfilment value if there is one, otherwise to the final defer
    cell, which will be `(None, ref)`, where `ref` is a Ref to the pair
    itself.

    Note: the forwarding links are read from a Ref before yielding it.
    Therefore, it is safe to write to the Refs; it will not interrupt
    the iteration.
    '''
    assert type(ref) is Ref, ref
    while True:
        tag, ptr = ref.read()
        yield ref
        if tag is not None:
            # Fulfilment value.
            break
        if ref.block is ptr:
            # Final defer cell.
            break
        ref = Ref(ptr, 0, 2)


def _canonicalise(ref):
    '''
    Returns a pair of words that ultimately represents the deferrable value
    referred to by `ref`. Also, short-circuits the forwarding pointers so that
    it can be read faster next time.
    '''
    for r in _defer_iter(ref):
        pass
    pair = r.read()
    for r in _defer_iter(ref):
        r.write(pair)
    return pair


def defer():
    '''Returns a Ref to a fresh defer cell.'''
    result = Ref.malloc(2)
    result.write((None, result.block))
    return result


def fulfil(deferred, fulfilment):
    '''
    Unify a deferred value with another value, hereafter called the
    "fulfilment". The fulfilment may in turn be deferred.

    Raises FulfilledValueVMError if `deferred` already has a value.

     - deferred - a Ref to the deferred value.
     - fulfilment - a Ref to the fulfilment.
    '''
    # Check that `deferred` actually is still deferred.
    for r in _defer_iter(deferred):
        pass
    tag, ref = r.read()
    if tag is not None:
        raise errors.FulfilledValueVMError()
    # Make `deferred` be/point to `fulfilment`.
    pair = _canonicalise(fulfilment)
    for r in _defer_iter(deferred):
        r.write(pair)


def read(ref, noun='the value'):
    '''
    Checks that a deferrable value is not deferred, and returns its value.
    The value is a pair of words.

    Raises UnfulfilledValueVMError if `ref` is deferred.

     - noun - a noun phrase describing the value being read, e.g.
       "the function value". This is used in error messages.
    '''
    assert type(ref) is Ref
    result = _canonicalise(ref)
    if result[0] is None:
        raise errors.UnfulfilledValueVMError(noun)
    return result
