'''
Interfaces the pretty printer to the Welly type system.
'''


import defer, const, wtypes, subtypes

from .errors import UnfulfilledValueVMError
from .memory import Ref, Array


__all__ = ['wrap']


def _constant(value):
    assert type(value) is unicode
    return lambda: (value, u'', u'', [])


_primitive_types = {
    'BOOL': u'bool',
    'FLOAT': u'float',
    'INT': u'int',
    'STR': u'str',
    'ARRAY0': u'array0',
    'DYNAMIC': u'dynamic',
    'MODULE': u'module'
}


def _constness_syntax(constness):
    '''
    Returns '', 'const ' or 'var '. This is used to construct strings such as
    "ref const int" or "array var str".
    '''
    c0 = const.switch(constness)
    if c0 == 'BASE':
        return ''
    elif c0 == 'CONST':
        return 'const '
    elif c0 == 'VAR':
        return 'var '
    assert False


def _bracketed(contents, tag=u''):
    '''
    Returns a printable of the form "tag(contents[0], contents[1], ...)"

     - contents - an iterable returning Refs to Welly types.
     - tag - (optional) a word to put before the open bracket.
    '''
    return lambda: (u'%s(' % tag, u', ', u')', [wrap(x) for x in contents])


def wrap(type_):
    '''
    Given a Welly type, returns a pretty-print-aware wrapper around it.

     - type_ - a Ref to a Welly type.
    '''
    try:
        tag, payload = wtypes.switch(type_)
    except UnfulfilledValueVMError:
        return _constant(u'unfulfilled')
    if tag in _primitive_types:
        return _constant(_primitive_types[tag])
    elif tag == 'REF':
        constness, referee_type = payload
        return lambda: (
            u'ref ' + _constness_syntax(constness),
            u'',
            u'',
            [wrap(referee_type)]
        )
    elif tag == 'ARRAY':
        constness, elem_type = payload
        return lambda: (
            u'array ' + _constness_syntax(constness),
            u'',
            u'',
            [wrap(elem_type)]
        )
    elif tag == 'STRUCT':
        (field_type_array,) = payload
        if subtypes.eq(type_, wtypes.void):
            return _constant(u'void')
        else:
            return _bracketed(field_type_array, tag=u'struct')
    elif tag == 'UNION':
        (case_array,) = payload
        if subtypes.eq(type_, wtypes.type_):
            return _constant(u'type')
        return lambda: (
            u'union {',
            u'; ',
            u'}',
            [
                _bracketed(types, tag=tag)
                for (tag, types) in wtypes.union_index(case_array).items()
            ]
        )
    elif tag == 'FUNCTION':
        param_type_array, return_type = payload
        return lambda: (
            u'fn', u': ', u'',
            [_bracketed(param_type_array), _bracketed([return_type])]
        )
    assert False, t0
