'''
Implements the subtype relation described in "doc/subtypes.txt", modulo the
differences in "doc/python_model.txt".
'''


import collections

import const
import wtypes

from errors import TypeVMError, TypeInferenceVMError
from memory import Ref, Array
from const import CONST, VAR, BASE
from wtypes import sizeof


__all__ = ['is_subtype', 'eq', 'check', 'join', 'meet', 'join_many']


# Tags of all types that are only subtypes of themselves.
_discrete_types = [
    'BOOL', 'FLOAT', 'INT', 'STR', 'DYNAMIC', 'MODULE']


_known_subtypes = set() # Set of pair of Ref to payload.


def is_subtype(t, u):
    '''
    Tests whether `t` is a subtype of `u`.
    Handles recursive types. Caches results.

     - t - a Ref to a type.
     - u - a Ref to a type.
    '''
#    print "is_subtype(%r, %r)" % (t, u)
    done = set() # Set of pair of Ref to type.
    todo = [] # List of (type, type).

    def enqueue(t, u):
        # print "enqueue(%r, %r)" % (t, u)
        if t == u:
            return
        key = (t, u)
        if key in _known_subtypes or key in done:
            return
        else:
            done.add(key)
            todo.append((t, u))

    enqueue(t, u)
    while todo:
        t, u = todo.pop()
#        print "comparing %r with %r" % (t, u)
        t0, t_payload = wtypes.switch(t)
        u0, u_payload = wtypes.switch(u)
        if t0 in _discrete_types:
            if u0 == t0:
                pass
            else:
                return False
        elif t0 == 'REF':
            t_constness, t_type = t_payload
            t_constness = const.switch(t_constness)
            if u0 == 'REF':
                u_constness, u_type = u_payload
                u_constness = const.switch(u_constness)
                if t_constness != u_constness and u_constness != 'BASE':
                    return False
                enqueue(t_type, u_type)
                if u_constness == 'VAR':
                    # We can assign via `u` and read it back via `t`, so
                    # `u_type` had better be a subtype of `t_type`.
                    enqueue(u_type, t_type)
            else:
                return False
        elif t0 == 'ARRAY0':
            if u0 == 'ARRAY0':
                pass
            elif u0 == 'ARRAY':
                pass
            else:
                return False
        elif t0 == 'ARRAY':
            t_constness, t_elem_type = t_payload
            t_constness = const.switch(t_constness)
            if u0 == 'ARRAY':
                u_constness, u_elem_type = u_payload
                u_constness = const.switch(u_constness)
                if t_constness != u_constness and u_constness != 'BASE':
                    return False
                enqueue(t_elem_type, u_elem_type)
                if u_constness == 'VAR':
                    enqueue(u_elem_type, t_elem_type)
            else:
                return False
        elif t0 == 'STRUCT':
            (t_fields,) = t_payload
            if u0 == 'STRUCT':
                (u_fields,) = u_payload
                if len(t_fields) != len(u_fields):
                    return False
                for t_field_type, u_field_type in zip(t_fields, u_fields):
                    enqueue(t_field_type, u_field_type)
            else:
                return False
        elif t0 == 'UNION':
            (t_cases,) = t_payload
            if u0 == 'UNION':
                (u_cases,) = u_payload
                t_index = wtypes.union_index(t_cases)
                u_index = wtypes.union_index(u_cases)
                for name, t_types in t_index.items():
                    if name not in u_index:
                        return False
                    u_types = u_index[name]
                    if len(t_types) != len(u_types):
                        return False
                    for t_field_type, u_field_type in zip(t_types, u_types):
                        enqueue(t_field_type, u_field_type)
            else:
                return False
        elif t0 == 'FUNCTION':
            t_param_array, t_return_type = t_payload
            if u0 == 'FUNCTION':
                u_param_array, u_return_type = u_payload
                if len(t_param_array) != len(u_param_array):
                    return False
                for t_param_type, u_param_type in zip(
                    t_param_array, u_param_array
                ):
                    enqueue(u_param_type, t_param_type)
                enqueue(t_return_type, u_return_type)
            else:
                return False
        else:
            return False
    _known_subtypes.update(done)
    return True


def eq(t, u):
    '''
    Tests whether `t` and `u` are structurally equivalent.
    '''
    return is_subtype(t, u) and is_subtype(u, t)


def check(t, u, noun, loc):
    '''
    Checks that `t` is a subtype of `u`, and raises a TypeVMError if not.

     - noun - a noun phrase that describes the value of type `t`, e.g.
       "return type of function", or `None` to use a default noun phrase.
     - loc - an opaque source code location for use in error messages.
    '''
    if not is_subtype(t, u):
        raise TypeVMError(t, u, noun, loc)


def join(t, u, noun, loc):
    '''
    Returns a Ref to the least supertype of `t` and `u`, if possible,
    otherwise raises TypeInferenceVMError.

     - noun - a noun phrase describing the expression whose type is being
       inferred, for use in error messages.
     - loc - an opaque source code location for use in error messages.
    '''
    def error():
        raise TypeInferenceVMError(t, u, 'supertype', noun, loc)
    # Return `t` or `u` if possible.
    # This is not just an optimization. For example,
    #   join(ref var x, ref var x) = ref var x
    # but if x != y then
    #   join(ref var x, ref var y) = ref join(x, y)
    # N.B. `ref var ...` in the former case but `ref ...` in the latter.
    if is_subtype(t, u):
        return u
    if is_subtype(u, t):
        return t
    # TODO: Cope with recursive types.
    t0, t_payload = wtypes.switch(t)
    u0, u_payload = wtypes.switch(u)
    if t0 == 'REF':
        t_constness, t_type = t_payload
        t_constness = const.switch(t_constness)
        if u0 == 'REF':
            u_constness, u_type = u_payload
            u_constness = const.switch(u_constness)
            # Compute the join of the referee types.
            new_t = join(t_type, u_type, noun, loc)
            # Return `ref const ...` if possible.
            if t_constness == 'CONST' and u_constness == 'CONST':
                return wtypes.ref(CONST, new_t)
            else:
                # Cannot return `ref var ...` because `t != u`.
                return wtypes.ref(BASE, new_t)
        else:
            raise error()
    elif t0 == 'ARRAY0':
        assert u0 != 'ARRAY0' and u0 != 'ARRAY'
        raise error()
    elif t0 == 'ARRAY':
        t_constness, t_elem_type = t_payload
        t_constness = const.switch(t_constness)
        if u0 == 'ARRAY':
            u_constness, u_elem_type = u_payload
            u_constness = const.switch(u_constness)
            # Compute the meet of the element types.
            new_t = join(t_elem_type, u_elem_type, noun, loc)
            # Return `array const ...` if possible.
            if t_constness == 'CONST' and u_constness == 'CONST':
                return wtypes.array(CONST, new_t)
            else:
                # Cannot return `array var ...` because `t != u`.
                return wtypes.array(BASE, new_t)
        else:
            assert u0 != 'ARRAY0'
            raise error()
    elif t0 == 'STRUCT':
        (t_fields,) = t_payload
        if u0 == 'STRUCT':
            (u_fields,) = u_payload
            if len(t_fields) != len(u_fields):
                raise error()
            return wtypes.struct([
                join(t_field_type, u_field_type, noun, loc)
                for t_field_type, u_field_type in zip(t_fields, u_fields)
            ])
        else:
            raise error()
    elif t0 == 'UNION':
        (t_cases,) = t_payload
        if u0 == 'UNION':
            (u_cases,) = u_payload
            t_index = wtypes.union_index(t_cases)
            u_index = wtypes.union_index(u_cases)
            new_index = collections.OrderedDict()
            for name in t_index:
                if name not in u_index:
                    new_index[name] = t_index[name]
                elif len(t_index[name]) == len(u_index[name]):
                    new_index[name] = [
                        join(x, y, noun, loc)
                        for x, y in zip(t_index[name], u_index[name])
                    ]
                else:
                    raise error()
            for name in u_index:
                if name not in t_index:
                    new_index[name] = u_index[name]
            return wtypes.union(new_index.items())
        else:
            raise error()
    elif t0 == 'FUNCTION':
        t_param_array, t_return_type = t_payload
        if u0 == 'FUNCTION':
            u_param_array, u_return_type = u_payload
            if len(t_param_array) != len(u_param_array):
                raise error()
            new_param_types = []
            for t_param_type, u_param_type in zip(
                t_param_array, u_param_array
            ):
                new_t = meet(t_param_type, u_param_type, noun, loc)
                new_param_types.append(new_t)
            new_return_type = join(t_return_type, u_return_type, noun, loc)
            return wtypes.function(new_param_types, new_return_type)
        else:
            raise error()
    else:
        assert t0 in _discrete_types
        assert t0 != u0
        raise error()


def meet(t, u, noun, loc):
    '''
    Returns a Ref to the greatest subtype of `t` and `u`, if possible,
    otherwise raises TypeInferenceVMError.

     - noun - a noun phrase describing the expression whose type is being
       inferred, for use in error messages.
     - loc - an opaque source code location for use in error messages.
    '''
    def error():
        raise TypeInferenceVMError(t, u, 'subtype', noun, loc)
    # Return `t` or `u` if possible.
    # This is not just an optimization. For example,
    #   meet(ref var x, ref var x) = ref var x
    # but if x != y then
    #   meet(ref var x, ref var y) = None
    if is_subtype(t, u):
        return t
    if is_subtype(u, t):
        return u
    # TODO: Cope with recursive types.
    t0, t_payload = wtypes.switch(t)
    u0, u_payload = wtypes.switch(u)
    if t0 == 'REF':
        t_constness, t_type = t_payload
        t_constness = const.switch(t_constness)
        if u0 == 'REF':
            u_constness, u_type = u_payload
            u_constness = const.switch(u_constness)
            if t_constness == 'VAR' or u_constness == 'VAR':
                # Cannot return `ref var ...` because `t != u`.
                raise error()
            # Compute the meet of the referee types.
            new_t = meet(t_type, u_type, noun, loc)
            # Return `ref const ...` if necessary.
            if t_constness == 'CONST' or u_constness == 'CONST':
                return wtypes.ref(CONST, new_t)
            else:
                return wtypes.ref(BASE, new_t)
        else:
            raise error()
    elif t0 == 'ARRAY0':
        assert u0 != 'ARRAY0' and u0 != 'ARRAY'
        raise error()
    elif t0 == 'ARRAY':
        t_constness, t_elem_type = t_payload
        t_constness = const.switch(t_constness)
        if u0 == 'ARRAY':
            u_constness, u_elem_type = u_payload
            u_constness = const.switch(u_constness)
            if t_constness == 'VAR' or u_constness == 'VAR':
                # Cannot return `array var ...` because `t != u`.
                return wtypes.array0
            # Compute the meet of the element types.
            try:
                new_t = meet(t_elem_type, u_elem_type, noun, loc)
            except TypeInferenceVMError:
                return wtypes.array0
            # Return `array const ...` if necessary.
            if t_constness == 'CONST' or u_constness == 'CONST':
                return wtypes.array(CONST, new_t)
            else:
                return wtypes.array(BASE, new_t)
        else:
            assert u0 != 'ARRAY0'
            raise error()
    elif t0 == 'STRUCT':
        (t_fields,) = t_payload
        if u0 == 'STRUCT':
            (u_fields,) = u_payload
            if len(t_fields) != len(u_fields):
                raise error()
            return wtypes.struct([
                meet(t_field_type, u_field_type, noun, loc)
                for t_field_type, u_field_type in zip(t_fields, u_fields)
            ])
        else:
            raise error()
    elif t0 == 'UNION':
        (t_cases,) = t_payload
        if u0 == 'UNION':
            (u_cases,) = u_payload
            t_index = wtypes.union_index(t_cases)
            u_index = wtypes.union_index(u_cases)
            new_index = collections.OrderedDict()
            for name in t_index:
                if name in u_index:
                    if len(t_index[name]) == len(u_index[name]):
                        try:
                            new_index[name] = [
                                meet(x, y, noun, loc)
                                for x, y in zip(t_index[name], u_index[name])
                            ]
                        except TypeInferenceVMError:
                            pass
            return wtypes.union(new_index.items())
        else:
            raise error()
    elif t0 == 'FUNCTION':
        t_param_array, t_return_type = t_payload
        if u0 == 'FUNCTION':
            u_param_array, u_return_type = u_payload
            if len(t_param_array) != len(u_param_array):
                raise error()
            new_param_types = []
            for t_param_type, u_param_type in zip(
                t_param_array, u_param_array
            ):
                new_t = join(t_param_type, u_param_type, noun, loc)
                new_param_types.append(new_t)
            new_return_type = meet(t_return_type, u_return_type, noun, loc)
            return wtypes.function(new_param_types, new_return_type)
        else:
            raise error()
    else:
        assert t0 in _discrete_types
        assert t0 != u0
        raise error()


def join_many(types, noun, loc):
    '''
    Returns the least supertype of all of `types`. If `types` is empty,
    returns `wtypes.never`. If there is no common supertype, raises
    TypeInferenceError.

     - noun - a noun phrase describing the expression whose type is being
       inferred, for use in error messages.
     - loc - an opaque source code location for use in error messages.
    '''
    if len(types) == 0:
        return wtypes.never
    ans = types[0]
    for u in types[1:]:
        try:
            ans = join(ans, u, noun, loc)
        except TypeInferenceVMError as e:
            # Try to find a pair of types in `types` that have no common
            # supertype, to make the error message easier to understand.
            for t in types:
                _ = join(t, u, noun, loc)
            # We failed. We don't think this is possible, but...
            raise e
    return ans


def join_many_with_constnesses(constnesses, types, noun, loc):
    '''
    Given the constnesses and types of a list of `n` values, returns the
    constness and type of their join.

     - constnesses - a list of `n` constnesses.
     - types - a list of `n` types.
     - noun - a noun phrase describing the expression whose type and constness
       is being inferred, for use in error messages.
     - loc - an opaque source code location for use in error messages.
    '''
    assert len(constnesses) == len(types)
    type_ = join_many([t for t in types], noun, loc)
    if all(c == CONST for c in constnesses):
        constness = CONST
    elif (
        all(c == VAR for c in constnesses) and
        all(is_subtype(type_, t) for t in types)
    ):
        constness = VAR
    else:
        constness = BASE
    return (constness, type_)
