import pytest

from errors import TypeInferenceVMError
from const import BASE, CONST, VAR
from wtypes import *
from subtypes import *


slow = pytest.mark.skipif(
    not (hasattr(pytest, 'config') and pytest.config.getoption("--run-slow")),
    reason="need --run-slow option to run"
)


def test_subtype_cache():
    assert not is_subtype(int_, dynamic) # Should not cache `(int_, dynamic)`.
    assert not is_subtype(int_, dynamic) # Should return `False` again.


def test_distinct():
    # Check that different types are not subtypes of each other.
    ts = [
        bool_, float_, int_, str_,
        ref(BASE, int_),
        array(BASE, int_),
        struct([int_]),
        union([('None', []), ('Some', [int_])]),
        function([], bool_),
        dynamic, module,
    ]
    for t in ts:
        for u in ts:
            assert is_subtype(t, u) == (t == u), (t, u)
    # 'ARRAY0' is special; it is a supertype of all array types.
    for t in ts:
        is_t_array = switch(t)[0] == 'ARRAY'
        assert is_subtype(array0, t) == is_t_array, t
        assert not is_subtype(t, array0), t


# Test arrays first, because they are useful for testing other things.
def test_array_subtypes():
    def check_array_subtypes(t, u):
        assert is_subtype(t, u), 'check_array_subtypes precondition'
        assert is_subtype(array(BASE, t), array(BASE, u))
        assert not is_subtype(array(BASE, t), array(CONST, u))
        assert not is_subtype(array(BASE, t), array(VAR, u))
        assert is_subtype(array(CONST, t), array(BASE, u))
        assert is_subtype(array(CONST, t), array(CONST, u))
        assert not is_subtype(array(CONST, t), array(VAR, u))
        assert is_subtype(array(VAR, t), array(BASE, u))
        assert not is_subtype(array(VAR, t), array(CONST, u))
        assert is_subtype(array(VAR, t), array(VAR, u)) == is_subtype(u, t)

    check_array_subtypes(int_, int_)
    check_array_subtypes(array0, array0)
    check_array_subtypes(array0, array(CONST, int_))
    check_array_subtypes(array(VAR, int_), array(BASE, int_))
    check_array_subtypes(array(CONST, int_), array(BASE, int_))


def test_ref_subtypes():
    def check_ref_subtypes(t, u):
        assert is_subtype(t, u), 'check_ref_subtypes precondition'
        assert is_subtype(ref(BASE, t), ref(BASE, u))
        assert not is_subtype(array(BASE, t), ref(CONST, u))
        assert not is_subtype(array(BASE, t), ref(VAR, u))
        assert is_subtype(ref(CONST, t), ref(BASE, u))
        assert is_subtype(ref(CONST, t), ref(CONST, u))
        assert not is_subtype(ref(CONST, t), ref(VAR, u))
        assert is_subtype(ref(VAR, t), ref(BASE, u))
        assert not is_subtype(array(VAR, t), ref(CONST, u))
        assert is_subtype(ref(VAR, t), ref(VAR, u)) == is_subtype(u, t)

    check_ref_subtypes(int_, int_)
    check_ref_subtypes(array0, array0)
    check_ref_subtypes(array0, array(CONST, int_))
    check_ref_subtypes(array(VAR, int_), array(BASE, int_))
    check_ref_subtypes(array(CONST, int_), array(BASE, int_))


def test_struct_subtypes():
    def check_struct_subtype(t_member_types, u_member_types):
        t = struct(t_member_types)
        u = struct(u_member_types)
        i = struct([int_])
        assert is_subtype(t, u)
        assert is_subtype(u, u)
        assert not is_subtype(u, t)
        assert is_subtype(i, i)
        assert not is_subtype(i, t)
        assert not is_subtype(i, u)
        assert not is_subtype(t, i)
        assert not is_subtype(u, i)
        assert not is_subtype(int_, t)
        assert not is_subtype(int_, u)
        assert not is_subtype(t, int_)
        assert not is_subtype(u, int_)

    check_struct_subtype([array0], [array(VAR, int_)])
    check_struct_subtype([array0, int_], [array(VAR, int_), int_])
    check_struct_subtype([int_, array0], [int_, array(VAR, int_)])


def test_union_subtypes():
    def check_union_subtype(t, u):
        assert is_subtype(never, t)
        assert not is_subtype(t, never)
        assert is_subtype(never, u)
        assert not is_subtype(u, never)
        assert is_subtype(t, t)
        assert is_subtype(t, u)
        assert not is_subtype(u, t)
        assert is_subtype(u, u)

    option0 = union([('None', [])])
    option_int_1 = union([('None', []), ('Some', [int_])])
    option_int_2 = union([('Some', [int_]), ('None', [])])

    assert is_subtype(never, never)
    check_union_subtype(option0, option_int_1)
    check_union_subtype(option0, option_int_2)
    assert is_subtype(option_int_1, option_int_2)
    assert is_subtype(option_int_2, option_int_1)


def test_function_subtypes():
    t = array0
    u = array(VAR, int_)
    assert is_subtype(t, u), 'check_function_subtypes precondition'
    assert is_subtype(function([], t), function([], t))
    assert is_subtype(function([], t), function([], u))
    assert not is_subtype(function([], u), function([], t))
    assert is_subtype(function([], u), function([], u))
    assert is_subtype(function([t], bool_), function([t], bool_))
    assert not is_subtype(function([t], bool_), function([u], bool_))
    assert is_subtype(function([u], bool_), function([t], bool_))
    assert is_subtype(function([u], bool_), function([u], bool_))


@slow
def test_meet_and_join():
    subtype = array0
    supertype = array(CONST, int_)
    # A set that is closed under meet and join (when they are defined)
    # and that covers many code paths.
    examples = [
        array0,
        array(BASE, subtype),
        array(CONST, subtype),
        array(VAR, subtype),
        array(BASE, supertype),
        array(CONST, supertype),
        array(VAR, supertype),
        ref(BASE, subtype),
        ref(CONST, subtype),
        ref(VAR, subtype),
        ref(BASE, supertype),
        ref(CONST, supertype),
        ref(VAR, supertype),
        struct([subtype, subtype]),
        struct([subtype, supertype]),
        struct([supertype, subtype]),
        struct([supertype, supertype]),
        struct([subtype]),
        struct([supertype]),
        struct([bool_]),
        struct([]),
        union([]),
        union([('FOO', [subtype, subtype])]),
        union([('FOO', [subtype, supertype])]),
        union([('FOO', [supertype, subtype])]),
        union([('FOO', [supertype, supertype])]),
        union([('BAR', [int_])]),
        union([('FOO', [subtype, subtype]), ('BAR', [int_])]),
        union([('FOO', [subtype, supertype]), ('BAR', [int_])]),
        union([('FOO', [supertype, subtype]), ('BAR', [int_])]),
        union([('FOO', [supertype, supertype]), ('BAR', [int_])]),
        union([('BAR', [str_])]),
        union([('FOO', [subtype, subtype]), ('BAR', [str_])]),
        union([('FOO', [subtype, supertype]), ('BAR', [str_])]),
        union([('FOO', [supertype, subtype]), ('BAR', [str_])]),
        union([('FOO', [supertype, supertype]), ('BAR', [str_])]),
        union([('FOO', [])]),
        union([('FOO', []), ('BAR', [int_])]),
        union([('FOO', []), ('BAR', [str_])]),
        union([('FOO', [bool_])]),
        union([('FOO', [bool_]), ('BAR', [int_])]),
        union([('FOO', [bool_]), ('BAR', [str_])]),
        function([], int_),
        function([], subtype),
        function([], supertype),
        function([int_], void),
        function([subtype], void),
        function([supertype], void),
        function([subtype, subtype], void),
        function([subtype, supertype], void),
        function([supertype, subtype], void),
        function([supertype, supertype], void),
    ]
    for t in examples:
        for u in examples:
            # Test meet.
            try:
                new_t = meet(t, u, 'the example type', None)
                assert new_t is not None, (t, u)
                is_in_examples = False
                for v in examples:
                    if is_subtype(v, t) and is_subtype(v, u):
                        assert is_subtype(v, new_t), (t, u, v)
                        if is_subtype(new_t, v):
                            is_in_examples = True
                assert is_in_examples, new_t
            except TypeInferenceVMError:
                assert not any(
                    is_subtype(v, t) and is_subtype(v, u)
                    for v in examples
                )
            # Test join.
            try:
                new_t = join(t, u, 'the example type', None)
                assert new_t is not None, (t, u)
                is_in_examples = False
                for v in examples:
                    if is_subtype(t, v) and is_subtype(u, v):
                        assert is_subtype(new_t, v), (t, u, v)
                        if is_subtype(v, new_t):
                            is_in_examples = True
                assert is_in_examples, new_t
            except TypeInferenceVMError:
                assert not any(
                    is_subtype(t, v) and is_subtype(u, v)
                    for v in examples
                )


def test_join_many():
    t = union([('FOO', [bool_])])
    u = union([('BAR', [])])
    v = union([('FOO', [int_])])
    _ = join_many([t, u], 'the example type', None)
    _ = join_many([u, v], 'the example type', None)
    with pytest.raises(TypeInferenceVMError):
        _ = join_many([t, v], 'the example type', None)
    with pytest.raises(TypeInferenceVMError):
        _ = join_many([t, u, v], 'the example type', None)
