'''
This implementation does computations as per the Welly-in-Python data model,
which differs in detail from the real Welly data model. See
"doc/python_model.txt" for details.

Welly types are deferrable values. To avoid complications in Python code, we
always pass types around by reference, i.e. as an instance of Ref.
'''


import collections

import defer
import pack
import const

from .errors import UnfulfilledValueVMError
from .memory import Ref, Array
from .const import BASE, CONST, VAR


# Public API.
__all__ = [
    'bool_', 'float_', 'int_', 'str_', 'ref', 'array0', 'array',
    'struct', 'union', 'function', 'dynamic', 'module',
    'type_', 'switch', 'check_valid',
    'union_case', 'union_index',
    'constness', 'void', 'never',
    'sizeof',
    'is_deferrable',
]


# Sizes of common types. We define them here to avoid circular dependencies.
_type_size = 2
_union_case_size = 3


# Type constructors. Ordered as in "doc/data_model.txt".
# These are intended as a convenient Python API for constructing types.
# They take Python data structures. They return a Welly data structure: a type.


bool_ = pack.tag('BOOL')


float_ = pack.tag('FLOAT')


int_ = pack.tag('INT')


str_ = pack.tag('STR') # Only in Welly-in-Python.


def ref(constness, type_):
    '''
    Constructs a reference type.

     - constness - one of BASE, CONST or VAR.
     - type_ - a Ref to the type of the value referred to.
    '''
    const.check_valid(constness)
    check_valid(type_)
    return pack.tag('REF', constness.read(), type_.read())


array0 = pack.tag('ARRAY0') # The type of `[]`.


def array(constness, elem_type):
    '''
    Constructs an array type.

     - constness - one of BASE, CONST or VAR.
     - elem_type - a Ref to the type of the array elements.
    '''
    const.check_valid(constness)
    check_valid(elem_type)
    return pack.tag('ARRAY', constness.read(), elem_type.read())


def struct(field_types):
    '''
    Constructs a struct type.

     - field_types - list of Ref to a Welly type.
    '''
    assert type(field_types) is list
    field_array = Array.from_list([t.read() for t in field_types], _type_size)
    # STRUCT(array const type)
    return pack.tag('STRUCT', field_array)


def union(cases):
    '''
    Constructs a union type.

     - cases - list of `(name, fields)` where `name` is a str and `fields` is
       a list of Ref to Welly type.

    The case names may include duplicates. We do not prevent this. Instead,
    we define the interpretation: only the first occurrence is used.
    '''
    assert type(cases) is list
    case_array = Array.malloc(len(cases), _union_case_size)
    for c, c_ref in zip(cases, case_array):
        assert type(c) is tuple
        name, field_types = c
        assert type(name) is str
        assert type(field_types) is list
        type_array = Array.malloc(len(field_types), _type_size)
        for t, t_ref in zip(field_types, type_array):
            check_valid(t)
            t_ref.write(t.read())
        c_ref.write(pack.pack(name, type_array))
    # UNION(array const union_case)
    return pack.tag('UNION', case_array)


def function(param_types, return_type):
    '''
    Constructs a function type.

     - param_types - list of Ref to Welly type.
     - return_type - Ref to Welly type.
    '''
    assert type(param_types) is list
    param_array = Array.malloc(len(param_types), _type_size)
    for t, t_ref in zip(param_types, param_array):
        check_valid(t)
        t_ref.write(t.read())
    check_valid(return_type)
    # FUNCTION(array const type, type)
    return pack.tag('FUNCTION', param_array, return_type.read())


dynamic = pack.tag('DYNAMIC')


module = pack.tag('MODULE')


def switch(t):
    '''
    Deconstructs type `t` to depth 1, returning `(tag, payload)` where `tag`
    is a str and `payload` is a tuple of Welly values.

    This is the recommended way to deconstruct a type.

     - t - a Ref to the Welly type.
    '''
    return pack.switch(t, {
        'BOOL': '',
        'FLOAT': '',
        'INT': '',
        'STR': '',
        'REF': 'r2r2', # Constness, type.
        'ARRAY0': '',
        'ARRAY': 'r2r2', # Constness, type.
        'STRUCT': 'a2', # Array of `type`.
        'UNION': 'a3', # Array of `(name, array of type)`.
        'FUNCTION': 'a2r2', # Array of type, type.
        'DYNAMIC': '',
        'MODULE': '',
    })


def union_index(union_cases):
    '''
    Iterates over an Array of union_cases and returns an OrderedDict
    mapping case name (str) to Python list of Ref to Welly type.

    This is the recommended way to examine union types, as it handles
    duplicated case names correctly. Specifically, if the same case name
    occurs more than once, the first occurrence is used.

    Params:
     - union_cases - an Array of union_cases.
    '''
    result = collections.OrderedDict()
    for ref in union_cases:
        tag, payload = pack.unpack(ref, 'w1a2')
        if tag not in result:
            result[tag] = list(payload)
    return result


def check_valid(t):
    '''
    Heuristically check that `t` (a Ref) could be a type, for debugging
    purposes.
    '''
    try:
        _ = switch(t)
    except UnfulfilledValueVMError:
        pass


# Common types.

# The type of constnesses.
constness = union([('BASE', []), ('CONST', []), ('VAR', [])])


# The type of types.
type_ = defer.defer()


# "const array_of_type = array const type"
array_of_type = array(CONST, type_)


# Used in the representation of union types.
# At this level, any string is a valid case name. The source language adds
# further restrictions.
# "struct union_case {name: str, field_types: array const type}"
union_case = struct([str_, array_of_type])


defer.fulfil(type_, union([
    ('BOOL', []),
    ('FLOAT', []),
    ('INT', []),
    ('STR', []),
    ('REF', [constness, type_]),
    ('ARRAY0', []),
    ('ARRAY', [constness, type_]),
    ('STRUCT', [array_of_type]),
    ('UNION', [array(CONST, union_case)]),
    ('FUNCTION', [array_of_type, type_]),
    ('DYNAMIC', []),
    ('MODULE', []),
]))


# "struct void {}", i.e. a type with one value.
# "union never {}", i.e. a type with no values. TODO: Choose a better name?
void = struct([])
never = union([])


# `sizeof()` and related things.

def sizeof(t):
    '''
    Returns number of words occupied by a location of type `t`.

     - t - a Ref to a type.
    '''
    t0, payload = switch(t)
    if t0 == 'BOOL':
        return 1 # Python bool.
    elif t0 == 'FLOAT':
        return 1 # Python float.
    elif t0 == 'INT':
        return 1 # Python int.
    elif t0 == 'STR':
        return 1 # Python string.
    elif t0 == 'REF':
        return 1 # Ref.
    elif t0 == 'ARRAY0':
        return 2 # (Empty) Ref, length.
    elif t0 == 'ARRAY':
        return 2 # Ref, length.
    elif t0 == 'STRUCT':
        (field_types,) = payload
        return sum(sizeof(u) for u in field_types)
    elif t0 == 'UNION':
        return 2 # Python str, Python list.
    elif t0 == 'FUNCTION':
        return 2 # Python function, Python list.
    elif t0 == 'DYNAMIC':
        return 5 # constness tag, payload, type tag, payload, Ref
    elif t0 == 'MODULE':
        return 1 # Python dict.
    assert False, t0


assert _type_size == sizeof(type_)
assert _union_case_size == sizeof(union_case)


# `is_deferrable()` and related things.

_deferrable_types = [
    'FUNCTION',
    'UNION',
]


def is_deferrable(t):
    return switch(t)[0] in _deferrable_types
