import re

from welly.parser import (
    grammar, lexer, brackets, atom, unary_expr, expr, expr_list, stmt,
    block, astbuilder
)

from welly.parser.token import ASCII, NonASCII
from welly.parser.astbuilder import ASTError


__all__ = ['WellyParser', 'ASTError']


# Matches valid and invalid UTF-8 sequences.
# An invalid sequence consumes all bytes up to the next start byte.
UTF8_REGEX = re.compile('''
    ( # Valid.
        [\x00-\x7f] |
        [\xc0-\xdf][\x80-\xbf]{1} |
        [\xe0-\xef][\x80-\xbf]{2} |
        [\xf0-\xf4][\x80-\xbf]{3}
    ) |
    ( # Invalid.
        [\x00-\xff][\x80-\xbf]*
    )
''', re.VERBOSE)


class WellyParser(object):
    '''
    Represents the state of a parser for the Welly language. Feed it bytes
    (UTF-8 encoded source code) using the `feed()` method. Call `flush()` to
    force all remaining input to be processed, e.g. at EOF. Complete
    statements and error tokens will be appended to `self.buffer`. Call
    `do_pending_stmts()` to process the contents of the buffer. Each
    statement in turn will be checked for errors, converted to an abstract
    syntax tree, and passed to `do_stmt`.

    Public fields:
     - buffer - a grammar.Buffer for collecting complete statements and
       errors.
     - parser - a grammar.Parser object.
     - do_stmt - a callback, which is called by `do_pending_stmts()`
       for every error-free statement in `self.buffer`.
     - pos - the number of bytes received so far.
    '''
    def __init__(self, do_stmt):
        '''
        Constructs a Welly parser, given its `do_stmt()` callback.
        '''

        def make_stmt_parser(p):
            p = grammar.Parser(block.block, p)
            p = grammar.Parser(stmt.stmt, p)
            p = grammar.Parser(expr.expr, p)
            p = grammar.Parser(unary_expr.unary_expr, p)
            p = grammar.Parser(atom.atom, p)
            return p

        def make_expr_parser(p):
            p = grammar.Parser(expr_list.expr_list, p)
            p = grammar.Parser(expr.expr, p)
            p = grammar.Parser(unary_expr.unary_expr, p)
            p = grammar.Parser(atom.atom, p)
            return p

        self.buffer = grammar.Buffer()
        p = make_stmt_parser(self.buffer)
        p = brackets.BracketMatcher(p, {
            '(': make_expr_parser,
            '[': make_expr_parser,
            '{': make_stmt_parser,
        })
        p = grammar.Parser(lexer.high_level_lexer, p)
        p = grammar.Parser(lexer.low_level_lexer, p)
        self.parser = p
        self.do_stmt = do_stmt
        self.pos = 0
        self.old_pos = 0 # `pos` at the moment we last started a unicode.

    def feed(self, bytes_):
        '''
        Parses `bytes_`. Complete statements and error tokens will be added to
        `self.buffer`. This method does not raise exceptions.

         - bytes_ - UTF-8 encoded source code. UTF-8 characters must not be
           split across calls to `feed()`.
        '''
        # Feed all the bytes, without raising any Exceptions.
        # We use a regex for UTF-8 decoding because it seems to be the best
        # way to keep track of the byte positions of the characters.
        for m in UTF8_REGEX.finditer(bytes_):
            c = m.group()
            old_pos = self.pos
            self.pos += len(c)
            if m.group(2):
                # Invalid sequence. Insert a unicode replacement character.
                t = NonASCII(old_pos, self.pos, u'\ufffd')
            elif len(c) == 1:
                t = ASCII(old_pos, c)
            else:
                t = NonASCII(old_pos, self.pos, c.decode('UTF-8'))
            self.parser.feed(t)

    def flush(self):
        '''
        Processes any buffered bytes and returns this WellyParser to its
        home state. Note that if the buffer contains incomplete data, it will
        be truncated, which may cause error tokens.

        This method does not raise exceptions.
        '''
        self.parser.flush()

    def is_flushed(self):
        return self.parser.is_flushed()

    def do_pending_stmts(self):
        '''
        Removes all stmts in `self.buffer`. For each one in turn, checks it
        for errors, and raises ASTError if necessary, otherwise constructs an
        abstract syntax tree and passes it to `self.do_stmt()`. Exceptions
        (including ASTError) cause the rest of the buffer to be discarded.
        '''
        to_run = self.buffer.contents[:]
        self.buffer.contents[:] = []
        for statement in to_run:
            ast = astbuilder.visit_stmt(statement)
            self.do_stmt(ast)
