"""
Defines a Grammar which recognises atoms (minimal sub-expressions and
expression operators) that occupy multiple tokens, and wraps them up as a
single token of a suitable type. Colons not consumed in this process are turned
into 'bin_op's. All other tags are passed on unchanged.
"""

from token import (
    Function, Module, TypeStruct, TypeUnion, Union, Field, Array, Ref)
from grammar import Grammar
from util import discard, pass_on, concat, Unfinished
from lexer import high_level_lexer_tags, keywords
from brackets import bracket_tags


__all__ = ['atom_tags', 'atom']

# The set of tags this parser can output, in addition to 'error' and
# those passed on from lower levels.

atom_tags = {
    'array', 'function', 'module', 'ref', 'field',
    'type_struct', 'type_union', 'union',
}

# The initial state of the atom grammar.

atom = Grammar(discard, name='atom')

# Unless overridden later, all tags are passed on unchanged.

atom.set_all(high_level_lexer_tags, Grammar(pass_on))
atom.set_all(keywords, Grammar(pass_on))
atom.set_all(bracket_tags, Grammar(pass_on))

# Errors, newlines and semicolons are passed on eagerly.

atom.set_all(['error', 'newline', ';'], Grammar(pass_on, is_eager=True))

# The keyword "fn" introduces a function definition, represented by the
# tag 'function'.

def make_function(multi_buffer):
    '''
    Returns a Function. This is used both for function definitions and
    function types.
    '''
    # keyword name? params (: type)? body?
    return Function(
        multi_buffer.get_only(stream_name='keyword'),
        multi_buffer.get_optional(stream_name='name'),
        multi_buffer.get_only(stream_name='params'),
        multi_buffer.get_optional(stream_name='colon'),
        multi_buffer.get_optional(stream_name='return_type'),
        multi_buffer.get_optional(stream_name='body'),
    )

function_end = Grammar(make_function)
function_block_or_end = Grammar(make_function)
function_block_or_end.set('{}', function_end, stream_name='body')
function_return_type = Unfinished('Expected a return type')
function_return_type.ignore('newline')
function_return_type.set_all(
    ['identifier', '()', 'module'],
    function_block_or_end,
    stream_name='return_type'
)
function_colon_or_block = Unfinished(
    'Expected a function body, or ":" and a return type')
function_colon_or_block.ignore('newline')
function_colon_or_block.set('{}', function_end, stream_name='body')
function_colon_or_block.set(':', function_return_type, stream_name='colon')
function_params = Unfinished('Expected "( ... )"')
function_params.ignore('newline')
function_params.set('()', function_colon_or_block, stream_name='params')
function_name_or_params = Unfinished('Expected an identifier or "( ... )"')
function_name_or_params.ignore('newline')
function_name_or_params.set('identifier', function_params, stream_name='name')
function_name_or_params.set('()', function_colon_or_block, stream_name='params')
atom.set('fn', function_name_or_params, stream_name='keyword')

# The keyword "module" introduces a module definition, represented by the
# tag 'module'.

def make_module(multi_buffer):
    '''
    Returns a Module. This is used both for module definitions and module
    types.
    '''
    return Module(
        multi_buffer.get_only(stream_name='keyword'),
        multi_buffer.get_optional(stream_name='name'),
        multi_buffer.get_optional(stream_name='body'),
    )

module_end = Grammar(make_module)
module_block = Unfinished('Expected "{ ... }"')
module_block.ignore('newline')
module_block.set('{}', module_end, stream_name='body')
module_name_or_block = Grammar(make_module)
module_name_or_block.set('identifier', module_block, stream_name='name')
module_name_or_block.set('{}', module_end, stream_name='body')
atom.set('module', module_name_or_block, stream_name='keyword')

# The keyword "struct" introduces a struct type, represented by the token
# type 'type_struct'.

def make_struct(multi_buffer):
    '''
    Returns a TypeStruct.
    '''
    return TypeStruct(
        multi_buffer.get_only(stream_name='keyword'),
        multi_buffer.get_optional(stream_name='name'),
        multi_buffer.get_only(stream_name='body'),
    )

struct_end = Grammar(make_struct)
struct_block = Unfinished('Expected "( ... )"')
struct_block.ignore('newline')
struct_block.set('()', struct_end, stream_name='body')
struct_name_or_block = Unfinished('Expected an identifier or "( ... )"')
struct_name_or_block.ignore('newline')
struct_name_or_block.set('identifier', struct_block, stream_name='name')
struct_name_or_block.set('()', struct_end, stream_name='body')
atom.set('struct', struct_name_or_block, stream_name='keyword')

# The keyword "union" introduces a union type, represented by the token
# type 'type_union'.

def make_type_union(multi_buffer):
    '''
    Returns a TypeUnion
    '''
    return TypeUnion(
        multi_buffer.get_only(stream_name='keyword'),
        multi_buffer.get_optional(stream_name='name'),
        multi_buffer.get_only(stream_name='body'),
    )

type_union_end = Grammar(make_type_union)
type_union_block = Unfinished('Expected "{ ... }"')
type_union_block.ignore('newline')
type_union_block.set('{}', type_union_end, stream_name='body')
type_union_name_or_block = Unfinished('Expected an identifier or "{ ... }"')
type_union_name_or_block.ignore('newline')
type_union_name_or_block.set('identifier', type_union_block, stream_name='name')
type_union_name_or_block.set('{}', type_union_end, stream_name='body')
atom.set('union', type_union_name_or_block, stream_name='keyword')

# A tag optionally followed by round brackets is a union value, represented
# by the tag 'union'.

def make_union(multi_buffer):
    '''
    Returns a Union.
    '''
    return Union(
        multi_buffer.get_only(stream_name='tag'),
        multi_buffer.get_optional(stream_name='payload'),
    )

union_end = Grammar(make_union)
union_payload_or_end = Grammar(make_union)
union_payload_or_end.set('()', union_end, stream_name='payload')
atom.set('tag', union_payload_or_end, stream_name='tag')

# "." followed by an identifier becomes a postfix operator, represented by
# the tag 'field'.

def make_field(multi_buffer):
    dot, name = multi_buffer.get()
    return Field(dot, name)

field_end = Grammar(make_field)
field_identifier = Unfinished('Expected an identifier after "."')
field_identifier.ignore('newline')
field_identifier.set('identifier', field_end)
atom.set('.', field_identifier)

# The keywords "array" optionally followed by one of "const" or "var" becomes
# a prefix operator, represented by the tag 'array'.

def make_array(multi_buffer):
    return Array(
        multi_buffer.get_only(stream_name='keyword'),
        multi_buffer.get_optional(),
    )

array_end = Grammar(make_array, name='array_end')
array_constness_or_end = Grammar(make_array, name='array_constness_or_end')
array_constness_or_end.set('const', array_end)
array_constness_or_end.set('var', array_end)
atom.set('array', array_constness_or_end, stream_name='keyword')

# The keywords "ref" optionally followed by one of "const" or "var" becomes
# a prefix operator, represented by the tag 'ref'.

def make_ref(multi_buffer):
    return Ref(
        multi_buffer.get_only(stream_name='keyword'),
        multi_buffer.get_optional(),
    )

ref_end = Grammar(make_ref, name='ref_end')
ref_constness_or_end = Grammar(make_ref, name='ref_constness_or_end')
ref_constness_or_end.set('const', ref_end)
ref_constness_or_end.set('var', ref_end)
atom.set('ref', ref_constness_or_end, stream_name='keyword')

# All other ":"s become 'bin_op' tokens.

atom.set(':', Grammar(concat('bin_op')))
