from token import Error, Concat, Bracket
from grammar import Listener, Buffer


__all__ = ['BracketMatcher', 'bracket_tags']

# Matched brackets will be passed to the higher level parser
# as one of these tags.
bracket_tags = ['()', '[]', '{}']

# The types of the tokens generated by the three kinds of bracket.
# For example, '(foo)' will generate a token `('()', ...)`.
_tags = dict((t[0], t) for t in bracket_tags)


class BracketMatcher(Listener):
    """
    A hand-written parser for matching brackets (round, square and curly).
    A hand-written parser is needed because the grammar of arbitrarily
    nested brackets is not a finite state machine.

    A BracketMatcher receives a stream of open bracket tokens, close
    bracket tokens, and other tokens. It identifies the tokens which are
    inside brackets, collects them, and collapses them down to a single
    token.

    Inside the brackets, brackets are recursively matched. The recursion is
    all handled by a single instance of BracketMatcher, using a stack.

    A separate Listener is needed for every bracket nesting level. The
    Listener for the outermost level is supplied to the constructor of the
    BracketMatcher. Listeners for the other nesting levels are constructed as
    needed, on finding an open bracket token. This is achieved using the
    `make_parser` callbacks.
    """

    def __init__(self, listener, make_parser):
        """
        Constructs a BracketMatcher.

         - listener - a Listener to which to feed the stream of tokens
            after collapsing any brackets.

         - make_parser - a dict giving, for each kind of open bracket, a
           callback that constructs a Listener to be fed the contents of the
           brackets. The callback is called passing one argument: a Listener
           to which to feed its output stream. It should return the Listener
           to which the bracket contents should be fed.

        Usage example:

            def make_statement_parser(p):
                return grammar.Parser(semicolon_separated, p)

            def make_expression_parser(p):
                return grammar.Parser(comma_separated, p)

            buf = grammar.Buffer()
            p = make_statement_parser(buf)
            p = BracketMatcher(p, {
                '{': make_statement_parser,
                '[': make_expression_parser,
                '(': lambda x: x,
            })
            p = grammar.Parser(lexer, p)

            for token in input:
                p.feed(token)
            p.flush()
            print buf.contents
        """
        self.make_parser = make_parser
        self.listener = listener
        self.stack = [] # List of (listener, buffer, bracketType)
        self.suppress_errors = False

    def feed(self, token):
        """Called for every input token."""
        if token.tag == 'open':
            self.open(token)
        elif token.tag == 'close':
            self.close(token)
        else:
            self.listener.feed(token)

    def flush(self):
        """
        Checks all brackets are closed. Recursively calls
        `listener.flush()`.
        """
        if self.stack:
            # No closing bracket.
            self.listener.flush()
            self.listener, buffer_, open_token = self.stack.pop()
            tag = _tags[open_token.s]
            error = Error(
                'Unmatched "%s"' % open_token.s, False, [open_token])
            self.listener.feed(Bracket(
                tag,
                open_token,
                [error] + buffer_.contents,
                None
            ))
            self._panic()
        self.listener.flush()

    def is_flushed(self):
        return not self.stack and self.listener.is_flushed()

    def open(self, token):
        """Called on encountering an open bracket."""
        assert type(token) is Concat
        assert token.tag == 'open'
        buffer_ = Buffer()
        self.stack.append((self.listener, buffer_, token))
        self.listener = self.make_parser[token.s](buffer_)

    def close(self, close_token):
        """Called on encountering an close bracket."""
        assert type(close_token) is Concat
        assert close_token.tag == 'close'
        if not self.stack:
            # No opening bracket.
            # We have already sent the bracket contents to the Listener.
            # Send an error token as well.
            if self.suppress_errors:
                self.listener.flush()
            else:
                self.listener.feed(Error(
                    'Unmatched "%s"' % close_token.s, False, [close_token]))
            return
        self.listener.flush()
        self.listener, buffer_, open_token = self.stack.pop()
        tag = _tags[open_token.s]
        if close_token.s == tag[1]:
            # Matched brackets.
            self.listener.feed(
                Bracket(tag, open_token, buffer_.contents, close_token))
        else:
            # Unmatched brackets.
            # We have already parsed the bracket contents, and cannot send the
            # result to the Listener without wrapping it in a bracket token.
            # Instead, we throw the contents away, and feed the Listener a
            # couple of error tokens.
            self.listener.feed(Error(
                'Unmatched "%s"' % open_token.s, False, [open_token]))
            self.listener.feed(Error(
                'Unmatched "%s"' % close_token.s, False, [close_token]))
            self._panic()

    def _panic(self):
        """
        Invents matching close brackets for all open brackets,
        and suppresses consequent errors.
        """
        while self.stack:
            self.listener.flush()
            self.listener, buffer_, open_token = self.stack.pop()
            tag = _tags[open_token.s]
            # We have already parsed the bracket contents, and cannot send the
            # result to the Listener without wrapping it in a bracket token.
            self.listener.feed(
                Bracket(tag, open_token, buffer_.contents, None))
        self.suppress_errors = True
