from token import Error, is_ascii, Escape, Concat, Keyword, String, Character
from grammar import Grammar
from util import char_range, discard, pass_on, concat, Unfinished

"""
Defines two Grammars, `low_level_lexer` and `high_level_lexer`.

The Grammar `low_level_lexer` implements a low-level lexer that accepts single
characters or the tag 'non_ascii'. Its main purpose is to collapse escape
sequences. It also groups together runs of similar characters such as digits,
letters and white-space characters. The low-level lexer does not distinguish
comments and strings. Therefore, it does not on the whole discard any
characters, and it generates few errors.

The Grammar `high_level_lexer` implements a high-level lexer that accepts the
output of `low_level_lexer`. Its main purpose is to remove comments and
white-space, and to collapse string and character literals. It also recognises
floats, brackets, identifiers, keywords and operators.
"""

__all__ = [
    'punctuation', 'low_level_lexer_tags',
    'keywords', 'high_level_lexer_tags',
    'low_level_lexer', 'high_level_lexer',
    'un_ops', 'bin_ops', 'un_bin_ops', 'assign_ops',
]

#######################################################################

# The set of all single characters passed on unchanged by `low_level_lexer`.
# They are all punctuation characters.

punctuation = set(char_range(chr(33), chr(126)))
punctuation.difference_update(char_range('0', '9'))
punctuation.difference_update(char_range('A', 'Z'))
punctuation.difference_update(char_range('a', 'z'))
punctuation.difference_update(['_', '\\'])

# The set of all Welly keywords recognised by `low_level_lexer`.

keywords = {
    'array', 'break', 'box', 'case', 'catch', 'const', 'continue', 'defer',
    'else', 'finally', 'for', 'fulfil', 'fn', 'if', 'in', 'module', 'new',
    'ref', 'return', 'struct', 'switch', 'throw', 'try', 'typeof', 'union',
    'var', 'while',
}

# The set of all `low_level_lexer` tags, apart from 'error' and
# those in `punctuation` and `keywords`.

low_level_lexer_tags = {
    'non_ascii', 'white_space', 'newline',
    'integer', 'hex_integer', 'float', 'identifier', 'tag', 'escape'
}

# The set of all `high_level_lexer` tags apart from 'error' and
# those in `keywords`.

high_level_lexer_tags = {
    'newline', '.', ',', ':', ';', 'open', 'close',
    'integer', 'hex_integer', 'float', 'string', 'character',
    'identifier', 'tag',
    'un_op', 'bin_op', 'un_bin_op', 'assign_op'
}

#######################################################################

# The initial state of the lower-level lexer.
# The post_process function will be used only when an empty token is
# forced by calling 'Parser.flush()'.

low_level_lexer = Grammar(discard, name='low_level_lexer')

# By default, all ASCII characters parse as errors.

illegal_character = Unfinished("Illegal character", name='illegal_character')
low_level_lexer.set_all(
    char_range(chr(0), chr(127)), illegal_character, stream_name=discard)

# All punctuation and non-ASCII characters are passed on unchanged.

low_level_lexer.set_all(punctuation, Grammar(pass_on))
low_level_lexer.set('non_ascii', Grammar(pass_on))

# White-space characters other than newline are grouped.

white_space = Grammar(concat('white_space'), name='white_space')
low_level_lexer.set(' ', white_space)

# A newline character becomes a 'newline' token, which is emitted eagerly.
# A carriage return before it is discarded.

newline = Grammar(concat('newline'), name=r'\n', is_eager=True)
low_level_lexer.set('\n', newline)
low_level_lexer.get_copy('\r').set('\n', newline)

# A semicolon is passed on eagerly.

low_level_lexer.set(';', Grammar(pass_on, is_eager=True))

# A string of decimal digits is an integer.

integer = Grammar(concat('integer'), name='integer')
low_level_lexer.set_all(char_range('0', '9'), integer)
integer.set_all(char_range('0', '9'), integer)

# After an integer, a dot makes it a float.
# After a dot, an integer makes it a float.
# The 'float' tag has:
# 1. An optional integer part (of low-level type 'integer').
# 2. A dot.
# 3. An optional fraction part (of low-level type 'integer').
# 4. An 'E' or an 'e'.
# 5. An optional '-' or '+'.
# 6. An exponent part (of low-level type 'integer').
# It may be truncated after parts 2 or 3, but not after parts 1, 4 or 5.

float_exponent_or_end = Grammar(concat('float'), name='float_exponent_or_end')
float_exponent_or_end.set_all(char_range('0', '9'), float_exponent_or_end)
float_exponent = Unfinished("Missing exponent")
float_exponent.set_all(char_range('0', '9'), float_exponent_or_end)
float_sign_or_exponent = Unfinished("Missing exponent")
float_sign_or_exponent.set_all('+-', float_exponent)
float_sign_or_exponent.set_all(char_range('0', '9'), float_exponent_or_end)
float_fraction = Grammar(concat('float'), name='float_fraction')
float_fraction.set_all(char_range('0', '9'), float_fraction)
float_fraction.set_all('Ee', float_sign_or_exponent)
integer.set_all('Ee', float_sign_or_exponent)
integer.set('.', float_fraction)
low_level_lexer.get_copy('.').set_all(char_range('0', '9'), float_fraction)

# A string of hexadecimal digits preceded by '0x' or '0X' is a hex integer.

hex_digit_or_end = Grammar(concat('hex_integer'), name='hex_digit_or_end')
hex_digit_or_end.set_all(char_range('0', '9'), hex_digit_or_end)
hex_digit_or_end.set_all(char_range('A', 'F'), hex_digit_or_end)
hex_digit_or_end.set_all(char_range('a', 'f'), hex_digit_or_end)
hex_digit = Unfinished('Bad hex integer', name='hex_digit')
hex_digit.set_all(char_range('0', '9'), hex_digit_or_end)
hex_digit.set_all(char_range('A', 'F'), hex_digit_or_end)
hex_digit.set_all(char_range('a', 'f'), hex_digit_or_end)
low_level_lexer.get_copy('0').set_all('xX', hex_digit)


# An alphabetic character or underscore begins an identifier or tag or keyword.
# It may also include digits after the first character.
# It's a keyword if it's in `keywords`, otherwise it's a tag if all letters
# are upper case and it doesn't start with an underscore, otherwise it's an
# identifier.

def make_identifier_etc(multi_buffer):
    asciis = multi_buffer.get()
    token = Concat('identifier', asciis)
    if token.s in keywords:
        token = Keyword(asciis)
    elif token.s == token.s.upper() and not token.s.startswith('_'):
        token = Concat('tag', asciis)
    return token


identifier = Grammar(make_identifier_etc, name='identifier')
identifier.set('_', identifier)
identifier.set_all(char_range('0', '9'), identifier)
identifier.set_all(char_range('A', 'Z'), identifier)
identifier.set_all(char_range('a', 'z'), identifier)

low_level_lexer.set('_', identifier)
low_level_lexer.set_all(char_range('A', 'Z'), identifier)
low_level_lexer.set_all(char_range('a', 'z'), identifier)

# Escapes are introduced by a backslash character.
# They normally only occur in string and character literals, but
# backslash characters are illegal everywhere else, so no ambiguity
# arises by lexing escape sequences wherever they occur.


def make_unicode_escape(multi_buffer):
    junk = multi_buffer.get('junk')
    asciis = multi_buffer.get()
    return Escape(junk + asciis, int(''.join(a.tag for a in asciis), 16))


unicode_escape = [
    Unfinished("Bad unicode escape sequence"),
    Unfinished("Bad unicode escape sequence"),
    Unfinished("Bad unicode escape sequence"),
    Unfinished("Bad unicode escape sequence"),
    Grammar(make_unicode_escape)
]
for n in range(4):
    unicode_escape[n].set_all(char_range('0', '9'), unicode_escape[n+1])
    unicode_escape[n].set_all(char_range('A', 'F'), unicode_escape[n+1])
    unicode_escape[n].set_all(char_range('a', 'f'), unicode_escape[n+1])


_escape_map = {'t': 9, 'n': 10, 'r': 13, '"': 34, "'": 39, '\\': 92}


def make_escape(multi_buffer):
    junk = multi_buffer.get('junk')
    ascii = multi_buffer.get_only()
    return Escape(junk + [ascii], _escape_map[ascii.tag])


escape = Unfinished("Bad escape sequence")
escape.set('u', unicode_escape[0], stream_name='junk')
escape.set_all(_escape_map, Grammar(make_escape))
escape.set('\n', Grammar(discard), stream_name='discard')
low_level_lexer.set('\\', escape, stream_name='junk')

#######################################################################

# The initial state of the higher-level lexer.
# The post_process function will be used only when an empty token is
# forced by calling 'Parser.flush()'.

high_level_lexer = Grammar(discard, name='high_level_lexer')

# Errors are passed on.

high_level_lexer.set('error', Grammar(pass_on, is_eager=True))

# By default, everything else is illegal.

illegal_character = Unfinished("Illegal character")
high_level_lexer.set_all(
    punctuation | keywords | low_level_lexer_tags, illegal_character)

# Improve the error message for illegal escape sequences.

high_level_lexer.set('escape', Unfinished(
    "Unexpected escape sequence. Escape sequences are only allowed "
    "inside string and character literals and comments"
))

# White-space tokens are silently ignored (except in strings etc).

high_level_lexer.set('white_space', high_level_lexer, stream_name='discard')

# Some tags are passed on, including some single characters.

high_level_lexer.set_all('.,:;', Grammar(pass_on))
high_level_lexer.set('integer', Grammar(pass_on))
high_level_lexer.set('float', Grammar(pass_on))
high_level_lexer.set('hex_integer', Grammar(pass_on))
high_level_lexer.set_all(keywords, Grammar(pass_on))
high_level_lexer.set('identifier', Grammar(pass_on))
high_level_lexer.set('tag', Grammar(pass_on))

# Newlines and semicolons are passed on eagerly.

high_level_lexer.set('newline', Grammar(pass_on, is_eager=True))
high_level_lexer.set(';', Grammar(pass_on, is_eager=True))

# Brackets are classified as 'open' or 'close'. Bracket flavour is
# recovered later.

high_level_lexer.set_all("([{", Grammar(concat('open')))
high_level_lexer.set_all(")]}", Grammar(concat('close')))

# Operators.

un_ops = {'+', '-', '~', '!', '++', '--', '&', '*'}
bin_ops = {
    '+', '-', '*', '**', '/', '%', '<<', '>>', '>>>', '&', '|', '^',
    '&&', '||', '<', '>', '<=', '>=', '==', '!=', '<>', '?',
    # ':' is not a binary operator as it is used in function literals.
    # Colons that are used as binary operators are turned into binary
    # operators at a higher level (atom).
}
un_bin_ops = un_ops & bin_ops
assign_ops = {
    '+=', '-=', '*=', '**=', '/=', '%=', '<<=', '>>=', '>>>=',
    '&=', '|=', '^=', '&&=', '||=', '='
}
all_ops = un_ops | bin_ops | assign_ops
prefixes = {}
for s in all_ops:
    for x in range(1, len(s)):
        t, u = s[:x], s[x:]
        if t not in prefixes:
            prefixes[t] = set()
        prefixes[t].add(u)
bad_ops = set()
for s in all_ops:
    for x in range(1, len(s)):
        t, u = s[:x], s[x:]
        if u in prefixes:
            for v in prefixes[u]:
                bad_ops.add(t+u+v)
bad_ops.difference_update(all_ops)
bad_ops = {s for s in bad_ops if s[:-1] not in bad_ops}
# print bad_ops
done = set()
for s in all_ops | bad_ops:
    state = high_level_lexer
    for n in range(len(s)):
        c, cs = s[n], s[:n+1]
        if cs not in done:
            done.add(cs)
            # print "Making state for %r[:%r] = %r" % (s, n+1, cs)
            if cs in assign_ops:
                new_state = Grammar(concat('assign_op'))
            elif cs in un_bin_ops:
                new_state = Grammar(concat('un_bin_op'))
            elif cs in un_ops:
                new_state = Grammar(concat('un_op'))
            elif cs in bin_ops:
                new_state = Grammar(concat('bin_op'))
            else:
                new_state = Unfinished("Ambiguous operator")
            state.set(c, new_state)
        state = state.get_copy(c)

high_level_lexer.get_copy(':').set('=', Unfinished(
    'Unknown operator ":=" (assignments use "=" and declarations start with '
    '"const" or "var")'
))

# '//' begins a one-line comment. Discarded.

one_line_comment = Grammar(discard)
high_level_lexer.get_copy('/').set('/', one_line_comment)
one_line_comment.set_all((
    (punctuation | keywords | low_level_lexer_tags | {'error'}) -
        {'newline'}
), one_line_comment, stream_name='discard')

# '/*' begins a multi-line comment. Discarded.

multi_line_comment = Unfinished("Unclosed comment")
high_level_lexer.get_copy('/').set('*', multi_line_comment)
multi_line_comment.set_all((
    punctuation | keywords | low_level_lexer_tags | {'error'}
), multi_line_comment, stream_name='discard')
multi_line_comment_star = multi_line_comment.get_copy('*')
multi_line_comment_star.set(
    '*', multi_line_comment_star, stream_name='discard')
multi_line_comment_star.set('/', Grammar(discard), stream_name='discard')

# Double quote introduces a string.


def make_string(multi_buffer):
    return String(
        multi_buffer.get_only(stream_name='open_quote'),
        multi_buffer.get(),
        multi_buffer.get_only(stream_name='close_quote'),
    )


string = Unfinished("Unclosed string literal")
string.set('"', Grammar(make_string), stream_name='close_quote')
string.set('error', string) # Preserve "Bad escape" errors.
string.set_all((
    (punctuation | keywords | low_level_lexer_tags) - {'newline', '"'}
), string)
high_level_lexer.set('"', string, stream_name='open_quote')

# Single quote introduces a character.


def make_character(multi_buffer):
    open_quote = multi_buffer.get_only(stream_name='open_quote')
    tokens = multi_buffer.get()
    close_quote = multi_buffer.get_only(stream_name='close_quote')
    length = 0
    for token in tokens:
        if token.tag == 'error':
            return token
        elif is_ascii(token) or token.tag in ('non_ascii', 'escape'):
            length += 1
        elif token.tag in keywords:
            length += len(token.tag)
        else:
            assert type(token) is Concat, token
            length += len(token.s)
    if length < 1:
        return Error(
            "Empty character literal",
            False,
            [open_quote] + tokens + [close_quote],
        )
    elif length > 1:
        return Error(
            "Character literals may only contain one character",
            False,
            [open_quote] + tokens + [close_quote],
        )
    else:
        assert len(tokens) == 1
        return Character(open_quote, tokens[0], close_quote)


character = Unfinished("Unclosed character literal")
character.set("'", Grammar(make_character), stream_name='close_quote')
character.set('error', character) # Preserve "Bad escape sequence" errors.
character.set_all((
    (punctuation | keywords | low_level_lexer_tags) - {'newline', "'"}
), character)
high_level_lexer.set("'", character, stream_name='open_quote')
