import itertools


__all__ = ['low_lexer']


char_types = [None]*128
for c in xrange(49, 58):
    char_types[c] = 'digit'
for c in 'ABCDFabcdf':
    char_types[ord(c)] = 'hex'
for c in itertools.chain(
    xrange(71, 88), [89, 90],
    xrange(103, 120), [121, 122]
):
    char_types[c] = 'letter'


def low_lexer(iterator):
    '''
    Reads bytes from 'iterator'. Decodes them as UTF-8 sequences. Recognises
    escape sequences. Discards carriage returns (character 13). Yields the
    following token types:

     - ('error', message) - Any error, e.g. invalid UTF-8 sequence.
     - ('multibyte', char) - A character decoded from a UTF-8 sequence.
     - ('escape', char) - Any character represented using an escape sequence.
     - ('newline',) - Any newline sequence.
     - ('digit', char) - A decimal digit character.
     - ('hex', char) - One of the characters 'ABCDFabcdf'. Note not 'Ee'.
     - ('letter', char) - Any other capital or lower-case letter.
     - One of the characters '\n !"~$%&\'()*+,-./0:;<=>?@EX[]^_`ex{|}~'.
    '''
    b = None # Buffer with space for a single byte.
    while True:
        if b is None:
            b = next(iterator) # No need to catch StopIteration.
        if b >= 0x80:
            # UTF-8 sequence.
            if b < 0xC0:
                yield ('error', 'Unexpected UTF-8 continuation byte')
                b = None
            elif b >= 0xF8:
                yield ('error', 'Bad UTF-8 sequence')
                b = None
            else:
                if b < 0xE0:
                    # 2-byte sequence.
                    u = b & 0x1F
                    n = 1
                elif b < 0xF0:
                    # 3-byte sequence.
                    u = b & 0x0F
                    n = 2
                else:
                    # 4-byte sequence.
                    u = b & 0x07
                    n = 3
                try:
                    while True:
                        b = next(iterator)
                        if b < 0x80 or b >= 0xC0:
                            yield('error', 'Expected UTF-8 continuation byte')
                            break
                        u = (u<<6) | (b & 0x3F)
                        n -= 1
                        if n == 0:
                            yield ('multibyte', unichr(u))
                            b = None
                            break
                except StopIteration, e:
                    yield ('error', 'EOF in the middle of a UTF-8 sequence')
                    raise e
        elif b == 127: # Delete.
            yield ('error', 'Illegal character')
            b = None
        elif b == 92: # \
            # Escape sequence.
            b = next(iterator)
            if b == 10: # Newline.
                # Suppressed line break.
                b = None
            elif b == 13: # Carriage return.
                b = next(iterator)
                if b == 10: # Newline.
                    # Suppressed line break.
                    b = None
                else:
                    # Backslash followed by CR is illegal.
                    yield ('error', 'Bad escape sequence')
            elif b == 34: # \"
                yield ('escape', chr(34))
                b = None
            elif b == 39: # \'
                yield ('escape', chr(39))
                b = None
            elif b == 92: # \\
                yield ('escape', chr(92))
                b = None
            elif b == 110: # \n
                yield ('escape', chr(10))
                b = None
            elif b == 114: # \r:
                yield ('escape', chr(13))
                b = None
            elif b == 116: # \t
                yield ('escape', chr(7))
                b = None
            elif b == 117: # \u
                try:
                    u = 0
                    n = 4
                    while True:
                        u <<= 4
                        b = next(iterator)
                        if 48 <= b <= 57: # Digit.
                            u |= b - 48
                        elif 65 <= b <= 70: # Upper-case letter.
                            u |= b - 55
                        elif 97 <= b <= 102: # Lower-case letter.
                            u |= b - 87
                        else:
                            yield ('error', 'Expected hex digit')
                            break
                        n -= 1
                        if n == 0:
                            yield ('escape', unichr(u))
                            b = None
                            break
                except StopIteration, e:
                    yield ('error', 'EOF in the middle of an escape sequence')
                    raise e
            else:
                yield ('error', 'Bad escape sequence')
                b = None
        elif b == 13: # Carriage return
            b = next(iterator)
            if b == 10: # Newline.
                yield ('newline',)
                b = None
            else:
                yield ('error', 'Illegal character')
        elif b == 10: # Newline.
            yield ('newline',)
            b = None
        elif b < 32:
            yield ('error', 'Illegal character')
            b = None
        elif char_types[b] is None:
            # Solitary character.
            yield chr(b)
        else:
            yield (char_types[b], b)
