"""
Defines a Grammar which recognises semicolon- or newline-separated statements.
"""


from token import (
    Assign, Case, Decl, Else, ElseIf, For, Fulfil, If, Return, Switch, Throw,
    While,
)
from grammar import Grammar
from util import discard, pass_on, Unfinished
from lexer import high_level_lexer_tags, keywords
from brackets import bracket_tags
from atom import atom_tags


__all__ = ['stmt', 'stmt_tags']

# The set of statement tags this parser can output. It can also output
# 'error' and those passed on from lower levels.

stmt_tags = [
    'assign', 'break', 'continue', 'case', 'decl', 'else', 'expr', 'for',
    'fulfil', 'if', 'return', 'switch', 'throw', 'while',
]

# Standard wording used for error message where an unexpected comma is
# found after an expression where a keyword is expected. This may indicate
# that the user is trying to write a tuple but has omitted the enclosing
# "{ ... }".

comma_hint = ' (structs, including tuples, must be enclosed in "( ... )")'

# The initial state of the statement grammars.

stmt = Grammar(discard, name='stmt')

# Unless overridden later, all tags are passed on unchanged.

stmt.set_all(high_level_lexer_tags, Grammar(pass_on))
stmt.set_all(keywords, Grammar(pass_on))
stmt.set_all(bracket_tags, Grammar(pass_on))
stmt.set_all(atom_tags, Grammar(pass_on))

# Errors, newlines and semicolons are passed on eagerly.

stmt.set_all(['error', 'newline', ';'], Grammar(pass_on, is_eager=True))

# Expressions.

stmt_expr_end = Grammar(pass_on, name='stmt_expr_end')
stmt.set('expr', stmt_expr_end)

# Assignments.

def make_assign(multi_buffer):
    '''
    Returns an Assign.
    '''
    lhs, op, rhs = multi_buffer.get()
    return Assign(lhs, op, rhs)


stmt_assign_end = Grammar(make_assign)
stmt_assign_rhs = Unfinished(
    "Expected an expression for the right-hand side of the assignment")
stmt_assign_rhs.ignore('newline')
stmt_assign_rhs.set('expr', stmt_assign_end)
stmt.get_copy('expr').set('assign_op', stmt_assign_rhs)


# "case" statements are only allowed inside "switch" statements. At this
# level, we can't tell if we're inside a switch statement, so we allow them
# everywhere.

def make_case(multi_buffer):
    '''
    Returns a Case.
    '''
    keyword, pattern, body = multi_buffer.get()
    return Case(keyword, pattern, body)


stmt_case_end = Grammar(make_case)
stmt_case_body = Unfinished('Expected "{ ... }" after "case" expression')
stmt_case_body.ignore('newline')
stmt_case_body.set('{}', stmt_case_end)
stmt_case_pattern = Unfinished('Expected an expression after "case"')
stmt_case_pattern.ignore('newline')
stmt_case_pattern.set('expr', stmt_case_body)
stmt.set('case', stmt_case_pattern)


# Declarations.

def make_decl(multi_buffer):
    '''
    Returns a Decl.
    '''
    decl, lhs, op, rhs = multi_buffer.get()
    return Decl(decl, lhs, op, rhs)


stmt_decl_end = Grammar(make_decl)
stmt_decl_rhs = Unfinished(
    "Expected an expression for the right-hand side of the declaration")
stmt_decl_rhs.ignore('newline')
stmt_decl_rhs.set('expr', stmt_decl_end)
stmt_decl_op = Unfinished('Expected "="')
stmt_decl_op.ignore('newline')
stmt_decl_op.set('assign_op', stmt_decl_rhs)
stmt_decl_lvalue = Unfinished('Expected a variable declaration')
stmt_decl_lvalue.ignore('newline')
stmt_decl_lvalue.set('expr', stmt_decl_op)
stmt.set_all(['var', 'const'], stmt_decl_lvalue)

# A grammar state to override errors from lower levels.
stmt_decl_op_error = Unfinished('Expected "="')
stmt_decl_op_error.ignore('error')
stmt_decl_op.set('error', stmt_decl_op_error)

# "else" statements are only allowed inside "switch" statements. At this
# level, we can't tell if we're inside a switch statement, so we allow them
# everywhere.
# Note that the "else" keyword can also be part of "if", "for" and "while";
# these don't parse as "else" statements.

def make_else(multi_buffer):
    '''
    Returns an Else.
    '''
    keyword, body = multi_buffer.get()
    return Else(keyword, body)


stmt_else_end = Grammar(make_else)
stmt_else_body = Unfinished('Expected "{ ... }" after "else"')
stmt_else_body.ignore('newline')
stmt_else_body.set('{}', stmt_else_end)
stmt.set('else', stmt_else_body)


# These grammar states exist to improve error messages when an "else if"
# clause becomes separated from its "if" statement.

else_if_hint = (
    '"else if" is only allowed as part of an "if" statement '
    '(on the same line)')

stmt_else_if_end = Unfinished(else_if_hint)
stmt_else_if_body = Unfinished(else_if_hint)
stmt_else_if_body.ignore('newline')
stmt_else_if_body.set('{}', stmt_else_if_end)
stmt_else_if_condition = Unfinished(else_if_hint)
stmt_else_if_condition.ignore('newline')
stmt_else_if_condition.set('expr', stmt_else_if_body)
stmt_else_body.set('if', stmt_else_if_condition)


# "for" statements.

def make_for(multi_buffer):
    '''
    Returns a For.
    '''
    else_keyword = multi_buffer.get_optional('else_keyword')
    if else_keyword is not None:
        else_ = Else(else_keyword, multi_buffer.get_only('else_body'))
    else:
        else_ = None
    return For(
        multi_buffer.get_only('for_keyword'),
        multi_buffer.get_only('variable'),
        multi_buffer.get_only('in_keyword'),
        multi_buffer.get_only('sequence'),
        multi_buffer.get_only('for_body'),
        else_,
    )


stmt_forelse_end = Grammar(make_for)
stmt_forelse_body = Unfinished('Expected "{ ... }" after "for ... else"')
stmt_forelse_body.ignore('newline')
stmt_forelse_body.set('{}', stmt_forelse_end, stream_name='else_body')
stmt_for_else_or_end = Grammar(make_for)
stmt_for_else_or_end.set('else', stmt_forelse_body, stream_name='else_keyword')
stmt_for_body = Unfinished('Expected "{ ... }" after "for ... in ..."')
stmt_for_body.ignore('newline')
stmt_for_body.set('{}', stmt_for_else_or_end, stream_name='for_body')
stmt_for_sequence = Unfinished('Expected a sequence after "for ... in"')
stmt_for_sequence.ignore('newline')
stmt_for_sequence.set('expr', stmt_for_body, stream_name='sequence')
stmt_for_in_comma_hint = Unfinished(
    'Expected "in" after the "for" loop variable' + comma_hint)
stmt_for_in_comma_hint.ignore_all(['in', 'else', 'expr', '{}'])
stmt_for_in = Unfinished('Expected "in" after the "for" loop variable')
stmt_for_in.ignore('newline')
stmt_for_in.set('in', stmt_for_sequence, stream_name='in_keyword')
# TODO: Enable this when "for (x, y) in ..." is allowed.
#stmt_for_in.set(',', stmt_for_in_comma_hint, stream_name='discard')
stmt_for_variable = Unfinished('Expected a loop variable after "for"')
stmt_for_variable.ignore('newline')
stmt_for_variable.set('expr', stmt_for_in, stream_name='variable')
stmt.set('for', stmt_for_variable, stream_name='for_keyword')


# "fulfil" statements.

def make_fulfil(multi_buffer):
    '''
    Returns a Fulfil.
    '''
    keyword, lhs, op, rhs = multi_buffer.get()
    return Fulfil(keyword, lhs, op, rhs)

stmt_fulfil_end = Grammar(make_fulfil)
stmt_fulfil_expr2 = Unfinished('Expected an expression after "fulfil"')
stmt_fulfil_expr2.ignore('newline')
stmt_fulfil_expr2.set('expr', stmt_fulfil_end)
stmt_fulfil_assign_op = Unfinished('Expected "=" after "fulfil"')
stmt_fulfil_assign_op.set('assign_op', stmt_fulfil_expr2)
stmt_fulfil_expr1 = Unfinished('Expected a deferred value after "fulfil"')
stmt_fulfil_expr1.ignore('newline')
stmt_fulfil_expr1.set('expr', stmt_fulfil_assign_op)
stmt.set('fulfil', stmt_fulfil_expr1)


# "if" statements.

def make_if(multi_buffer):
    '''
    Returns ``('if', condition, then_body, else_body)`` representing:

        if <condition> <then_body>
        [else <else_body>]

    '''
    if_keywords = multi_buffer.get('if_keywords')
    else_keywords = multi_buffer.get('else_keywords')
    conditions = multi_buffer.get('conditions')
    if_bodies = multi_buffer.get('if_bodies')
    else_body = multi_buffer.get_optional('else_body')
    if else_body is None:
        else_ = None
    else:
        else_ = Else(else_keywords.pop(), else_body)
    while True:
        result = If(
            if_keywords.pop(),
            conditions.pop(),
            if_bodies.pop(),
            else_,
        )
        if else_keywords:
            else_ = ElseIf(else_keywords.pop(), result)
        else:
            break
    assert if_keywords == []
    assert conditions == []
    assert if_bodies == []
    assert else_keywords == []
    return result


stmt_if_condition = Unfinished('Expected a condition after "if"')
stmt_ifelse_end = Grammar(make_if)
stmt_ifelse_if_or_body = Unfinished(
    'Expected "{ ... }" or "if" after "if ... else"')
stmt_ifelse_if_or_body.ignore('newline')
stmt_ifelse_if_or_body.set('if', stmt_if_condition, stream_name='if_keywords')
stmt_ifelse_if_or_body.set('{}', stmt_ifelse_end, stream_name='else_body')
stmt_if_else_or_end = Grammar(make_if)
stmt_if_else_or_end.set(
    'else', stmt_ifelse_if_or_body, stream_name='else_keywords')
stmt_if_body = Unfinished('Expected "{ ... }" after "if"')
stmt_if_body.ignore('newline')
stmt_if_body.set('{}', stmt_if_else_or_end, stream_name='if_bodies')
stmt_if_condition.ignore('newline')
stmt_if_condition.set('expr', stmt_if_body, stream_name='conditions')
stmt.set('if', stmt_if_condition, stream_name='if_keywords')

# "return" statements.

def make_return(multi_buffer):
    '''
    Returns a Return.
    '''
    return Return(
        multi_buffer.get_only(stream_name='keyword'),
        multi_buffer.get_optional(),
    )


stmt_return_end_comma_hint = Unfinished(
    'Expected the end of the statement' + comma_hint)
stmt_return_end_comma_hint.ignore_all([',', 'expr'])
stmt_return_end = Grammar(make_return)
stmt_return_end.set(',', stmt_return_end_comma_hint)
stmt_return_expr_or_end = Grammar(make_return)
stmt_return_expr_or_end.set('expr', stmt_return_end)
stmt.set('return', stmt_return_expr_or_end, stream_name='keyword')


# "switch" statements. See also "case" statements and "else" statements, which
# are really part of "switch" statements.

def make_switch(multi_buffer):
    '''
    Returns a Switch.
    '''
    keyword, condition, body = multi_buffer.get()
    return Switch(keyword, condition, body)


stmt_switch_end = Grammar(make_switch)
stmt_switch_body = Unfinished('Expected "{ ... }" after "switch" expression')
stmt_switch_body.ignore('newline')
stmt_switch_body.set('{}', stmt_switch_end)
stmt_switch_expr = Unfinished(
    'Expected an expression to examine after "switch"')
stmt_switch_expr.ignore('newline')
stmt_switch_expr.set('expr', stmt_switch_body)
stmt.set('switch', stmt_switch_expr)


# "throw" statements.

def make_throw(multi_buffer):
    '''
    Returns a Throw.
    '''
    return Throw(
        multi_buffer.get_only(stream_name='keyword'),
        multi_buffer.get_only(),
    )


stmt_throw_end = Grammar(make_throw)
stmt_throw_expr = Unfinished('Expected an identifier after "throw"')
stmt_throw_expr.ignore('newline')
stmt_throw_expr.set('expr', stmt_throw_end)
stmt.set('throw', stmt_throw_expr, stream_name='keyword')


# "while" statements.

def make_while(multi_buffer):
    '''
    Returns a While.
    '''
    else_keyword = multi_buffer.get_optional('else_keyword')
    if else_keyword is not None:
        else_ = Else(else_keyword, multi_buffer.get_only('else_body'))
    else:
        else_ = None
    return While(
        multi_buffer.get_only('while_keyword'),
        multi_buffer.get_only('condition'),
        multi_buffer.get_only('while_body'),
        else_,
    )

stmt_whileelse_end = Grammar(make_while)
stmt_whileelse_body = Unfinished('Expected "{ ... }" after "while ... else"')
stmt_whileelse_body.ignore('newline')
stmt_whileelse_body.set('{}', stmt_whileelse_end, stream_name='else_body')
stmt_while_else_or_end = Grammar(make_while)
stmt_while_else_or_end.set(
    'else', stmt_whileelse_body, stream_name='else_keyword')
stmt_while_body = Unfinished('Expected "{ ... }" after "while"')
stmt_while_body.ignore('newline')
stmt_while_body.set('{}', stmt_while_else_or_end, stream_name='while_body')
stmt_while_condition = Unfinished('Expected a condition after "while"')
stmt_while_condition.ignore('newline')
stmt_while_condition.set('expr', stmt_while_body, stream_name='condition')
stmt.set('while', stmt_while_condition, stream_name='while_keyword')
