from token import matches, Token, ASCII
from lexer import *
from brackets import BracketMatcher, bracket_tags
from atom import atom, atom_tags
from grammar import Parser, Buffer


allowed_tags = {'error'}.union(
    high_level_lexer_tags, keywords,
    bracket_tags,
    atom_tags,
)


observed_tags = set()


def check(source, expected_result):
    listener = Buffer()
    def make_parser(p):
        # Call-back used to make a Parser for each bracket nesting level.
        p = Parser(atom, p)
        return p
    p = make_parser(listener)  # top-level outside brackets
    p = BracketMatcher(p, {
        '(': make_parser,
        '[': make_parser,
        '{': make_parser,
    })
    p = Parser(high_level_lexer, p)
    p = Parser(low_level_lexer, p)
    for i, c in enumerate(source):
        p.feed(ASCII(i, c))
    p.flush()
    result = listener.contents
    for token in result:
        assert isinstance(token, Token), token
        assert token.tag in allowed_tags, token
        observed_tags.add(token.tag)
    assert matches(result, expected_result), (
        "Parsed:\n"
        "    %r\n"
        "as:\n"
        "    %r\n"
        "Expected:\n"
        "    %r\n" % (source, result, expected_result)
    )


def test_irrelevant():
    check('@', [
        ('error', 'Illegal character'),
    ])

    check('hello', [
        ('identifier', 'hello'),
    ])

    check('hello { world }', [
        ('identifier', 'hello'),
        ('{}', [('identifier', 'world')]),
    ])

    check('for', [
        ('for',),
    ])

    check('+', [
        ('un_bin_op', '+'),
    ])

    check('+=', [
        ('assign_op', '+='),
    ])

    check(':', [
        ('bin_op', ':'),
    ])

    check('::', [
        ('bin_op', ':'),
        ('bin_op', ':'),
    ])

    check(';', [(';',)])

    check('// break', [])


def test_functions():
    id_bool = ('identifier', 'bool')
    params_0 = ('()', [])
    type_0 = ('()', [])
    block_0 = ('{}', [])
    module = ('module',)
    error_missing_return_type_or_function_body = (
        'error', 'Expected a function body, or ":" and a return type')

    check('fn', [('error', 'Expected an identifier or "( ... )"')])
    check('fn@', [('error', 'Illegal character')])
    check('fn;', [('error', 'Expected an identifier or "( ... )"'), (';',)])
    check('fn foo;', [('error', 'Expected "( ... )"'), (';',)])
    check('fn foo@;', [('error', 'Illegal character'), (';',)])
    check('fn foo();', [error_missing_return_type_or_function_body, (';',)])
    check('fn();', [error_missing_return_type_or_function_body, (';',)])
    check('fn():;', [('error', 'Expected a return type'), (';',)])
    check('fn():@', [('error', 'Illegal character')])
    check('fn():bool;', [
        ('function', None, params_0, id_bool, None),
        (';',),
    ])
    check('fn():();', [
        ('function', None, params_0, type_0, None),
        (';',),
    ])
    check('fn():module;', [
        ('function', None, params_0, module, None),
        (';',),
    ])
    check('fn():bool{};', [
        ('function', None, params_0, id_bool, block_0),
        (';',)
    ])
    check('fn():(){};', [
        ('function', None, params_0, type_0, block_0),
        (';',)
    ])
    check('fn():module{};', [
        ('function', None, params_0, module, block_0),
        (';',)
    ])
    check('fn(){};', [
        ('function', None, params_0, None, block_0),
        (';',)
    ])
    check('fn()\n{}', [
        ('function', None, params_0, None, block_0),
    ])
    # `fn () {} : bool;` is `(fn () {}) : bool;`
    check('fn () {} : bool;', [
        ('function', None, params_0, None, block_0),
        ('bin_op', ':'),
        id_bool,
        (';',)
    ])
    # `fn () {} + bool;` is `(fn () {}) + bool;`
    check('fn () {} + bool;', [
        ('function', None, params_0, None, block_0),
        ('un_bin_op', '+'),
        id_bool,
        (';',)
    ])
    # `fn () + bool;` is `(fn () #ERROR#) + bool;`
    check('fn () + bool;', [
        error_missing_return_type_or_function_body,
        ('un_bin_op', '+'),
        id_bool,
        (';',)
    ])
    # `fn () + bool {};` is `(fn () #ERROR#) + bool {};`.
    check('fn () + bool {};', [
        error_missing_return_type_or_function_body,
        ('un_bin_op', '+'),
        id_bool,
        block_0,
        (';',)
    ])


def test_module():
    check('module', [('module', None, None)])
    check('module foo', [('error', 'Expected "{ ... }"')])
    check('module foo bar', [
        ('error', 'Expected "{ ... }"'),
        ('identifier', 'bar')
    ])
    check('module {}', [('module', None, ('{}', []))])
    check('module foo {}', [('module', ('identifier', 'foo'), ('{}', []))])
    check('module\nfoo\n{}', [
        ('module', None, None), ('newline', '\n'),
        ('identifier', 'foo'), ('newline', '\n'),
        ('{}', [])
    ])
    check('module foo {} bar', [
        ('module', ('identifier', 'foo'), ('{}', [])),
        ('identifier', 'bar')
    ])


def test_type_struct():
    check('struct', [('error', 'Expected an identifier or "( ... )"')])
    check('struct foo', [('error', 'Expected "( ... )"')])
    check('struct foo bar', [
        ('error', 'Expected "( ... )"'),
        ('identifier', 'bar')
    ])
    check('struct ()', [('type_struct', None, ('()', []))])
    check('struct\nfoo\n()', [
        ('type_struct', ('identifier', 'foo'), ('()', []))
    ])
    check('struct foo () bar', [
        ('type_struct', ('identifier', 'foo'), ('()', [])),
        ('identifier', 'bar')
    ])


def test_type_union():
    check('union', [('error', 'Expected an identifier or "{ ... }"')])
    check('union foo', [('error', 'Expected "{ ... }"')])
    check('union foo bar', [
        ('error', 'Expected "{ ... }"'),
        ('identifier', 'bar')
    ])
    check('union {}', [('type_union', None, ('{}', []))])
    check('union\nfoo\n{}', [
        ('type_union', ('identifier', 'foo'), ('{}', []))
    ])
    check('union foo {} bar', [
        ('type_union', ('identifier', 'foo'), ('{}', [])),
        ('identifier', 'bar')
    ])


def test_union():
    check('C', [('union', ('tag', 'C'), None)])
    check('C()', [('union', ('tag', 'C'), ('()', []))])
    check('C(1)', [('union', ('tag', 'C'), ('()', [('integer', '1')]))])
    check('c', [('identifier', 'c')])
    check('c()', [('identifier', 'c'), ('()', [])])
    check('c(1)', [('identifier', 'c'), ('()', [('integer', '1')])])


def test_field():
    check('hello.world', [
        ('identifier', 'hello'),
        ('field', ('identifier', 'world')),
    ])
    check('hello.', [
        ('identifier', 'hello'),
        ('error', 'Expected an identifier after "."'),
    ])


def test_array():
    check('array', [('array', None)])
    check('array const', [('array', ('const',))])
    check('array var', [('array', ('var',))])


def test_ref():
    check('ref', [('ref', None)])
    check('ref const', [('ref', ('const',))])
    check('ref var', [('ref', ('var',))])


def test_colon():
    check(':', [('bin_op', ':')])


def test_coverage():
    for t0 in atom_tags:
        assert t0 in observed_tags, t0
