from token import Token, matches, ASCII
from grammar import Buffer, Parser
from lexer import *
from brackets import BracketMatcher, bracket_tags
from atom import atom, atom_tags
from unary_expr import unary_expr, unary_expr_tags
from expr import expr, expr_tags


allowed_tags = {'error'}.union(
    high_level_lexer_tags, keywords,
    bracket_tags,
    atom_tags,
    unary_expr_tags,
    expr_tags,
)


observed_tags = set()


def check(source, expected_result):
    # Parse `tokens`.
    listener = Buffer()
    def make_parser(p):
        # Call-back used to make a Parser for each bracket nesting level.
        p = Parser(expr, p)
        p = Parser(unary_expr, p)
        p = Parser(atom, p)
        return p
    p = make_parser(listener) # Top-level outside brackets.
    p = BracketMatcher(p, {
        '(': make_parser,
        '[': make_parser,
        '{': make_parser,
    })
    p = Parser(high_level_lexer, p)
    p = Parser(low_level_lexer, p)
    for i, c in enumerate(source):
        p.feed(ASCII(i, c))
    p.flush()
    result = listener.contents
    # Study results.
    for token in result:
        assert isinstance(token, Token)
        assert token.tag in allowed_tags, token
        observed_tags.add(token.tag)
    assert matches(result, expected_result), (
        "Parsed:\n"
        "    %r\n"
        "as:\n"
        "    %r\n"
        "Expected:\n"
        "    %r\n" % (source, result, expected_result)
    )


def test_errors():
    check('@', [('error', 'Illegal character')])
    check('1+@', [('error', 'Illegal character')])


def test_irrelevant():
    check('for', [
        ('for',),
    ])


def test_expr():
    check('hello', [
        ('expr', None, ('identifier', 'hello'), None),
    ])

    check('( )', [
        ('expr', None, ('()', []), None),
    ])

    check('( ( ) )', [
        ('expr', None, ('()', [('expr', None, ('()', []), None)]), None),
    ])

    check('( world )', [
        ('expr', None, ('()', [
            ('expr', None, ('identifier', 'world'), None)
        ]), None),
    ])

    check('0==1', [
        ('expr',
            ('expr', None, ('integer', '0'), None),
            ('bin_op', '=='),
            ('expr', None, ('integer', '1'), None))
    ])

    check('hello + world', [
        ('expr',
            ('expr', None, ('identifier', 'hello'), None),
            ('un_bin_op', '+'),
            ('expr', None, ('identifier', 'world'), None),
        )
    ])

    # a ? b : c is parsed as two separate operators.
    check('hello ? world', [
        ('expr',
            ('expr', None, ('identifier', 'hello'), None),
            ('bin_op', '?'),
            ('expr', None, ('identifier', 'world'), None),
        )
    ])

    check('hello : world', [
        ('expr',
            ('expr', None, ('identifier', 'hello'), None),
            ('bin_op', ':'),
            ('expr', None, ('identifier', 'world'), None),
        )
    ])

    check('hello { world }', [
        ('expr', None, ('identifier', 'hello'), None),
        ('{}', [
            ('expr', None, ('identifier', 'world'), None)
        ]),
    ])

    check('hello ( world )', [
        ('expr',
            ('expr', None, ('identifier', 'hello'), None),
            ('()', [
                ('expr', None, ('identifier', 'world'), None)
            ]),
            None,
        )
    ])

    check('hello [ world ]', [
        ('expr',
            ('expr', None, ('identifier', 'hello'), None),
            ('[]', [
                ('expr', None, ('identifier', 'world'), None)
            ]),
            None,
        )
    ])

    check('hello.world', [
        ('expr',
            ('expr', None, ('identifier', 'hello'), None),
            ('field', ('identifier', 'world')),
            None,
        )
    ])


def test_ternary():
    check('goodbye ? cruel : world', [
        ('expr',
            ('expr', None, ('identifier', 'goodbye'), None),
            ('bin_op', '?'),
            ('expr',
                ('expr', None, ('identifier', 'cruel'), None),
                ('bin_op', ':'),
                ('expr', None, ('identifier', 'world'), None),
            ),
        )
    ])


def test_nested():
    check('goodbye.cruel.world', [
        ('expr',
            ('expr',
                ('expr', None, ('identifier', 'goodbye'), None),
                ('field', ('identifier', 'cruel')),
                None,
            ),
            ('field', ('identifier', 'world')),
            None,
        )
    ])

    check('sizeof typeof array int', [
        # "sizeof" is not a keyword, but "typeof" and "array" are.
        ('expr', None, ('identifier', 'sizeof'), None),
        ('expr',
            None,
            ('typeof',),
            ('expr',
                None,
                ('array', None),
                ('expr', None, ('identifier', 'int'), None)
            ),
        ),
    ])

    check('sizeof(typeof array int)', [
        ('expr',
            ('expr', None, ('identifier', 'sizeof'), None),
            ('()', [
                ('expr',
                    None,
                    ('typeof',),
                    ('expr',
                        None,
                        ('array', None),
                        ('expr', None, ('identifier', 'int'), None)
                    ),
                ),
            ]),
            None,
        ),
    ])


def test_silly():
    check(r'~"foo".x^', [
        ('error', "Expected an expression")
    ])

    check(r'~"foo".x^0x1', [
        ('expr',
            ('expr',
                None,
                ('un_op', '~'),
                ('expr',
                    ('expr', None, ('string', [u'foo']), None),
                    ('field', ('identifier', 'x')),
                    None,
                ),
            ),
            ('bin_op', '^'),
            ('expr', None, ('hex_integer', '0x1'), None),
        )
    ])


def test_precedence():
    check('goodbye+cruel*world', [
        ('expr',
            ('expr', None, ('identifier', 'goodbye'), None),
            ('un_bin_op', '+'),
            ('expr',
                ('expr', None, ('identifier', 'cruel'), None),
                ('un_bin_op', '*'),
                ('expr', None, ('identifier', 'world'), None),
            ),
        ),
    ])

    check('goodbye*cruel+world', [
        ('expr',
            ('expr',
                ('expr', None, ('identifier', 'goodbye'), None),
                ('un_bin_op', '*'),
                ('expr', None, ('identifier', 'cruel'), None),
            ),
            ('un_bin_op', '+'),
            ('expr', None, ('identifier', 'world'), None),
        ),
    ])

    check('-goodbye-cruel.world', [
        ('expr',
            ('expr',
                None,
                ('un_bin_op', '-'),
                ('expr', None, ('identifier', 'goodbye'), None),
            ),
            ('un_bin_op', '-'),
            ('expr',
                ('expr', None, ('identifier', 'cruel'), None),
                ('field', ('identifier', 'world')),
                None,
            ),
        ),
    ])


def test_associativity():
    check('goodbye*cruel*world', [
        ('expr',
            ('expr',
                ('expr', None, ('identifier', 'goodbye'), None),
                ('un_bin_op', '*'),
                ('expr', None, ('identifier', 'cruel'), None),
            ),
            ('un_bin_op', '*'),
            ('expr', None, ('identifier', 'world'), None),
        ),
    ])

    check('goodbye**cruel**world', [
        ('expr',
            ('expr', None, ('identifier', 'goodbye'), None),
            ('bin_op', '**'),
            ('expr',
                ('expr', None, ('identifier', 'cruel'), None),
                ('bin_op', '**'),
                ('expr', None, ('identifier', 'world'), None),
            ),
        ),
    ])


def test_for():
    # Check that we haven't done anything that breaks the "for" syntax.
    check('for x in sequence {}', [
        ('for',),
        ('expr', None, ('identifier', 'x'), None),
        ('in',),
        ('expr', None, ('identifier', 'sequence'), None),
        ('{}', []),
    ])


def test_interruption():
    check('-', [('error', 'Expected an expression')])

    check('1-', [('error', 'Expected an expression')])


def test_coverage():
    for t0 in expr_tags:
        assert t0 in observed_tags, t0
