import pprint

from token import Token, matches, ASCII
from grammar import Buffer, Parser
from lexer import *
from brackets import BracketMatcher
from atom import atom
from unary_expr import unary_expr
from expr import expr
from expr_list import expr_list


allowed_tags = {'error', 'expr_list', 'expr'}


def check(source, expected_result):
    listener = Buffer()
    p = Parser(expr_list, listener)
    p = Parser(expr, p)
    p = Parser(unary_expr, p)
    p = Parser(atom, p)
    p = BracketMatcher(p, {
        '(': lambda x: x,
        '[': lambda x: x,
        '{': lambda x: x,
    })
    p = Parser(high_level_lexer, p)
    p = Parser(low_level_lexer, p)
    for i, c in enumerate(source):
        p.feed(ASCII(i, c))
    p.flush()
    result = listener.contents
    if expected_result is None:
        assert len(result) == 0, (
            "Parsed:\n"
            "    %r\n"
            "as:\n"
            "    %s\n"
            "which is not empty." % (
                source,
                pprint.pformat(result).replace('\n', '\n    '),
            )
        )
    else:
        assert len(result) == 1, (
            "Parsed:\n"
            "    %r\n"
            "as:\n"
            "    %s\n"
            "which is not one token." % (
                source,
                pprint.pformat(result).replace('\n', '\n    '),
            )
        )
        (token,) = result
        assert isinstance(token, Token)
        assert token.tag in allowed_tags, token
        assert matches(token, expected_result), (
            "Parsed:\n"
            "    %r\n"
            "as:\n"
            "    %s\n"
            "Expected:\n"
            "    %s\n" % (
                source,
                pprint.pformat(token).replace('\n', '\n    '),
                pprint.pformat(expected_result).replace('\n', '\n    ')
            )
        )


error_illegal = ('error', 'Illegal character')
error_comma = ('error', 'Expected "," or a closing bracket')
error_expr = ('error', 'Expected an expression or a closing bracket')
expr_1 = ('expr', None, ('integer', '1'), None)
comma = (',',)

def test_empty():
    check('', None)
    check('\n', None)
    check('\n\n', None)

def test_expr():
    check('1', ('expr_list', [expr_1], []))
    check('\n1\n', ('expr_list', [expr_1], []))

def test_list():
    # Correct cases. Newlines are to demonstrate that they don't matter.
    check('1,', ('expr_list', [expr_1], [comma]))
    check('\n1\n,\n', ('expr_list', [expr_1], [comma]))
    check('1,1', ('expr_list', [expr_1]*2, [comma]))
    check('\n1\n,\n1\n', ('expr_list', [expr_1]*2, [comma]))
    check('1,1,', ('expr_list', [expr_1]*2, [comma]*2))
    check('\n1\n,\n1\n,\n', ('expr_list', [expr_1]*2, [comma]*2))
    check('1,1,1', ('expr_list', [expr_1]*3, [comma]*2))
    check('\n1\n,\n1\n,\n1\n', ('expr_list', [expr_1]*3, [comma]*2))

def test_illegal():
    # An error from a lower level replaces the entire list. Anything after
    # the error ("+ x" below) is ignored.
    for x in ['@', '1', ',', ';']:
        check('@' + x, error_illegal)
        check('1 @' + x, error_illegal)
        check('1 , @' + x, error_illegal)
        check('1 , 1 @' + x, error_illegal)
        check('1 , 1 , @' + x, error_illegal)
        check('1 , 1 , 1 @' + x, error_illegal)

def test_bad():
    # Any token other than expressions and commas generates a new error
    # which replaces the entire list. Anything after the error ("+ x" below)
    # is ignored.
    for x in ['@', '1', ',', ';']:
        check('; ' + x, error_expr)
        check('1 ; ' + x, error_comma)
        check('1 , ; ' + x, error_expr)
        check('1 , 1 ; ' + x, error_comma)
        check('1 , 1 , ; ' + x, error_expr)
        check('1 , 1 , 1 ; ' + x, error_comma)

def test_wrong():
    # If expressions and commas don't alternate properly it generates a new
    # error which replaces the entire list. Anything after the error ("+ x"
    # below) is ignored.
    for x in ['@', '1', ',', ';']:
        check(', ' + x, error_expr)
        check('1 1 ' + x, error_comma)
        check('1 , , ' + x, error_expr)
        check('1 , 1 1 ' + x, error_comma)
        check('1 , 1 , , ' + x, error_expr)
        check('1 , 1 , 1 1 ' + x, error_comma)
