import pprint

from token import matches, Token, ASCII
from grammar import Buffer, Parser
from lexer import *
from brackets import BracketMatcher, bracket_tags
from atom import atom, atom_tags
from unary_expr import unary_expr, unary_expr_tags
from expr import expr, expr_tags
from expr_list import expr_list
from stmt import stmt, stmt_tags


allowed_tags = {'error'}.union(
    high_level_lexer_tags, keywords,
    bracket_tags,
    atom_tags,
    unary_expr_tags,
    expr_tags,
    stmt_tags,
)


observed_tags = set()


def check(source, expected_result):
    result = parse_stmts(source)
    for token in result:
        assert isinstance(token, Token), token
        assert token.tag in allowed_tags, token
        observed_tags.add(token.tag)
    assert matches(result, expected_result), (
        "Parsed:\n"
        "    %r\n"
        "as:\n"
        "    %s\n"
        "Expected:\n"
        "    %s\n" % (
            source,
            pprint.pformat(result).replace('\n', '\n    '),
            pprint.pformat(expected_result).replace('\n', '\n    ')
        )
    )


# Utility functions.


def parse_stmts(source):
    '''Note that parsing stops at stmts, which may be exprs.'''
    listener = Buffer()
    def make_stmt_parser(p):
        p = Parser(stmt, p)
        p = Parser(expr, p)
        p = Parser(unary_expr, p)
        p = Parser(atom, p)
        return p
    def make_expr_parser(p):
        p = Parser(expr_list, p)
        p = Parser(expr, p)
        p = Parser(unary_expr, p)
        p = Parser(atom, p)
        return p
    p = make_stmt_parser(listener) # Top-level outside brackets.
    p = BracketMatcher(p, {
        '(': make_expr_parser,
        '[': make_expr_parser,
        '{': make_stmt_parser,
    })
    p = Parser(high_level_lexer, p)
    p = Parser(low_level_lexer, p)
    for i, c in enumerate(source):
        p.feed(ASCII(i, c))
    p.flush()
    return listener.contents


def nonfix(op):
    return ('expr', None, op, None)


def prefix(op, e):
    return ('expr', None, op, e)


def postfix(e, op):
    return ('expr', e, op, None)


def infix(e1, op, e2):
    return ('expr', e1, op, e2)


def int_(n):
    return nonfix(('integer', str(n)))


def str_(s):
    return nonfix(('string', [unicode(s)]))


def id_(name):
    return nonfix(('identifier', name))


def call(f, exprs, commas):
    assert type(exprs) is list
    assert type(commas) is list
    assert len(exprs) - 1 <= len(commas) <= len(exprs)
    return postfix(f, bracketed('()', exprs, commas))


def dot(e, name):
    return postfix(e, ('field', ('identifier', name)))


def braced(*tokens):
    '''Helper for constructing the parse-tree of "{ tokens }".'''
    return ('{}', list(tokens))


def bracketed(tag, exprs, commas):
    '''Helper for constructing the parse-tree of "(expr, ...)".'''
    assert type(exprs) is list
    assert type(commas) is list
    assert len(exprs) - 1 <= len(commas) <= len(exprs)
    if len(exprs) == 0:
        contents = []
    else:
        contents = [('expr_list', exprs, commas)]
    return (tag, contents)


def union(name, exprs, commas):
    return nonfix(('union', ('tag', name), bracketed('()', exprs, commas)))


def test_expr_comma_semicolon():
    check('3', [int_(3)])
    check('3;', [int_(3), (';',)])
    check('3,', [int_(3), (',',)])
    check('x', [id_('x')])
    check('x;', [id_('x'), (';',)])
    check('x y', [id_('x'), id_('y')])
    check('x; y', [id_('x'), (';',), id_('y')])
    check('3x', [int_(3), id_('x')])
    check('3 x', [int_(3), id_('x')])
    check('0x', [('error', 'Bad hex integer')])
    check('0 x', [int_(0), id_('x')])
    check('print("Hello")', [call(id_('print'), [str_("Hello")], [])])
    check('f(x, x, x)', [call(id_('f'), [id_('x')]*3, [(',',)]*2)])
    check('()', [nonfix(bracketed('()', [], []))])
    check('[]', [nonfix(bracketed('[]', [], []))])


def test_assignments():
    check('x =', [('error',
        'Expected an expression for the right-hand side of the assignment'
    )])

    check('x +=', [('error',
        'Expected an expression for the right-hand side of the assignment'
    )])

    check('x = something', [('assign',
        id_('x'),
        ('assign_op', '='),
        id_('something'),
    )])

    check('x += something', [('assign',
        id_('x'),
        ('assign_op', '+='),
        id_('something'),
    )])


def test_braced():
    check('{}', [braced()])
    check('{3}', [braced(int_(3))])
    check('{ return 2 }', [braced(('return', int_(2)))])


# We do "return" statements early so we can use them in other tests.
def test_return():
    check('return', [('return', None)])
    check('return 2', [('return', int_(2))])


def test_break():
    check('break', [('break',)])


def test_case():
    check('case', [('error', 'Expected an expression after "case"')])
    check('case VAR', [
        ('error', 'Expected "{ ... }" after "case" expression')])
    check('case VAR()', [
        ('error', 'Expected "{ ... }" after "case" expression')])
    check('case VAR() {}', [('case', union('VAR', [], []), braced())])


def test_continue():
    check('continue', [('continue',)])


def test_declarations():
    check('var', [('error', 'Expected a variable declaration')])

    check('var x', [('error', 'Expected "="')])

    check('var x =', [('error',
        'Expected an expression for the right-hand side of the declaration'
    )])

    check('var x +=', [('error',
        'Expected an expression for the right-hand side of the declaration'
    )])

    check('var x = something', [('decl',
        ('var',),
        id_('x'),
        ('assign_op', '='),
        id_('something'),
    )])

    check('const x = something', [('decl',
        ('const',),
        id_('x'),
        ('assign_op', '='),
        id_('something'),
    )])

    # "var x += ..." and "const x += ..." fail in the AST builder.
    check('var x += something', [('decl',
        ('var',),
        id_('x'),
        ('assign_op', '+='),
        id_('something'),
    )])


def test_else():
    else_if_hint = ('error',
        '"else if" is only allowed as part of an "if" statement '
        '(on the same line)')

    check('else', [('error', 'Expected "{ ... }" after "else"')])
    check('else {}', [('else', ('{}', []))])
    check('else if', [else_if_hint])
    check('else if 1', [else_if_hint])
    check('else if 1 {}', [else_if_hint])
    check('else if {}', [else_if_hint, ('{}', [])])


def test_for():
    party = braced(id_('party'))
    fly = braced(id_('fly'))
    complain = braced(id_('complain'))
    sleep = braced(id_('sleep'))

    check('for x in seq { sum += x; }', [
        ('for', id_('x'), id_('seq'), braced(
            ('assign',
                id_('sum'),
                ('assign_op', '+='),
                id_('x')
            ),
            (';',),
        ), None),
    ])

    check('for', [
        ('error', 'Expected a loop variable after "for"'),
    ])

    check('for x', [
        ('error', 'Expected "in" after the "for" loop variable'),
    ])

    check('for x in', [
        ('error', 'Expected a sequence after "for ... in"'),
    ])

    check('for x in seq', [
        ('error', 'Expected "{ ... }" after "for ... in ..."'),
    ])

    check('for x in seq;', [
        ('error', 'Expected "{ ... }" after "for ... in ..."'),
        (';',),
    ])

    check('for x in seq { party }', [
        ('for', id_('x'), id_('seq'), party, None),
    ])

    check('for x in seq { party } else', [
        ('error', 'Expected "{ ... }" after "for ... else"'),
    ])

    check('for x in seq { party } else if pigs { fly }', [
        ('error', 'Expected "{ ... }" after "for ... else"'),
        ('if', id_('pigs'), fly, None),
    ])

    check('for x in seq { party } else for fly in soup { complain }', [
        ('error', 'Expected "{ ... }" after "for ... else"'),
        ('for', id_('fly'), id_('soup'), complain, None),
    ])

    check('for x in seq { party } else { sleep }', [
        ('for', id_('x'), id_('seq'), party, ('else', sleep)),
    ])

    check('for x in seq { party }; else { sleep }', [
        ('for', id_('x'), id_('seq'), party, None),
        (';',),
        ('else', sleep),
    ])


def test_if():
    sulk = braced(id_('sulk'))
    cheer = braced(id_('cheer'))
    sulk_but_harder = braced(id_('sulk_but_harder'))

    check('if 1 { return 2 }', [
        ('if', int_(1), braced(('return', int_(2))), None),
    ])

    check('if', [
        ('error', 'Expected a condition after "if"'),
    ])

    check('if 0==1', [
        ('error', 'Expected "{ ... }" after "if"'),
    ])

    check('if 0==1 { sulk }', [
        ('if', infix(int_(0), ('bin_op', '=='), int_(1)), sulk, None),
    ])

    check('if 0==1 { sulk };', [
        ('if', infix(int_(0), ('bin_op', '=='), int_(1)), sulk, None),
        (';',),
    ])

    check('if 0==1 { sulk } else', [
        ('error', 'Expected "{ ... }" or "if" after "if ... else"'),
    ])

    check('if 0==1 { sulk } else { cheer }', [
        ('if',
            infix(int_(0), ('bin_op', '=='), int_(1)),
            sulk,
            ('else', cheer),
        ),
    ])

    check('if 0==1 { sulk }; else { cheer }', [
        ('if', infix(int_(0), ('bin_op', '=='), int_(1)), sulk, None),
        (';',),
        ('else', cheer),
    ])

    check('if 0==1 { sulk } else if 1==1', [
        ('error', 'Expected "{ ... }" after "if"'),
    ])

    check('if 0==1 { sulk } else if 1==1 { cheer }', [
        ('if',
            infix(int_(0), ('bin_op', '=='), int_(1)),
            sulk,
            ('else_if',
                ('if',
                    infix(int_(1), ('bin_op', '=='), int_(1)),
                    cheer,
                    None,
                ),
            ),
        ),
    ])

    check(
        'if 0==1 { sulk } else if 1==1 { cheer } else { sulk_but_harder }',
        [
            ('if',
                infix(int_(0), ('bin_op', '=='), int_(1)),
                sulk,
                ('else_if',
                    ('if',
                        infix(int_(1), ('bin_op', '=='), int_(1)),
                        cheer,
                        ('else', sulk_but_harder),
                    ),
                ),
            ),
        ],
    )

    check(
        'if 0==1 { sulk }; else if 1==1 { cheer }; else { sulk_but_harder }',
        [
            ('if', infix(int_(0), ('bin_op', '=='), int_(1)), sulk, None),
            (';',),
            ('error',
                '"else if" is only allowed as part of an '
                '"if" statement (on the same line)'
            ),
            (';',),
            ('else', sulk_but_harder),
        ],
    )

    check('else', [('error', 'Expected "{ ... }" after "else"')])


def test_fulfil():
    check('fulfil x = t', [
        ('fulfil', id_('x'), ('assign_op', '='), id_('t'))
    ])

    check('fulfil x =', [
        ('error', 'Expected an expression after "fulfil"')
    ])

    check('fulfil x', [
        ('error', 'Expected "=" after "fulfil"')
    ])

    check('fulfil', [
        ('error', 'Expected a deferred value after "fulfil"')
    ])


def test_switch():
    check('switch', [
        ('error', 'Expected an expression to examine after "switch"')
    ])

    check('switch 3', [
        ('error', 'Expected "{ ... }" after "switch" expression')
    ])

    check('switch 3 else {}', [
        ('error', 'Expected "{ ... }" after "switch" expression'),
        ('else', braced()),
    ])

    check('switch 3 case VAR(name) {}', [
        ('error', 'Expected "{ ... }" after "switch" expression'),
        ('case', union('VAR', [id_('name')], []), braced()),
    ])

    # The malformed pattern is not detected at this level.
    check('''\
        switch t {
            case Var(name) { return 1 }
        }\
    ''', [
        ('switch',
            id_('t'),
            braced(
                ('newline', '\n'),
                ('case',
                    call(id_('Var'), [id_('name')], []),
                    braced(('return', int_(1)))
                ),
                ('newline', '\n'),
            ),
        )
    ])

    check('''\
        switch t {
            case VAR(name) { return 1 }
            case FUN(arg, body) { return 2 }
            case APP(f, v) { return 3 }
        }\
    ''', [
        ('switch',
            id_('t'),
            braced(
                ('newline', '\n'),
                ('case',
                    union('VAR', [id_('name')], []),
                    braced(('return', int_(1)))
                ),
                ('newline', '\n'),
                ('case',
                    union('FUN', [id_('arg'), id_('body')], [(',',)]),
                    braced(('return', int_(2)))
                ),
                ('newline', '\n'),
                ('case',
                    union('APP', [id_('f'), id_('v')], [(',',)]),
                    braced(('return', int_(3)))
                ),
                ('newline', '\n'),
            ),
        )
    ])


def test_throw():
    check('throw', [
        ('error', 'Expected an identifier after "throw"'),
    ])
    check('throw foo', [
        ('throw', id_('foo')),
    ])


def test_while():
    sulk = braced(id_('sulk'))
    fly = braced(id_('fly'))
    cheer = braced(id_('cheer'))

    check('while it.hasNext() { it.next(); }', [
        ('while',
            call(dot(id_('it'), 'hasNext'), [], []),
            braced(call(dot(id_('it'), 'next'), [], []), (';',)),
            None
        ),
    ])

    check('while', [
        ('error', 'Expected a condition after "while"'),
    ])

    check('while 1', [
        ('error', 'Expected "{ ... }" after "while"'),
    ])

    check('while 1;', [
        ('error', 'Expected "{ ... }" after "while"'),
        (';',),
    ])

    check('while 1==1 { sulk }', [
        ('while', infix(int_(1), ('bin_op', '=='), int_(1)), sulk, None),
    ])

    check('while 1==1 { sulk };', [
        ('while', infix(int_(1), ('bin_op', '=='), int_(1)), sulk, None),
        (';',),
    ])

    check('while 1==1 { sulk } else', [
        ('error', 'Expected "{ ... }" after "while ... else"'),
    ])

    check('while 1==1 { sulk } else if pigs { fly }', [
        ('error', 'Expected "{ ... }" after "while ... else"'),
        ('if', id_('pigs'), fly, None),
    ])

    check('while 1==1 { sulk } else while pigs { fly }', [
        ('error', 'Expected "{ ... }" after "while ... else"'),
        ('while', id_('pigs'), fly, None),
    ])

    check('while 1==1 { sulk } else { cheer }', [
        (
            'while',
            infix(int_(1), ('bin_op', '=='), int_(1)),
            sulk,
            ('else', cheer)
        ),
    ])

    check('while 1==1 { sulk }; else { cheer }', [
        ('while', infix(int_(1), ('bin_op', '=='), int_(1)), sulk, None),
        (';',),
        ('else', cheer),
    ])


def test_silly():
    check('''if 1 { return 2 } else if 3 { while 4 { if 5 { 6 } else if { 7 } 8 } } else { 9; return 10 }''', [
        ('if',
            int_(1),
            braced(('return', int_(2))),
            ('else_if',
                ('if',
                    int_(3),
                    braced(
                        ('while',
                            int_(4),
                            braced(
                                ('error', 'Expected a condition after "if"'),
                                braced(int_(7)),
                                int_(8),
                            ),
                            None,
                        ),
                    ),
                    ('else',
                        braced(int_(9), (';',), ('return', int_(10)),
                        ),
                    ),
                ),
            ),
        ),
    ])


def test_coverage():
    for t0 in stmt_tags:
        assert t0 in observed_tags, t0
