from token import *


def make_asciis(source):
    return [ASCII(pos, c) for pos, c in enumerate(source)]


def test_error():
    tokens = make_asciis(' @ ')
    e = Error('Illegal character', False, tokens[1:2])
    assert e.tag == 'error'
    assert e.start == 1
    assert e.end == 2
    assert e.message == 'Illegal character'


def test_ascii():
    a = ASCII(5, 'a')
    assert a.tag == 'a'
    assert a.start == 5
    assert a.end == 6


def test_nonascii():
    a = NonASCII(5, 8, u'\N{snowman}')
    assert a.tag == 'non_ascii'
    assert a.start == 5
    assert a.end == 8
    assert a.c == u'\N{snowman}'


def test_escape():
    tokens = make_asciis(r'"world\n"')
    e = Escape(tokens[7:9], 10)
    assert e.tag == 'escape'
    assert e.start == 7
    assert e.end == 9
    assert e.code_point == 10


def test_concat():
    tokens = make_asciis(r' print ')
    c = Concat('identifier', tokens[1:6])
    assert c.tag == 'identifier'
    assert c.start == 1
    assert c.end == 6
    assert type(c.s) == str
    assert c.s == 'print'


def test_keyword():
    tokens = make_asciis(r' throw ')
    k = Keyword(tokens[1:6])
    assert k.tag == 'throw'
    assert k.start == 1
    assert k.end == 6


def test_string():
    # Test ASCII characters and escapes.
    tokens = make_asciis(r'print("Hello, world!\n")')
    start = len(r'print(')
    end = start + len(r'"Hello, world!\n"')
    e = Escape(tokens[end-3:end-1], 10)
    s = String(
        tokens[start],
        tokens[start+1: end-3] + [e],
        tokens[end-1],
    )
    assert s.tag == 'string'
    assert s.start == start
    assert s.end == end
    assert s.open_quote is tokens[start]
    assert type(s.parts) is list
    assert len(s.parts) == 1
    assert s.parts[0] == u'Hello, world!\n'
    assert s.close_quote is tokens[end-1]
    # Test unicode characters and errors.
    tokens = [
        ASCII(0, '"'),
        NonASCII(1, 4, u'\N{snowman}'),
        ASCII(4, '\\'),
        ASCII(5, '@'),
        ASCII(6, '"'),
    ]
    tokens[2:4] = [Error('Bad escape', False, tokens[2:4])]
    s = String(
        tokens[0],
        tokens[1:3],
        tokens[3],
    )
    assert s.tag == 'string'
    assert s.start == 0
    assert s.end == 7
    assert s.open_quote is tokens[0]
    assert type(s.parts) is list
    assert len(s.parts) == 2
    assert s.parts[0] == u'\N{snowman}'
    assert s.parts[1] == tokens[2]
    assert s.close_quote is tokens[3]


def test_character():
    # Test ASCII.
    tokens = [
        ASCII(0, "'"),
        ASCII(1, '@'),
        ASCII(2, "'"),
    ]
    c = Character(tokens[0], tokens[1], tokens[2])
    assert c.tag == 'character'
    assert c.start == 0
    assert c.end == 3
    assert c.open_quote is tokens[0]
    assert type(c.c) is unicode
    assert c.c == u'@'
    assert c.close_quote is tokens[2]
    # Test NonASCII.
    tokens = [
        ASCII(0, "'"),
        NonASCII(1, 4, u'\N{snowman}'),
        ASCII(4, "'"),
    ]
    c = Character(tokens[0], tokens[1], tokens[2])
    assert c.tag == 'character'
    assert c.start == 0
    assert c.end == 5
    assert c.open_quote is tokens[0]
    assert type(c.c) is unicode
    assert c.c == u'\N{snowman}'
    assert c.close_quote is tokens[2]
    # Test Escape
    tokens = [
        ASCII(0, "'"),
        ASCII(1, "\\"),
        ASCII(2, "n"),
        ASCII(3, "'"),
    ]
    tokens[1:3] = [Escape(tokens[1:3], 10)]
    c = Character(tokens[0], tokens[1], tokens[2])
    assert c.tag == 'character'
    assert c.start == 0
    assert c.end == 4
    assert c.open_quote is tokens[0]
    assert type(c.c) is unicode
    assert c.c == u'\n'
    assert c.close_quote is tokens[2]


def test_bracket():
    tokens = make_asciis('p(x+1);')
    ob = Concat('open', [tokens[1]])
    cb = Concat('close', [tokens[5]])
    b = Bracket('()', ob, tokens[2:5], cb)
    assert b.tag == '()'
    assert b.start == 1
    assert b.end == 6
    assert type(b.contents) is list
    assert len(b.contents) == 3
    assert b.contents[0] is tokens[2]
    assert b.contents[1] is tokens[3]
    assert b.contents[2] is tokens[4]


def test_function():
    # Test with all optional parts omitted.
    tokens = make_asciis('fn(x:t)')
    k = Keyword(tokens[0:2])
    ob = Concat('open', [tokens[2]])
    cb = Concat('close', [tokens[6]])
    b = Bracket('()', ob, tokens[3:6], cb)
    f = Function(k, None, b, None, None, None)
    assert f.tag == 'function'
    assert f.start == 0
    assert f.end == 7
    assert f.name is None
    assert f.params is b
    assert f.return_type is None
    assert f.body is None
    # Test with all optional parts present.
    tokens = make_asciis('fn p(x:t):u {x}')
    k = Keyword(tokens[0:2])
    i1 = Concat('identifier', tokens[3:4])
    ob1 = Concat('open', [tokens[4]])
    cb1 = Concat('close', [tokens[8]])
    b1 = Bracket('()', ob1, tokens[5:8], cb1)
    i2 = Concat('identifier', tokens[10:11])
    ob2 = Concat('open', [tokens[12]])
    cb2 = Concat('close', [tokens[14]])
    b2 = Bracket('{}', ob2, tokens[13:14], cb2)
    f = Function(k, i1, b1, tokens[9], i2, b2)
    assert f.tag == 'function'
    assert f.start == 0
    assert f.end == 15
    assert f.name is i1
    assert f.params is b1
    assert f.return_type is i2
    assert f.body is b2


def test_module():
    # Test with all optional parts omitted.
    tokens = make_asciis('module')
    k = Keyword(tokens[0:6])
    m = Module(k, None, None)
    assert m.tag == 'module'
    assert m.start == 0
    assert m.end == 6
    assert m.keyword is k
    assert m.name is None
    assert m.body is None
    # Test with all optional parts present.
    tokens = make_asciis('module m {1}')
    k = Keyword(tokens[0:6])
    i = Concat('identifier', tokens[7:8])
    ob = Concat('open', [tokens[9]])
    cb = Concat('close', [tokens[11]])
    b = Bracket('{}', ob, tokens[10:11], cb)
    m = Module(k, i, b)
    assert m.tag == 'module'
    assert m.start == 0
    assert m.end == 12
    assert m.keyword is k
    assert m.name is i
    assert m.body is b


def test_type_struct():
    # Test with all optional parts omitted.
    tokens = make_asciis('struct()')
    k = Keyword(tokens[0:6])
    ob = Concat('open', [tokens[6]])
    cb = Concat('close', [tokens[7]])
    b = Bracket('()', ob, tokens[7:7], cb)
    s = TypeStruct(k, None, b)
    assert s.tag == 'type_struct'
    assert s.start == 0
    assert s.end == 8
    assert s.keyword is k
    assert s.name is None
    assert s.params is b
    # Test with all optional parts present.
    tokens = make_asciis('struct x()')
    k = Keyword(tokens[0:6])
    i = Concat('identifier', tokens[7:8])
    ob = Concat('open', [tokens[8]])
    cb = Concat('close', [tokens[9]])
    b = Bracket('()', ob, tokens[9:9], cb)
    s = TypeStruct(k, i, b)
    assert s.tag == 'type_struct'
    assert s.start == 0
    assert s.end == 10
    assert s.keyword is k
    assert s.name is i
    assert s.params is b


def test_type_union():
    # Test with all optional parts omitted.
    tokens = make_asciis('union {}')
    k = Keyword(tokens[0:5])
    ob = Concat('open', [tokens[6]])
    cb = Concat('close', [tokens[7]])
    b = Bracket('{}', ob, tokens[7:7], cb)
    s = TypeUnion(k, None, b)
    assert s.tag == 'type_union'
    assert s.start == 0
    assert s.end == 8
    assert s.keyword is k
    assert s.name is None
    assert s.cases is b
    # Test with all optional parts present.
    tokens = make_asciis('union x {}')
    k = Keyword(tokens[0:5])
    i = Concat('identifier', tokens[6:7])
    ob = Concat('open', [tokens[8]])
    cb = Concat('close', [tokens[9]])
    b = Bracket('{}', ob, tokens[9:9], cb)
    s = TypeUnion(k, i, b)
    assert s.tag == 'type_union'
    assert s.start == 0
    assert s.end == 10
    assert s.keyword is k
    assert s.name is i
    assert s.cases is b


def test_field():
    tokens = make_asciis('.x')
    i = Concat('identifier', tokens[1:2])
    f = Field(tokens[0], i)
    assert f.tag == 'field'
    assert f.start == 0
    assert f.end == 2
    assert f.name is i


def test_array():
    # Test with all optional parts omitted.
    tokens = make_asciis('array')
    k = Keyword(tokens[0:5])
    a = Array(k, None)
    assert a.tag == 'array'
    assert a.start == 0
    assert a.end == 5
    assert a.keyword is k
    assert a.constness is None
    # Test with all optional parts present.
    tokens = make_asciis('array const')
    k1 = Keyword(tokens[0:5])
    k2 = Keyword(tokens[6:11])
    a = Array(k1, k2)
    assert a.tag == 'array'
    assert a.start == 0
    assert a.end == 11
    assert a.keyword is k1
    assert a.constness is k2


def test_ref():
    # Test with all optional parts omitted.
    tokens = make_asciis('ref')
    k = Keyword(tokens[0:3])
    a = Ref(k, None)
    assert a.tag == 'ref'
    assert a.start == 0
    assert a.end == 3
    assert a.keyword is k
    assert a.constness is None
    # Test with all optional parts present.
    tokens = make_asciis('ref const')
    k1 = Keyword(tokens[0:3])
    k2 = Keyword(tokens[4:9])
    a = Ref(k1, k2)
    assert a.tag == 'ref'
    assert a.start == 0
    assert a.end == 9
    assert a.keyword is k1
    assert a.constness is k2


def test_unary_expr_infix():
    tokens = make_asciis('-x.y*')
    o1 = Concat('un_bin_op', tokens[0:1])
    i1 = Concat('identifier', tokens[1:2])
    i2 = Concat('identifier', tokens[3:4])
    f = Field(tokens[2], i2)
    o2 = Concat('un_bin_op', tokens[4:5])
    u = UnaryExpr([o1], i1, [f], o2)
    assert u.tag == 'unary_expr_infix'
    assert u.start == 0
    assert u.end == 5
    assert type(u.prefixes) is list
    assert len(u.prefixes) == 1
    assert u.prefixes[0] is o1
    assert u.nonfix is i1
    assert type(u.postfixes) is list
    assert len(u.postfixes) == 1
    assert u.postfixes[0] is f
    assert u.infix is o2


def test_unary_expr():
    tokens = make_asciis('-x.y')
    o = Concat('un_bin_op', tokens[0:1])
    i1 = Concat('identifier', tokens[1:2])
    i2 = Concat('identifier', tokens[3:4])
    f = Field(tokens[2], i2)
    u = UnaryExpr([o], i1, [f])
    assert u.tag == 'unary_expr'
    assert u.start == 0
    assert u.end == 4
    assert type(u.prefixes) is list
    assert len(u.prefixes) == 1
    assert u.prefixes[0] is o
    assert u.nonfix is i1
    assert type(u.postfixes) is list
    assert len(u.postfixes) == 1
    assert u.postfixes[0] is f


def test_expr():
    tokens = make_asciis('-1<<x.y')
    o1 = Concat('un_bin_op', tokens[0:1])
    o2 = Concat('bin_op', tokens[2:4])
    i1 = Concat('integer', tokens[1:2])
    i2 = Concat('identifier', tokens[4:5])
    i3 = Concat('identifier', tokens[6:7])
    f = Field(tokens[5], i3)
    assert f.tag == 'field'
    assert f.start == 5
    assert f.end == 7
    e1 = Expr(None, i1, None)
    assert e1.tag == 'expr'
    assert e1.start == 1
    assert e1.end == 2
    e2 = Expr(None, o1, e1)
    assert e2.tag == 'expr'
    assert e2.start == 0
    assert e2.end == 2
    e3 = Expr(None, i2, None)
    assert e3.tag == 'expr'
    assert e3.start == 4
    assert e3.end == 5
    e4 = Expr(e3, f, None)
    assert e4.tag == 'expr'
    assert e4.start == 4
    assert e4.end == 7
    e5 = Expr(e2, o2, e4)
    assert e5.tag == 'expr'
    assert e5.start == 0
    assert e5.end == 7
    assert e5.left is e2
    assert e5.op is o2
    assert e5.right is e4


def test_expr_list():
    tokens = make_asciis('x, y,')
    i1 = Concat('identifier', tokens[0:1])
    i2 = Concat('identifier', tokens[3:4])
    e1 = Expr(None, i1, None)
    e2 = Expr(None, i2, None)
    l = ExprList([e1, e2], [tokens[1], tokens[4]])
    assert l.tag == 'expr_list'
    assert l.start == 0
    assert l.end == 5
    assert type(l.exprs) is list
    assert len(l.exprs) == 2
    assert l.exprs[0] is e1
    assert l.exprs[1] is e2
    assert type(l.commas) is list
    assert len(l.commas) == 2
    assert l.commas[0] is tokens[1]
    assert l.commas[1] is tokens[4]


def test_assign():
    tokens = make_asciis('x += 1')
    i = Concat('identifier', tokens[0:1])
    o = Concat('assign_op', tokens[2:4])
    n = Concat('integer', tokens[5:6])
    e1 = Expr(None, i, None)
    e2 = Expr(None, n, None)
    a = Assign(e1, o, e2)
    assert a.tag == 'assign'
    assert a.start == 0
    assert a.end == 6
    assert a.lhs is e1
    assert a.op is o
    assert a.rhs is e2


def test_case():
    tokens = make_asciis('case FOO {}')
    k = Keyword(tokens[0:4])
    t = Concat('tag', tokens[5:8])
    ob = Concat('open', [tokens[9]])
    cb = Concat('close', [tokens[10]])
    b = Bracket('{}', ob, [], cb)
    u = Union(t, None)
    e = Expr(None, u, None)
    c = Case(k, e, b)
    assert c.tag == 'case'
    assert c.start == 0
    assert c.end == 11
    assert c.pattern is e
    assert c.body is b


def test_decl():
    tokens = make_asciis('var x = 1')
    k = Keyword(tokens[0:3])
    i = Concat('identifier', tokens[4:5])
    o = Concat('assign_op', tokens[6:7])
    n = Concat('integer', tokens[8:9])
    e1 = Expr(None, i, None)
    e2 = Expr(None, n, None)
    d = Decl(k, e1, o, e2)
    assert d.tag == 'decl'
    assert d.start == 0
    assert d.end == 9
    assert d.decl is k
    assert d.lhs is e1
    assert d.op is o
    assert d.rhs is e2


def test_else():
    tokens = make_asciis('else {}')
    k = Keyword(tokens[0:4])
    ob = Concat('open', [tokens[5]])
    cb = Concat('close', [tokens[6]])
    b = Bracket('{}', ob, [], cb)
    e = Else(k, b)
    assert e.tag == 'else'
    assert e.start == 0
    assert e.end == 7
    assert e.keyword is k
    assert e.body is b


def test_for():
    tokens = make_asciis('for x in 1 {} else {}')
    k1 = Keyword(tokens[0:3])
    i = Concat('identifier', tokens[4:5])
    k2 = Keyword(tokens[6:8])
    n = Concat('integer', tokens[9:10])
    ob1 = Concat('open', [tokens[11]])
    cb1 = Concat('close', [tokens[12]])
    b1 = Bracket('{}', ob1, [], cb1)
    k3 = Keyword(tokens[14:18])
    ob2 = Concat('open', [tokens[19]])
    cb2 = Concat('close', [tokens[20]])
    b2 = Bracket('{}', ob2, [], cb2)
    e1 = Expr(None, i, None)
    e2 = Expr(None, n, None)
    else_ = Else(k3, b2)
    f = For(k1, e1, k2, e2, b1, else_)
    assert f.tag == 'for'
    assert f.start == 0
    assert f.end == 21
    assert f.for_keyword is k1
    assert f.variable is e1
    assert f.in_keyword is k2
    assert f.sequence is e2
    assert f.body is b1
    assert f.else_ is else_


def test_fulfil():
    tokens = make_asciis('fulfil x = 1')
    k = Keyword(tokens[0:6])
    i = Concat('identifier', tokens[7:8])
    o = Concat('assign_op', tokens[9:10])
    n = Concat('integer', tokens[11:12])
    e1 = Expr(None, i, None)
    e2 = Expr(None, n, None)
    d = Fulfil(k, e1, o, e2)
    assert d.tag == 'fulfil'
    assert d.start == 0
    assert d.end == 12
    assert d.lhs is e1
    assert d.op is o
    assert d.rhs is e2


def test_if():
    tokens = make_asciis('if x {} else if y {}')
    k1 = Keyword(tokens[0:2])
    i1 = Concat('identifier', tokens[3:4])
    ob1 = Concat('open', [tokens[5]])
    cb1 = Concat('close', [tokens[6]])
    b1 = Bracket('{}', ob1, [], cb1)
    k2 = Keyword(tokens[8:12])
    k3 = Keyword(tokens[13:15])
    i2 = Concat('identifier', tokens[16:17])
    ob2 = Concat('open', [tokens[18]])
    cb2 = Concat('close', [tokens[19]])
    b2 = Bracket('{}', ob2, [], cb2)
    e1 = Expr(None, i1, None)
    e2 = Expr(None, i2, None)
    if2 = If(k3, e2, b2, None)
    assert if2.tag == 'if'
    assert if2.start == 13
    assert if2.end == 20
    assert if2.keyword is k3
    assert if2.condition is e2
    assert if2.then is b2
    assert if2.else_ is None
    else_if = ElseIf(k2, if2)
    assert else_if.tag == 'else_if'
    assert else_if.start == 8
    assert else_if.end == 20
    assert else_if.keyword is k2
    assert else_if.if_ is if2
    if1 = If(k1, e1, b1, else_if)
    assert if1.tag == 'if'
    assert if1.start == 0
    assert if1.end == 20
    assert if1.keyword == k1
    assert if1.condition is e1
    assert if1.then is b1
    assert if1.else_ is else_if


def test_return():
    tokens = make_asciis('return x')
    k = Keyword(tokens[0:6])
    i = Concat('identifier', tokens[7:8])
    e = Expr(None, i, None)
    r = Return(k, e)
    assert r.tag == 'return'
    assert r.start == 0
    assert r.end == 8
    assert r.keyword is k
    assert r.expr is e


def test_switch():
    tokens = make_asciis('switch x {}')
    k = Keyword(tokens[0:6])
    i = Concat('identifier', tokens[7:8])
    ob = Concat('open', [tokens[9]])
    cb = Concat('close', [tokens[10]])
    b = Bracket('{}', ob, [], cb)
    e = Expr(None, i, None)
    s = Switch(k, e, b)
    assert s.tag == 'switch'
    assert s.start == 0
    assert s.end == 11
    assert s.condition is e
    assert s.cases is b


def test_throw():
    tokens = make_asciis('throw x')
    k = Keyword(tokens[0:5])
    i = Concat('identifier', tokens[6:7])
    e = Expr(None, i, None)
    t = Throw(k, e)
    assert t.tag == 'throw'
    assert t.start == 0
    assert t.end == 7
    assert t.keyword is k
    assert t.expr is e


def test_while():
    tokens = make_asciis('while x {} else {}')
    k1 = Keyword(tokens[0:5])
    i = Concat('identifier', tokens[6:7])
    ob1 = Concat('open', [tokens[8]])
    cb1 = Concat('close', [tokens[9]])
    b1 = Bracket('{}', ob1, [], cb1)
    k2 = Keyword(tokens[11:15])
    ob2 = Concat('open', [tokens[16]])
    cb2 = Concat('close', [tokens[17]])
    b2 = Bracket('{}', ob2, [], cb2)
    e = Expr(None, i, None)
    else_ = Else(k2, b2)
    w = While(k1, e, b1, else_)
    assert w.tag == 'while'
    assert w.start == 0
    assert w.end == 18
    assert w.keyword is k1
    assert w.condition is e
    assert w.body is b1
    assert w.else_ is else_
