from token import Token, matches, ASCII
from lexer import *
from brackets import BracketMatcher, bracket_tags
from atom import atom, atom_tags
from unary_expr import unary_expr, unary_expr_tags
from grammar import Parser, Buffer


allowed_tags = {'error'}.union(
    high_level_lexer_tags, keywords,
    bracket_tags,
    atom_tags,
    unary_expr_tags
)


observed_tags = set()


def check(source, expected_result):
    listener = Buffer()
    def make_parser(p):
        # Call-back used to make a Parser for each bracket nesting level.
        p = Parser(unary_expr, p)
        p = Parser(atom, p)
        return p
    p = make_parser(listener)  # top-level outside brackets
    p = BracketMatcher(p, {
        '(': make_parser,
        '[': make_parser,
        '{': make_parser,
    })
    p = Parser(high_level_lexer, p)
    p = Parser(low_level_lexer, p)
    for i, c in enumerate(source):
        p.feed(ASCII(i, c))
    p.flush()
    result = listener.contents
    for token in result:
        assert isinstance(token, Token)
        assert token.tag in allowed_tags, token
        observed_tags.add(token.tag)
    assert matches(result, expected_result), (
        "Parsed:\n"
        "    %r\n"
        "as:\n"
        "    %r\n"
        "Expected:\n"
        "    %r\n" % (source, result, expected_result)
    )


id_t = ('identifier', 't')


def test_irrelevant():
    check('@', [('error', 'Illegal character')])

    check('for', [
        ('for',),
    ])


def test_atoms():
    check('hello', [
        ('unary_expr', [], ('identifier', 'hello'), []),
    ])

    check('{ }', [
        ('{}', []),
    ])

    check('( )', [
        ('unary_expr', [], ('()', []), []),
    ])

    check('[ ]', [
        ('unary_expr', [], ('[]', []), []),
    ])

    check('fn (): t {}', [
        ('unary_expr',
            [],
            ('function', None, ('()', []), id_t, ('{}', [])),
            []
        )
    ])

    check('module {}', [
        ('unary_expr', [], ('module', None, ('{}', [])), [])
    ])

    check('struct ()', [
        ('unary_expr', [], ('type_struct', None, ('()', [])), [])
    ])

    check('union {}', [
        ('unary_expr', [], ('type_union', None, ('{}', [])), [])
    ])

    check('C', [
        ('unary_expr', [], ('union', ('tag', 'C'), None), [])
    ])

    check('C()', [
        ('unary_expr', [], ('union', ('tag', 'C'), ('()', [])), [])
    ])

    check('C(1)', [
        ('unary_expr',
            [],
            ('union',
                ('tag', 'C'),
                ('()', [('unary_expr', [], ('integer', '1'), [])])
            ),
            []
        )
    ])


def test_neighbours():
    check('hello world', [
        ('unary_expr', [], ('identifier', 'hello'), []),
        ('unary_expr', [], ('identifier', 'world'), []),
    ])


def test_interruptions():
    check('1@', [
        ('unary_expr', [], ('integer', '1'), []),
        ('error', 'Illegal character')
    ])

    check('1+@', [
        ('unary_expr_infix', [], ('integer', '1'), [], ('un_bin_op', '+')),
        ('error', 'Illegal character')
    ])


def test_infix():
    check('hello + world', [
        ('unary_expr_infix', [], ('identifier', 'hello'), [], ('un_bin_op', '+')),
        ('unary_expr', [], ('identifier', 'world'), []),
    ])

    check('hello : world', [
        ('unary_expr_infix', [], ('identifier', 'hello'), [], ('bin_op', ':')),
        ('unary_expr', [], ('identifier', 'world'), []),
    ])

    check('0==1', [
        ('unary_expr_infix', [], ('integer', '0'), [], ('bin_op', '==')),
        ('unary_expr', [], ('integer', '1'), [])
    ])


def test_postfix():
    check('hello { world }', [
        ('unary_expr', [], ('identifier', 'hello'), []),
        ('{}', [
            ('unary_expr', [], ('identifier', 'world'), [])
        ]),
    ])

    check('hello ( world )', [
        ('unary_expr', [], ('identifier', 'hello'), [
            ('()', [('unary_expr', [], ('identifier', 'world'), [])]),
        ]),
    ])

    check('hello [ world ]', [
        ('unary_expr', [], ('identifier', 'hello'), [
            ('[]', [('unary_expr', [], ('identifier', 'world'), [])]),
        ]),
    ])

    check('hello.world', [
        ('unary_expr', [], ('identifier', 'hello'), [
            ('field', ('identifier', 'world')),
        ]),
    ])

    check('goodbye.cruel.world', [
        ('unary_expr', [], ('identifier', 'goodbye'), [
            ('field', ('identifier', 'cruel')),
            ('field', ('identifier', 'world')),
        ]),
    ])

    check('goodbye(cruel.world)', [
        ('unary_expr', [], ('identifier', 'goodbye'), [
            ('()', [
                ('unary_expr', [], ('identifier', 'cruel'), [
                    ('field', ('identifier', 'world')),
                ]),
            ]),
        ]),
    ])


def test_prefix():
    check('sizeof ref defer typeof new array int', [
        # "sizeof" and "int" are not keywords, but "ref", "defer",
        # "typeof", "new" and "array" are.
        ('unary_expr', [], ('identifier', 'sizeof'), []),
        ('unary_expr', [
            ('ref', None), ('defer',), ('typeof',),
            ('new',), ('array', None)
        ], ('identifier', 'int'), [])
    ])

    check('-', [('error', 'Expected an expression')])



def test_silly():
    check(r'~"foo".x^', [(
        'unary_expr_infix',
        [('un_op', '~')],
        ('string', [u'foo']),
        [('field', ('identifier', 'x'))],
        ('bin_op', '^')
    )])


def test_coverage():
    for t0 in unary_expr_tags:
        assert t0 in observed_tags, t0
