class Token(object):
    '''
    Represents some approximation of the parse tree of a contiguous string of
    bytes of source code.

     - tag - a str that can be tested by the parser to discriminate
       syntactic elements.
     - start - the byte offset in the source code of the start of the
       string (inclusive).
     - end - the byte offset in the source code of the end of the string
       (exclusive).
    '''
    def __init__(self, tag, start, end):
        assert type(tag) is str
        assert type(start) in (int, long)
        assert type(end) in (int, long)
        self.tag = tag
        self.start = start
        self.end = end

    def matches(self, tag, *pattern):
        '''
        Returns `True` if `self.tag` equals `tag` and if this Token matches
        `pattern` (in some sense to be defined by the subclass).
        '''
        raise NotImplementedError("abstract method")


def matches(token, pattern):
    '''
    If `token` is a Token, returns `token.matches(*pattern)`. If it's a list,
    recursively matches the contents. Otherwise returns `token == pattern`.

     - token - a Token or a value that can be tested using "==".
    '''
    if type(token) is list:
        return (
            type(pattern) is list and
            len(token) == len(pattern) and
            all(matches(t, p) for t, p in zip(token, pattern))
        )
    elif isinstance(token, Token):
        return type(pattern) is tuple and token.matches(*pattern)
    else:
        return token == pattern


class Compound(Token):
    '''
    A Token that is composed of smaller Tokens.
    '''
    def __init__(self, tag, children):
        '''
         - children - a list of Tokens and/or `None`s, with at least one
           Token.
        '''
        start = min(child.start for child in children if child is not None)
        end = max(child.end for child in children if child is not None)
        super(Compound, self).__init__(tag, start, end)


class Error(Compound):
    '''
    A Compound that represents a sequence of characters that could not be
    parsed.

    Public fields:
     - message - a human-readable error message.
     - is_after - `True` if the error was something missing after
       `children`, or `False` if the error was within `children`.
    '''
    def __init__(self, message, is_after, children):
        '''
         - children - a list of Tokens.
        '''
        assert type(message) is str
        super(Error, self).__init__('error', children)
        self.message = message
        self.is_after = is_after

    def matches(self, tag, message):
        return self.tag == tag and self.message == message

    def __repr__(self):
        return 'Error(%r)' % self.message


class ASCII(Token):
    '''
    A Token that represents a single ASCII character. Different characters can
    be discriminated by the parser.
    '''
    def __init__(self, start, c):
        '''
         - start - the byte offset in the source code of the start of the
           string (inclusive). `end` is `start+1`.
         - c - the character, a str of length 1.
        '''
        assert type(c) is str
        assert len(c) == 1
        end = start + 1
        super(ASCII, self).__init__(c, start, end)

    def matches(self, tag):
        return self.tag == tag

    def __repr__(self):
        return 'ASCII(%r)' % self.tag


def is_ascii(token):
    # Equivalent to but more efficient than `token.tag in ' !"#$%...'`
    return type(token) is ASCII


class NonASCII(Token):
    '''
    A Token that represents a single non-ASCII character. Different characters
    cannot be discriminated by the parser.

    Public fields:
     - c - the character, a unicode of length 1.
    '''
    def __init__(self, start, end, c):
        '''
         - start - the byte offset in the source code of the start of the
           string (inclusive).
         - end - the byte offset in the source code of the end of the string
           (exclusive).
        '''
        assert type(c) is unicode
        assert len(c) == 1
        super(NonASCII, self).__init__('non_ascii', start, end)
        self.c = c

    def matches(self, tag, c):
        return self.tag == tag and self.c == c

    def __repr__(self):
        return 'NonASCII(%r)' % self.c


class Escape(Compound):
    '''
    A Token that represents an escape sequence.

    Public fields:
     - code_point - the unicode code point represented by the escape
       sequence.
    '''
    def __init__(self, asciis, code_point):
        '''
         - asciis - a list of ASCIIs representing the entire escape sequence.
        '''
        assert type(code_point) is int
        super(Escape, self).__init__('escape', asciis)
        self.code_point = code_point

    def matches(self, tag, code_point):
        return self.tag == tag and self.code_point == code_point

    def __repr__(self):
        return 'Escape(%r)' % self.code_point


class Concat(Compound):
    '''
    A Token that represents a contiguous string of ASCII characters.

    Public fields:
     - s - the characters as a str.
    '''
    def __init__(self, tag, asciis):
        '''
         - asciis - a list of ASCIIs.
        '''
        assert all(is_ascii(x) for x in asciis)
        super(Concat, self).__init__(tag, asciis)
        self.s = ''.join(a.tag for a in asciis)

    def matches(self, tag, s):
        return self.tag == tag and self.s == s

    def __repr__(self):
        return 'Concat(%r, %r)' % (self.tag, self.s)


class Keyword(Compound):
    '''
    A Token that represents a contiguous string of ASCII characters and
    whose `tag` is equal to that string of characters.
    '''
    def __init__(self, asciis):
        '''
         - asciis - a list of ASCIIs.
        '''
        assert all(is_ascii(x) for x in asciis)
        tag = ''.join(a.tag for a in asciis)
        super(Keyword, self).__init__(tag, asciis)

    def matches(self, tag):
        return self.tag == tag

    def __repr__(self):
        return 'Keyword(%r)' % self.tag


class String(Compound):
    '''
    A Token that represents a string literal.

    Public fields:
     - open_quote - the ASCII representing the opening quotation mark.
     - parts - a list of unicodes and Errors. If there are no errors there
       will be at most one unicode.
     - close_quote - the ASCII representing the closing quotation mark.
    '''
    def __init__(self, open_quote, contents, close_quote):
        '''
         - contents - a list of ASCIIs, NonASCIIs, Escapes and Errors.
        '''
        super(String, self).__init__(
            'string', [open_quote] + contents + [close_quote])
        assert type(open_quote) is ASCII and open_quote.tag == '"'
        assert type(close_quote) is ASCII and close_quote.tag == '"'
        self.open_quote = open_quote
        self.close_quote = close_quote
        self.parts = []
        cs = [] # List of str and unicode (no Errors).
        for c in contents:
            if is_ascii(c):
                cs.append(c.tag)
            elif c.tag == 'non_ascii':
                cs.append(c.c)
            elif c.tag == 'escape':
                cs.append(unichr(c.code_point))
            elif isinstance(c, Concat):
                # TODO: Rewrite the lexer so that Concats are only produced
                # by the high-level lexer, then remove this case.
                cs.append(c.s)
            elif isinstance(c, Keyword):
                # TODO: Rewrite the lexer so that Keywords are only produced
                # by the high-level lexer, then remove this case.
                cs.append(c.tag)
            elif isinstance(c, Error):
                # It's an error.
                if cs:
                    self.parts.append(u''.join(cs))
                cs = []
                self.parts.append(c)
            else:
                assert False, c
        if cs:
            self.parts.append(u''.join(cs))

    def matches(self, tag, parts):
        '''
        Tests that:
         - `self.tag` equals `tag`
         - the unicodes in `self.parts` equal the corresponding elements of
           `parts`.
         - the Errors in `self.parts` match the corresponding elements of
           `parts`, which are `(tag, message)` pairs.
        '''
        return self.tag == tag and matches(self.parts, parts)

    def __repr__(self):
        return 'String(%r)' % self.parts


class Character(Compound):
    '''
    A Token that represents a character literal.

    Public fields:
     - open_quote - the ASCII representing the opening quotation mark.
     - c - the character, a unicode of length 1.
     - close_quote - the ASCII representing the closing quotation mark.
    '''
    def __init__(self, open_quote, token, close_quote):
        '''
         - token - an ASCII, a NonASCII, or an Escape.
        '''
        super(Character, self).__init__(
            'character', [open_quote, token, close_quote])
        assert type(open_quote) is ASCII and open_quote.tag == "'"
        assert type(close_quote) is ASCII and close_quote.tag == "'"
        self.open_quote = open_quote
        self.close_quote = close_quote
        if is_ascii(token):
            self.c = unicode(token.tag)
        elif token.tag == 'non_ascii':
            self.c = token.c
        elif token.tag == 'escape':
            self.c = unichr(token.code_point)
        elif isinstance(token, Concat):
            # TODO: Rewrite the lexer so that Concats are only produced
            # by the high-level lexer, then remove this case.
            self.c = unicode(token.s)
        else:
            assert False, token
        assert len(self.c) == 1
        assert type(self.c) is unicode

    def matches(self, tag, c):
        return self.tag == tag and self.c == c

    def __repr__(self):
        return 'Character(%r)' % self.c


class Bracket(Compound):
    '''
    A Token that represents some matched brackets and their contents.

    Public fields:
     - open_bracket - a Concat with tag 'open'.
     - contents - a list of Tokens.
     - close_bracket - a Concat with tag 'close', or `None` if it's missing.
    '''
    def __init__(self, tag, open_bracket, contents, close_bracket):
        '''
         - contents - a list of Tokens.
        '''
        assert open_bracket.tag == 'open'
        assert close_bracket is None or close_bracket.tag == 'close'
        super(Bracket, self).__init__(
            tag, [open_bracket] + contents + [close_bracket])
        self.open_bracket = open_bracket
        self.contents = contents
        self.close_bracket = close_bracket

    def matches(self, tag, contents):
        '''
        Tests that `self.tag` equals `tag` and each element of `self.contents`
        matches the corresponding element of `contents`.

         - contents - a list of (tag, ...) tuples.
        '''
        return self.tag == tag and matches(self.contents, contents)

    def __repr__(self):
        return 'Bracket(%r, %r)' % (self.tag, self.contents)


class Function(Compound):
    '''
    A Token that represents a function definition or function type.

    Public fields:
     - keyword - the "fn" Keyword.
     - name - the 'identifier' Token that represents the name of the
       function, or `None` if omitted.
     - params - the '()' Token that represents the function parameters.
     - colon - the ':' ASCII, or `None` if omitted.
     - return_type - the 'identifier' or '()' Token that represents the
       explicit return type, or `None` if omitted.
     - body - the '{}' Token that represents the function body, or `None`
       if omitted.
    '''
    def __init__(self, keyword, name, params, colon, return_type, body):
        assert keyword.tag == 'fn'
        assert name is None or name.tag == 'identifier'
        assert params.tag == '()'
        assert colon is None or colon.tag == ':'
        assert return_type is None or return_type.tag in (
            'identifier', 'module', '()')
        assert (colon is None) == (return_type is None)
        assert body is None or body.tag == '{}'
        super(Function, self).__init__(
            'function', [keyword, name, params, colon, return_type, body])
        self.keyword = keyword
        self.name = name
        self.params = params
        self.colon = colon
        self.return_type = return_type
        self.body = body

    def matches(self, tag, name, params, return_type, body):
        return (
            self.tag == tag and
            matches(self.name, name) and
            matches(self.params, params) and
            matches(self.return_type, return_type) and
            matches(self.body, body)
        )

    def __repr__(self):
        return 'Function(%r, %r, %r, %r)' % (
            self.name, self.params, self.return_type, self.body)


class Module(Compound):
    '''
    A Token that represents a module definition or module type.

    Public fields:
     - keyword - the "module" Keyword.
     - name - the 'identifier' Token that represents the name of the
       module, or `None` if omitted.
     - body - the '{}' Token that represents the module body, or `None`
       if omitted.
    '''
    def __init__(self, keyword, name, body):
        assert keyword.tag == 'module'
        assert name is None or name.tag == 'identifier'
        assert body is None or body.tag == '{}'
        assert name is None or body is not None
        super(Module, self).__init__('module', [keyword, name, body])
        self.keyword = keyword
        self.name = name
        self.body = body

    def matches(self, tag, name, body):
        return (
            self.tag == tag and
            matches(self.name, name) and
            matches(self.body, body)
        )

    def __repr__(self):
        return 'Module(%r, %r)' % (self.name, self.body)


class TypeStruct(Compound):
    '''
    A Token that represents a struct type definition, e.g. "struct foo(...)".

    Public fields:
     - keyword - the "struct" Keyword.
     - name - the 'identifier' Token that represents the name of the
       struct type, or `None` if omitted.
     - params - the '()' Token that represents the struct fields.
    '''
    def __init__(self, keyword, name, params):
        assert keyword.tag == 'struct'
        assert name is None or name.tag == 'identifier'
        assert params.tag == '()'
        super(TypeStruct, self).__init__(
            'type_struct', [keyword, name, params])
        self.keyword = keyword
        self.name = name
        self.params = params

    def matches(self, tag, name, params):
        return (
            self.tag == tag and
            matches(self.name, name) and
            matches(self.params, params)
        )

    def __repr__(self):
        return 'TypeStruct(%r, %r)' % (self.name, self.params)


class TypeUnion(Compound):
    '''
    A Token that represents a union type definition, e.g. "union foo {...}".

    Public fields:
     - keyword - the "union" Keyword.
     - name - the 'identifier' Token that represents the name of the
       union type, or `None` if omitted.
     - cases - the '()' Token that represents the union cases.
    '''
    def __init__(self, keyword, name, cases):
        assert keyword.tag == 'union'
        assert name is None or name.tag == 'identifier'
        assert cases.tag == '{}'
        super(TypeUnion, self).__init__('type_union', [keyword, name, cases])
        self.keyword = keyword
        self.name = name
        self.cases = cases

    def matches(self, tag, name, cases):
        return (
            self.tag == tag and
            matches(self.name, name) and
            matches(self.cases, cases)
        )

    def __repr__(self):
        return 'TypeUnion(%r, %r)' % (self.name, self.cases)


class Union(Compound):
    '''
    A Token that represents a union value, e.g. "NIL" or "CONS(...)".

    Public fields:
     - name - the 'tag' Token that represents the case name.
     - payload - the '()' Token that represents the payload values, or
       `None` if the payload is omitted.
    '''
    def __init__(self, name, payload):
        assert name.tag == 'tag'
        assert payload is None or payload.tag == '()'
        super(Union, self).__init__('union', [name, payload])
        self.name = name
        self.payload = payload

    def matches(self, tag, name, payload):
        return (
            self.tag == tag and
            matches(self.name, name) and
            matches(self.payload, payload)
        )

    def __repr__(self):
        return 'Union(%r, %r)' % (self.name, self.payload)


class Field(Compound):
    '''
    A Token that represents a field access, e.g. ".foo".

    Public fields:
     - dot - the '.' Token (an ASCII).
     - name - the 'identifier' Token that represents the field name.
    '''
    def __init__(self, dot, name):
        assert dot.tag == '.'
        assert name.tag == 'identifier'
        super(Field, self).__init__('field', [dot, name])
        self.dot = dot
        self.name = name

    def matches(self, tag, name):
        return (
            self.tag == tag and
            matches(self.name, name)
        )

    def __repr__(self):
        return 'Field(%r)' % (self.name,)


class Array(Compound):
    '''
    A Token that represents an array type constructor, e.g. "array const".

    Public fields:
     - keyword - the "array" Keyword.
     - constness - the 'const' or 'var' keyword, or `None` if omitted.
    '''
    def __init__(self, keyword, constness):
        super(Array, self).__init__('array', [keyword, constness])
        self.keyword = keyword
        self.constness = constness

    def matches(self, tag, constness):
        return (
            self.tag == tag and
            matches(self.constness, constness)
        )

    def __repr__(self):
        return 'Array(%r)' % (self.constness,)


class Ref(Compound):
    '''
    A Token that represents a reference type constructor, e.g. "ref const".

    Public fields:
     - keyword - the "ref" Keyword.
     - constness - the 'const' or 'var' keyword, or `None` if omitted.
    '''
    def __init__(self, keyword, constness):
        super(Ref, self).__init__('ref', [keyword, constness])
        self.keyword = keyword
        self.constness = constness

    def matches(self, tag, constness):
        return (
            self.tag == tag and
            matches(self.constness, constness)
        )

    def __repr__(self):
        return 'Ref(%r)' % (self.constness,)


PREFIX_TAGS = {
    'un_op', 'un_bin_op',
    'array', 'box', 'defer', 'new', 'ref', 'typeof',
}
NONFIX_TAGS = {
    '()', '[]',
    'integer', 'hex_integer', 'float', 'string', 'character', 'identifier',
    'function', 'module', 'type_struct', 'type_union', 'union',
}
POSTFIX_TAGS = {
    '()', '[]', 'field',
}
INFIX_TAGS = {
    'bin_op', 'un_bin_op',
}


class UnaryExpr(Compound):
    '''
    A Token that represents a unary expression optionally followed by an infix
    operator. The tag is 'unary_expr' if the infix operator is absent, and
    'unary_expr_infix' if it is present.

    '''
    def __init__(self, prefixes, nonfix, postfixes, infix=None):
        '''
         - prefixes - a list of Tokens representing prefix operators.
         - nonfix - a Token representing an expression without any operators.
         - postfixes - a list of Tokens representing postfix operators.
         - infix - a Token representing an infix operator, or `None`.
        '''
        assert all(x.tag in PREFIX_TAGS for x in prefixes)
        assert nonfix.tag in NONFIX_TAGS
        assert all(x.tag in POSTFIX_TAGS for x in postfixes)
        assert infix is None or infix.tag in INFIX_TAGS
        if infix is None:
            super(UnaryExpr, self).__init__(
                'unary_expr', prefixes + [nonfix] + postfixes)
        else:
            super(UnaryExpr, self).__init__(
                'unary_expr_infix', prefixes + [nonfix] + postfixes + [infix])
        self.prefixes = prefixes
        self.nonfix = nonfix
        self.postfixes = postfixes
        self.infix = infix

    def matches(self, tag, prefixes, nonfix, postfixes, infix=None):
        return (
            self.tag == tag and
            matches(self.prefixes, prefixes) and
            matches(self.nonfix, nonfix) and
            matches(self.postfixes, postfixes) and
            matches(self.infix, infix)
        )

    def __repr__(self):
        return 'UnaryExpr(%r, %r, %r, %r)' % (
            self.prefixes, self.nonfix, self.postfixes, self.infix)


class Expr(Compound):
    '''
    A Token that represents an expression.
    '''
    def __init__(self, left, op, right):
        '''
         - left - an 'expr' Token or `None`.
         - op - a Token representing an operator.
         - right - an 'expr' Token or `None`.

        If `op` is nonfix, then both `left` and `right` must be `None`.
        If `op` is prefix, then `left` must be `None` and `right` must not.
        If `op` is postfix, then `right` must be `None` and `left` must not.
        If `op` is infix, then both `left` and `right` must not be `None`.
        '''
        if left is None:
            if right is None:
                assert op.tag in NONFIX_TAGS
            else:
                assert op.tag in PREFIX_TAGS
        else:
            if right is None:
                assert op.tag in POSTFIX_TAGS
            else:
                assert op.tag in INFIX_TAGS
        super(Expr, self).__init__('expr', [left, op, right])
        self.left = left
        self.op = op
        self.right = right

    def matches(self, tag, left, op, right):
        return (
            self.tag == tag and
            matches(self.left, left) and
            matches(self.op, op) and
            matches(self.right, right)
        )

    def __repr__(self):
        return 'Expr(%r, %r, %r)' % (self.left, self.op, self.right)


class ExprList(Compound):
    '''
    A Token that represents a comma-separated list of 'expr' Tokens.
    '''
    def __init__(self, exprs, commas):
        '''
         - exprs - a list of 'expr' Tokens.
         - commas - a list of the ',' ASCIIs.
        '''
        assert all(x.tag == 'expr' for x in exprs)
        assert all(x.tag == ',' for x in commas)
        super(ExprList, self).__init__('expr_list', exprs + commas)
        self.exprs = exprs
        self.commas = commas

    def matches(self, tag, exprs, commas):
        return (
            self.tag == tag and
            matches(self.exprs, exprs) and
            matches(self.commas, commas)
        )

    def __repr__(self):
        return 'ExprList(%r, %r)' % (self.exprs, self.commas)


class Assign(Compound):
    '''
    A Token that represents an assignment statement, e.g. "x = 1".

    There is further checking in `astbuilder` to ensure that the left-hand
    expr of an 'assign' must be an lvalue.

    See also "doc/grammar/decl-cases.txt".
    '''
    # We do not support "a = b = c" syntax. Should we?

    def __init__(self, lhs, op, rhs):
        '''
         - lhs - an 'expr' Token representing the left-hand side.
         - op - an 'assign_op' Token representing the assignment operator.
           Note that the operator need not be '='; it can be e.g. '+='
           instead.
         - rhs - an 'expr' Token representing the right-hand side.
        '''
        assert lhs.tag == 'expr'
        assert op.tag == 'assign_op'
        assert rhs.tag == 'expr'
        super(Assign, self).__init__('assign', [lhs, op, rhs])
        self.lhs = lhs
        self.op = op
        self.rhs = rhs

    def matches(self, tag, lhs, op, rhs):
        return (
            self.tag == tag and
            matches(self.lhs, lhs) and
            matches(self.op, op) and
            matches(self.rhs, rhs)
        )

    def __repr__(self):
        return 'Assign(%r, %r, %r)' % (self.lhs, self.op, self.rhs)


class Case(Compound):
    '''
    A Token that represents an "case" statement, e.g. "case FOO(...) {...}".
    '''
    def __init__(self, keyword, pattern, body):
        '''
         - keyword - the "case" Keyword.
         - pattern - an 'expr' Token representing the "FOO(...)" part.
         - body - a '{}' Token representing the "{...}" part.
        '''
        assert keyword.tag == 'case'
        assert pattern.tag == 'expr'
        assert body.tag == '{}'
        super(Case, self).__init__('case', [keyword, pattern, body])
        self.keyword = keyword
        self.pattern = pattern
        self.body = body

    def matches(self, tag, pattern, body):
        return (
            self.tag == tag and
            matches(self.pattern, pattern) and
            matches(self.body, body)
        )

    def __repr__(self):
        return 'Case(%r, %r)' % (self.pattern, self.body)


class Decl(Compound):
    '''
    A Token that represents a declaration, e.g. "const x = 1".

    There is further checking in `astbuilder` to ensure that the left-hand
    expr of an 'assign' must be a binding, and that the assignment operator
    must be "=".

    See also "doc/grammar/decl-cases.txt".
    '''
    def __init__(self, decl, lhs, op, rhs):
        '''
         - decl - one of the Keywords "const" or "var".
         - lhs - an 'expr' Token representing the left-hand side.
         - op - an 'assign_op' Token representing the "=" operator.
         - rhs - an 'expr' Token representing the right-hand side.
        '''
        assert decl.tag in ('const', 'var')
        assert lhs.tag == 'expr'
        assert op.tag == 'assign_op'
        assert rhs.tag == 'expr'
        super(Decl, self).__init__('decl', [decl, lhs, op, rhs])
        self.decl = decl
        self.lhs = lhs
        self.op = op
        self.rhs = rhs

    def matches(self, tag, decl, lhs, op, rhs):
        return (
            self.tag == tag and
            matches(self.decl, decl) and
            matches(self.lhs, lhs) and
            matches(self.op, op) and
            matches(self.rhs, rhs)
        )

    def __repr__(self):
        return 'Decl(%r, %r, %r, %r)' % (
            self.decl, self.lhs, self.op, self.rhs)


class Else(Compound):
    '''
    A Token that represents "else {...}". An Else can appear inside a "switch"
    statement, or as part of a For, If or While.

    Public fields:
     - keyword - the "else" Keyword.
     - body - a '{}' Token representing the "{...}" part.
    '''
    def __init__(self, keyword, body):
        assert keyword.tag == 'else'
        assert body.tag == '{}'
        super(Else, self).__init__('else', [keyword, body])
        self.keyword = keyword
        self.body = body

    def matches(self, tag, body):
        return (
            self.tag == tag and
            matches(self.body, body)
        )

    def __repr__(self):
        return 'Else(%r)' % (self.body,)


class ElseIf(Compound):
    '''
    A Token that represents "else if ...". An ElseIf can only appear inside an
    If, and it contains another If.

    Public fields:
     - keyword - the "else" Keyword.
     - if_ - an If.
    '''
    def __init__(self, keyword, if_):
        assert keyword.tag == 'else'
        assert type(if_) is If
        super(ElseIf, self).__init__('else_if', [keyword, if_])
        self.keyword = keyword
        self.if_ = if_

    def matches(self, tag, if_):
        return (
            self.tag == tag and
            matches(self.if_, if_)
        )

    def __repr__(self):
        return 'ElseIf(%r)' % (self.if_,)


class For(Compound):
    '''
    A Token that represents a "for" statement, e.g.
    "for x in e {...} else {...}".

    The final "else" clause is optional; its absence is indicated by `None`.

    Public fields:
     - for_keyword - the "for" Keyword.
     - variable - an 'expr' Token representing the "for" variable.
     - in_keyword - the "in" Keyword.
     - sequence - an 'expr' Token representing the sequence to loop through.
     - body - a '{}' Token representing the loop body.
     - else_ - an Else representing the else body, or `None`.
    '''
    def __init__(
        self,
        for_keyword,
        variable,
        in_keyword,
        sequence,
        body,
        else_,
    ):
        assert for_keyword.tag == 'for'
        assert variable.tag == 'expr'
        assert in_keyword.tag == 'in'
        assert sequence.tag == 'expr'
        assert body.tag == '{}'
        assert else_ is None or (type(else_) is Else and else_.tag == 'else')
        super(For, self).__init__(
            'for', [
                for_keyword,
                variable,
                in_keyword,
                sequence,
                body,
                else_,
            ])
        self.for_keyword = for_keyword
        self.variable = variable
        self.in_keyword = in_keyword
        self.sequence = sequence
        self.body = body
        self.else_ = else_

    def matches(self, tag, variable, sequence, body, else_):
        return (
            self.tag == tag and
            matches(self.variable, variable) and
            matches(self.sequence, sequence) and
            matches(self.body, body) and
            matches(self.else_, else_)
        )

    def __repr__(self):
        return 'For(%r, %r, %r, %r)' % (
            self.variable, self.sequence, self.body, self.else_)


class Fulfil(Compound):
    '''
    A Token that represents a "fulfil" statement, e.g. "fulfil x = y".

    Public fields:
     - keyword - the "fulfil" keyword.
     - lhs - an 'expr' Token representing the left-hand side.
     - op - an 'assign_op' Token representing the "=" operator.
     - rhs - an 'expr' Token representing the right-hand side.
    '''
    def __init__(self, keyword, lhs, op, rhs):
        assert keyword.tag == 'fulfil'
        assert lhs.tag == 'expr'
        assert op.tag == 'assign_op'
        assert rhs.tag == 'expr'
        super(Fulfil, self).__init__('fulfil', [keyword, lhs, op, rhs])
        self.keyword = keyword
        self.lhs = lhs
        self.op = op
        self.rhs = rhs

    def matches(self, tag, lhs, op, rhs):
        return (
            self.tag == tag and
            matches(self.lhs, lhs) and
            matches(self.op, op) and
            matches(self.rhs, rhs)
        )

    def __repr__(self):
        return 'Fulfil(%r, %r, %r)' % (self.lhs, self.op, self.rhs)


class If(Compound):
    '''
    A Token that represents an "if" statement, e.g.
    "if x {...} else {...}".

    The final "else" clause is optional; its absence is indicated by `None`.
    The final "else" may alternatively begin an "else if", in which case
    its `else_` is a token of type 'if'.

    Public fields:
     - keyword - the "if" Keyword.
     - condition - an 'expr' Token representing the condition.
     - then - a '{}' Token representing the then body.
     - else_ - an Else or ElseIf representing the else body, or `None`.
    '''
    def __init__(self, keyword, condition, then, else_):
        assert keyword.tag == 'if'
        assert condition.tag == 'expr'
        assert then.tag == '{}'
        assert else_ is None or else_.tag in ('else', 'else_if')
        super(If, self).__init__('if', [keyword, condition, then, else_])
        self.keyword = keyword
        self.condition = condition
        self.then = then
        self.else_ = else_

    def matches(self, tag, condition, then, else_):
        return (
            self.tag == tag and
            matches(self.condition, condition) and
            matches(self.then, then) and
            matches(self.else_, else_)
        )

    def __repr__(self):
        return 'If(%r, %r, %r)' % (
            self.condition, self.then, self.else_)


class Return(Compound):
    '''
    A Token that represents a "return" statement, e.g. "return x".

    The expression is optional; its absence is indicated by `None`.

    Public fields:
     - keyword - the "return" Keyword.
     - expr - an 'expr' Token representing the return value, or `None`.
    '''
    def __init__(self, keyword, expr):
        assert keyword.tag == 'return'
        assert expr is None or expr.tag == 'expr'
        super(Return, self).__init__('return', [keyword, expr])
        self.keyword = keyword
        self.expr = expr

    def matches(self, tag, expr):
        return (
            self.tag == tag and
            matches(self.expr, expr)
        )

    def __repr__(self):
        return 'Return(%r)' % (self.expr,)


class Switch(Compound):
    '''
    A Token that represents a "switch" statement, e.g. "switch x {...}".
    See also Case and Else, which are really part of "switch" statements.

    There is further checking in `astbuilder` to ensure that the body only
    contains zero or more "case" statements optionally followed by an "else"
    statement, after stripping out newlines.
    '''
    def __init__(self, keyword, condition, cases):
        '''
         - keyword - the "switch" keyword.
         - condition - an 'expr' Token representing the condition.
         - cases - a '{}' Token representing the "{...}" part.
        '''
        assert keyword.tag == 'switch'
        assert condition.tag == 'expr'
        assert cases.tag == '{}'
        super(Switch, self).__init__('switch', [keyword, condition, cases])
        self.keyword = keyword
        self.condition = condition
        self.cases = cases

    def matches(self, tag, condition, cases):
        return (
            self.tag == tag and
            matches(self.condition, condition) and
            matches(self.cases, cases)
        )

    def __repr__(self):
        return 'Switch(%r, %r)' % (self.condition, self.cases)


class Throw(Compound):
    '''
    A Token that represents a "throw" statement, e.g. "throw x".

    Public fields:
     - keyword - the "throw" Keyword.
     - expr - an 'expr' Token representing the exception.
    '''
    def __init__(self, keyword, expr):
        assert keyword.tag == 'throw'
        assert expr.tag == 'expr'
        super(Throw, self).__init__('throw', [keyword, expr])
        self.keyword = keyword
        self.expr = expr

    def matches(self, tag, expr):
        return (
            self.tag == tag and
            matches(self.expr, expr)
        )

    def __repr__(self):
        return 'Throw(%r)' % (self.expr,)


class While(Compound):
    '''
    A Token that represents a "while" statement, e.g.
    "while x { ... } else { ... }".

    The final "else" clause is optional; its absence is indicated by `None`.

    Public fields:
     - keyword - the "while" Keyword.
     - condition - an 'expr' Token representing the condition.
     - body - a '{}' Token representing the loop body.
     - else_ - an Else representing the else body, or `None`.
    '''
    def __init__(self, keyword, condition, body, else_):
        assert keyword.tag == 'while'
        assert condition.tag == 'expr'
        assert body.tag == '{}'
        assert else_ is None or (type(else_) is Else and else_.tag == 'else')
        super(While, self).__init__('while', [keyword, condition, body, else_])
        self.keyword = keyword
        self.condition = condition
        self.body = body
        self.else_ = else_

    def matches(self, tag, condition, body, else_):
        return (
            self.tag == tag and
            matches(self.condition, condition) and
            matches(self.body, body) and
            matches(self.else_, else_)
        )

    def __repr__(self):
        return 'While(%r, %r, %r)' % (self.condition, self.body, self.else_)
