"""
Defines a Grammar which recognises unary expressions, i.e. expressions which contain only unary operators (at least down to the outermost brackets). The abstract grammar of a unary expression is::

    <prefix>* <nonfix> <postfix>*

The Grammar also optionally absorbs a trailing <bin-op>, i.e. the grammar implemented is::

    <prefix>* <nonfix> <postfix>* <infix>?
"""


from token import (
    Token, ASCII, UnaryExpr,
    PREFIX_TAGS, NONFIX_TAGS, POSTFIX_TAGS, INFIX_TAGS)
from grammar import Grammar
from util import discard, pass_on, Unfinished
from lexer import high_level_lexer_tags, keywords
from brackets import bracket_tags
from atom import atom_tags


__all__ = ['unary_expr', 'unary_expr_tags']


# The set of tags this parser can output, in addition to 'error' and
# those passed on from lower levels.

unary_expr_tags = {'unary_expr', 'unary_expr_infix'}


# The initial state of the unary expression grammar.

unary_expr = Grammar(discard, name="unary_expr")


# Unless overridden later, all tags are passed on unchanged.

unary_expr.set_all(high_level_lexer_tags, Grammar(pass_on))
unary_expr.set_all(keywords, Grammar(pass_on))
unary_expr.set_all(bracket_tags, Grammar(pass_on))
unary_expr.set_all(atom_tags, Grammar(pass_on))


# Errors, newlines and semicolons are passed on eagerly.

unary_expr.set_all(['error', 'newline', ';'], Grammar(pass_on, is_eager=True))


def make_unary_expr(multi_buffer):
    return UnaryExpr(
        multi_buffer.get('prefix'),
        multi_buffer.get_only('nonfix'),
        multi_buffer.get('postfix'),
        multi_buffer.get_optional('infix')
    )


ue_end = Grammar(make_unary_expr, name='ue_end')
ue_postfix_or_infix_or_end = Grammar(make_unary_expr)
ue_postfix_or_infix_or_end.set_all(
    INFIX_TAGS, ue_end, stream_name='infix')
ue_postfix_or_infix_or_end.set_all(
    POSTFIX_TAGS, ue_postfix_or_infix_or_end, stream_name='postfix')
ue_prefix_or_nonfix = Unfinished("Expected an expression")
ue_prefix_or_nonfix.ignore('newline')
ue_prefix_or_nonfix.set_all(
    NONFIX_TAGS, ue_postfix_or_infix_or_end, stream_name='nonfix')
ue_prefix_or_nonfix.set_all(
    PREFIX_TAGS, ue_prefix_or_nonfix, stream_name='prefix')
unary_expr.set_all(
    NONFIX_TAGS, ue_postfix_or_infix_or_end, stream_name='nonfix')
unary_expr.set_all(
    PREFIX_TAGS, ue_prefix_or_nonfix, stream_name='prefix')
