import sys
import bisect

from welly.printer import pretty

import run
from ..model import wtypes, pretty_value, pretty_type
from ..compiler import builtins

from ..parser import WellyParser, ASTError
from ..compiler.flattener import (
    LexicalScope, Module, from_id, FlattenerError)
from ..model.errors import *
from ..model.memory import Array


dynamic_size = wtypes.sizeof(wtypes.dynamic)


def short_pretty_type(type_):
    return pretty(pretty_type.wrap(type_), height=1, max_depth=5)


def describe_exception(exc):
    if isinstance(exc, TypeVMError):
        # 1: str
        rvalue_type = short_pretty_type(exc.rvalue_type)
        lvalue_type = short_pretty_type(exc.lvalue_type)
        if exc.noun is None:
            message = 'value of type %s is not of type %s' % (
                rvalue_type, lvalue_type)
        else:
            message = '%s is not of type %s' % (
                exc.noun, lvalue_type)
    elif isinstance(exc, TypeInferenceVMError):
        # ["x", 1]
        type1 = short_pretty_type(exc.type1)
        type2 = short_pretty_type(exc.type2)
        message = 'cannot compute %s; %s and %s have no common %s' % (
            exc.noun, type1, type2, exc.problem_type)
    elif isinstance(exc, CallTypeVMError):
        # 1()
        type_ = short_pretty_type(exc.type_)
        message = 'cannot call values of type %s' % (type_,)
    elif isinstance(exc, DeferTypeVMError):
        # defer int
        type_ = short_pretty_type(exc.type_)
        message = 'cannot "defer" or "fulfil" values of type %s' % (type_,)
    elif isinstance(exc, DereferenceTypeVMError):
        # *1
        type_ = short_pretty_type(exc.type_)
        message = 'cannot dereference values of type %s' % (type_,)
    elif isinstance(exc, SubscriptTypeVMError):
        # 1[1]
        type_ = short_pretty_type(exc.type_)
        message = 'cannot subscript values of type %s' % (type_,)
    elif isinstance(exc, SwitchTypeVMError):
        # switch 1 {}
        type_ = short_pretty_type(exc.type_)
        message = 'cannot "switch" on values of type %s' % (type_,)
    elif isinstance(exc, UnpackTypeVMError):
        # () = 1
        type_ = short_pretty_type(exc.type_)
        message = 'cannot unpack values of type %s' % (type_,)
    elif isinstance(exc, StructIndexVMError):
        # ()[1]
        struct_type = pretty(pretty_type.wrap(exc.struct_type))
        message = 'index %d is out of range for type %s' % (
            exc.index, struct_type)
    elif isinstance(exc, FieldAccessVMError):
        # 1 .x
        type_ = short_pretty_type(exc.type_)
        message = "type %s has no field '%s'" % (type_, exc.name)
    elif isinstance(exc, BinaryOperatorVMError):
        # 1 + ()
        type1 = short_pretty_type(exc.type1)
        type2 = short_pretty_type(exc.type2)
        message = 'binary operator "%s" is not defined for (%s, %s)' % (
            exc.op, type1, type2)
    elif isinstance(exc, UnaryOperatorVMError):
        # -()
        type_ = short_pretty_type(exc.type_)
        message = 'unary operator "%s" is not defined for %s' % (
            exc.op, type_)
    else:
        # No parts of the error need to be pretty-printed, so use `str()`.
        message = str(exc)
    return '%s: %s' % (exc.__class__.__name__, message)


class ReportedError(Exception):
    '''Wraps a Welly error that has already been reported.'''


class Repl(object):
    '''
    Represents the state of a Welly shell. The globals are initialised to a
    copy of `builtins.builtins`.

    Feed bytes to the shell using `feed()` and `flush()`. Completed stmts
    will be buffered, and can be run using `run_pending_stmts()`.

    Public fields.
     - globals - a dict from str to Dynamic.
     - parser - a WellyParser.
     - lines - a list of the lines of code seen so far, i.e. the strs passed
       to `feed()`.
     - line_starts - a list of ints, representing byte offsets of the starts
       of the `lines`. This is a sorted list. If there are any elements then
       the first is `0`.
    '''
    def __init__(self, on_output):
        '''
         - on_output - a callback used print out expressions used as
           statements. Takes a vm.Slot and returns nothing,
        '''
        self.parser = WellyParser(self.run_stmt)
        self.on_output = on_output
        self.globals = dict(builtins.builtins)
        self.lines = [] # List of str.
        self.line_starts = [] # List of int.

    def feed(self, line):
        '''
        Calls `self.parser.feed(line)` and increments `self.line_number()`.

         - line - a str representing a line of code, UTF-8 encoded, including
           the newline sequence if any.
        '''
        assert type(line) is str
        self.line_starts.append(self.parser.pos)
        self.lines.append(line)
        self.parser.feed(line)

    def line_number(self):
        return len(self.lines)

    def to_text_pos(self, pos):
        '''
        Given a byte position, returns a pair of ints
        `(line_number, byte_number)`, where `line_number` is the line number
        (zero-based, i.e. an index into `self.lines`) and `byte_number` is
        the offset within the line (in bytes, also zero-based).

        Conversion to the common 1-based convention for source code line
        and column numbers is the caller's responsibility.

         - pos - a position in the source code, measured in bytes from the
           start of the REPL session.
        '''
        line_number = bisect.bisect(self.line_starts, pos) - 1
        byte_number = pos - self.line_starts[line_number]
        return (line_number, byte_number)

    def flush(self):
        '''
        Calls `self.parser.flush()`.
        '''
        self.parser.flush()

    def is_flushed(self):
        return self.parser.is_flushed()

    def report_error(self, exc, loc, is_after=False):
        '''
        Report an error from Welly. Do not use this method to report
        implementation bugs other than NotImplementedError.

        The default implementation prints the error to stdout. Subclasses can
        override this method to change the way errors are reported.

         - exc - the Exception.
         - loc - a (start, end) pair representing byte offsets of the
           offending source code.
         - is_after - `True` to indicate the point just after `loc`, or
           `False` to indicate `loc` itself.
        '''
        print describe_exception(exc)
        if loc:
            start, end = loc
            line_number, byte_number = self.to_text_pos(start)
            line = self.lines[line_number].rstrip()
            # Measure string widths.
            start_col = get_display_width(
                line[:byte_number].decode('utf-8', 'replace'))
            num_bytes = min(end - start, len(line) - byte_number)
            num_cols = get_display_width(
                line[byte_number:][:num_bytes].decode('utf-8', 'replace'))
            print 'at line %d, col %d' % (line_number + 1, start_col + 1)
            print line
            if is_after:
                underline = '~' * num_cols + '^'
            else:
                underline = '^' * num_cols
            print ' ' * start_col + underline

    def run_pending_stmts(self):
        '''
        Removes and runs (or reports errors in) all buffered statements.
        Welly errors will be passed to `self.report_error()`, and turned into
        ReportedErrors. ASTErrors cause the rest of the buffer to be
        discarded. Other errors are fatal.
        '''
        try:
            self.parser.do_pending_stmts() # Calls `self.run_stmt()`.
            return
        except ASTError as e:
            assert e.errors
            # Suppress all errors after the first.
            token = e.errors[0]
            e.errors = [token]
            self.report_error(
                e, (token.start, token.end), is_after=token.is_after)
        except FlattenerError as e:
            # Syntax usage error.
            self.report_error(e, e.ast.error_loc)
        except VMError as e:
            # Error detected using static checking or at runtime.
            self.report_error(e, e.loc)
        except NotImplementedError as e:
            # Something we haven't done yet.
            self.report_error(e, None) # No location information available.
        except MemoryError as e:
            self.report_error(e, None) # No location information available.
        except OverflowError as e:
            self.report_error(e, None) # No location information available.
        raise ReportedError(e)

    def run_stmt(self, ast):
        '''
        Attempts to run `ast`, and updates `globals` if successful.

         - ast - the abstract syntax tree of a statement.

        Raises:
         - ASTError - for any kind of syntax error.
         - FlattenerError - for syntax usage errors.
         - StaticVMError - for errors detected using static checking.
         - RuntimeVMError - for errors detected at runtime.
         - NotImplementedError - for unimplemented features.
         - MemoryError, OverflowError - in extremis.
        Other Exception types are bugs.
        '''
        # Flatten.
        scope = LexicalScope(Module())
        scope.flatten_stmt(ast, 'the top-level scope')
        # Append code to push newly defined values. Also, record their names.
        new_names = []
        for token in scope.locals_.as_list():
            name = from_id(token)
            if name is not None:
                scope.flatten_identifier(token)
                new_names.append(name)
        # Construct captures array.
        capture_ids = scope.top_level.captures.as_list()
        captures = Array.malloc(len(capture_ids), dynamic_size)
        for identifier, ref in zip(capture_ids, captures):
            name = from_id(identifier)
            if name not in self.globals:
                raise UndefinedNameVMError(name, identifier.error_loc)
            ref.write(self.globals[name].to_welly())
        # Run.
        result = run.run_module_statement(
            scope.block.get_code(is_break_allowed=False),
            captures,
            on_output,
        )
        # Update `globals`.
        assert len(new_names) == len(result), (new_names, result)
        self.globals.update(zip(new_names, result))


def get_display_width(ustr):
    '''
    Counts the width of `ustr` (a unicode string) when displayed in a
    terminal using a fixed-width font.
    '''
    # WONTFIX: This should be done by a specialized library. Instead, we
    # just count unicode code points, which is not always correct.
    # https://denisbider.blogspot.co.uk/2015/09/when-monospace-fonts-arent-unicode.html
    # You need to handle at least:
    # - Combining characters. http://www.unicode.org/reports/tr29/
    # - "Full width East Asian" characters, which includes other full
    # width characters too. http://unicode.org/reports/tr11/
    # - Left-to-right vs right-to-left text.
    # This is too complicated, and not important enough.
    return len(ustr)


def on_output(is_static, value):
    print pretty(lambda: (u'', u': ', u'', [
        pretty_value.wrap(value),
        pretty_type.wrap(value.type_),
    ]))


def run_lines(read_line, stop_on_error=True):
    '''
    Starts a Welly shell, feeds it each of `lines`, then discards the shell.
    Runs all pending stmts after each line and at EOF, and catches and reports
    Welly errors. Other errors are fatal.

     - read_line - a callback that takes a boolean `is_flushed` and returns
       a line of source code, or `None` to indicate that there are no more
       lines. `is_flushed` can be used to decide which prompt to show; `True`
       indicates the start of a new statement.

     - stop_on_error - if true, stops at Welly errors.
    '''
    repl = Repl(on_output)

    def run_pending_stmts():
        '''Returns `True` to stop.'''
        try:
            repl.run_pending_stmts()
            return False
        except ReportedError:
            return stop_on_error

    while True:
        line = read_line(repl.is_flushed())
        if line is None:
            # EOF.
            repl.flush()
            run_pending_stmts()
            break
        repl.feed(line)
        if run_pending_stmts():
            break


def run_repl():
    def read_line(is_flushed):
        prompt = '>>>> ' if is_flushed else '.... '
        try:
            return raw_input(prompt) + '\n'
        except EOFError:
            print
            return None

    run_lines(read_line, stop_on_error=False)


def run_file(file_like):
    line_iter = iter(file_like)
    def read_line(is_flushed):
        try:
            return next(line_iter)
        except StopIteration:
            return None

    run_lines(read_line, stop_on_error=True)


def main(args):
    if len(args) == 0:
        run_repl()
    elif len(args) == 1:
        if args[0] == '-':
            run_file(sys.stdin)
        else:
            with open(args[0]) as f:
                run_file(f)
    else:
        sys.exit('error: wrong number of arguments')


if __name__ == '__main__':
    sys.setrecursionlimit(10000)
    main(sys.argv[1:])
