import sys
import traceback

import pytest

from ..model import const, wtypes, pretty_type
from ..compiler import builtins

from .run import *
from ..model.memory import Ref, Array
from ..model.const import BASE, CONST, VAR
from ..model.wtypes import *
from ..model.ops import *
from ..parser import WellyParser
from ..compiler.flattener import LexicalScope, Module, from_id
from ..compiler.typed_code import Block
from ..printer import pretty


sys.setrecursionlimit(10000)


slow = pytest.mark.skipif(
    not (hasattr(pytest, 'config') and pytest.config.getoption("--run-slow")),
    reason="need --run-slow option to run"
)


_test_builtins = builtins.builtins.copy()
_test_builtins['three'] = Dynamic.new(wtypes.int_, (3,), constness=VAR)


def test_num_locals():
    '''
    Tests that `run()` removes any locals it declares, no matter how it
    returns.
    '''
    captures = []
    locals_ = [Ref.new((1,))]
    # Test 'fallthrough'.
    reason = run([
        ('local', 0),
        ('init', 1),
    ], captures, locals_, None)
    assert reason == ('fallthrough', [])
    assert len(locals_) == 1
    # Test 'break'.
    reason = run([
        ('local', 0),
        ('init', 1),
        ('break',),
    ], captures, locals_, None)
    assert reason == ('break',)
    assert len(locals_) == 1
    # TODO: Test 'continue'.
    # Test 'return'.
    reason = run([
        ('local', 0),
        ('init', 1),
        ('local', 0),
        ('return', 1),
    ], captures, locals_, None)
    assert reason == ('return', locals_[0])
    assert len(locals_) == 1
    # TODO: Test 'throw'.


def make_do_stmt(on_output):
    def do_stmt(ast):
        # Flatten.
        scope = LexicalScope(Module())
        scope.flatten_stmt(ast, 'the top-level scope')
        # Construct captures array.
        capture_ids = scope.top_level.captures.as_list()
        captures = linker.Stack()
        for identifier in capture_ids:
            name = from_id(identifier)
            captures.push_static(_test_builtins[name])
        flat_code = scope.block.get_code(is_break_allowed=False)
        # Link.
        block = Block()
        locals_ = linker.Stack()
        linker.link(
            block,
            flat_code,
            captures,
            locals_,
            make_vm_function(on_output),
        )
        typed_code = block.get_code(is_break_allowed=False)
#        print "Typed code:"
#        print pretty(typed_code)
        # Run.
        result = run(
            typed_code,
            [], # captures
            [], # locals_.
            on_output,
        )
#        print result
    return do_stmt


def format_output(is_static, constness, type_, words):
    return "%s; %s; %s; %r" % (
        is_static,
        const.switch(constness),
        pretty(pretty_type.wrap(type_), height=1, max_depth=5),
        words
    )


def check(source, expected, error_message=None):
    expected = [format_output(*x) for x in expected]
    outputs = []
    def on_output(is_static, value):
        outputs.append(format_output(
            is_static, value.constness, value.type_, value.ref.read()))
    parser = WellyParser(make_do_stmt(on_output))
    parser.feed(source)
    parser.flush()
    try:
        parser.do_pending_stmts()
    except Exception as e:
        # Terminated abnormally. Look at the exception.
        actual_error = '%s: %s' % (e.__class__.__name__, e)
        if actual_error == error_message:
            return
        traceback.print_exc()
        assert not outputs
        print "Observed error:"
        print actual_error
        print "Expected error:"
        print error_message
        assert False, "Error mismatch!"
    # Terminated normally. Look at the outputs.
    if outputs == expected:
        return
    print "Outputs:"
    print pretty(outputs)
    assert error_message is None
    print "Expected:"
    print pretty(expected)
    assert False, "Mismatch!"


# Useful output fragments.

int_size = wtypes.sizeof(wtypes.int_)
union_size = wtypes.sizeof(wtypes.union([]))

def static_const(t, value):
    assert type(value) is tuple
    return (True, CONST, t, value)

def const_int(n, is_static=True):
    return (is_static, CONST, wtypes.int_, (n,))

def const_bool(tf, is_static=True):
    return (is_static, CONST, wtypes.bool_, (tf,))

def const_str(s, is_static=True):
    assert type(s) is str
    return (is_static, CONST, wtypes.str_, (s,))

def var_int(n, is_static=True):
    return (is_static, VAR, wtypes.int_, (n,))

def int_list(*ints):
    return Array.from_list([(i,) for i in ints], int_size).to_welly()

empty_array = int_list()

intf = wtypes.function([], wtypes.int_)

type_repr = repr(wtypes.type_.read())


# Tests.

def test_integers():
    check('3', [const_int(3)])


def test_bools():
    check('true', [const_bool(True)])
    check('false', [const_bool(False)])


def test_strings():
    check('"hello"', [const_str('hello')])


def test_lines():
    check('3;', [const_int(3)])
    check('3\n', [const_int(3)])
    check('3\\\n', [const_int(3)])
    check('3\n3', [const_int(3), const_int(3)])
    check('3\\\n3', [], 'ASTError: Expected ";" or a newline')


def test_globals():
    check('int', [static_const(wtypes.type_, wtypes.int_.read())])
    check('three', [var_int(3)])


def test_unary():
    check('+3', [const_int(3)])
    check('-3', [const_int(-3)])
    check('+three', [const_int(3, is_static=False)])
    check('-three', [const_int(-3, is_static=False)])


def test_binary():
    check('3 + 1', [const_int(4)])
    check('3 - 1', [const_int(2)])
    check('three + 1', [const_int(4, is_static=False)])
    check('three - 1', [const_int(2, is_static=False)])
    check('3 < 1', [const_bool(False)])
    check('"he" + "llo"', [const_str('hello')])


def test_arithmetic_error():
    check('1 / 0', [], 'ArithmeticVMError: cannot divide by zero')


def test_variables():
    check('const x = 3', [])
    check('var x = 3', [])
    check('const x = three', [])
    check('var x = three', [])


def test_structs():
    check('()', []) # Output is suppressed because its type is `void`.
    check('1+()', [],
        'BinaryOperatorVMError: '
        'binary operator "+" is not defined for (INT, STRUCT)')
    check('(1, ())', [static_const(
        wtypes.struct([wtypes.int_, wtypes.void]),
        (1, ),
    )])
    check('(1, 2, 3)', [static_const(
        wtypes.struct([wtypes.int_, wtypes.int_, wtypes.int_]),
        (1, 2, 3),
    )])
    check('(1, 2, three)', [(
        False,
        CONST,
        wtypes.struct([wtypes.int_, wtypes.int_, wtypes.int_]),
        (1, 2, 3),
    )])


def test_struct_types():
    check('struct(int)', [(
        True,
        CONST,
        wtypes.type_,
        wtypes.struct([wtypes.int_]).read(),
    )])
    check('struct(int, str)', [(
        True,
        CONST,
        wtypes.type_,
        wtypes.struct([wtypes.int_, wtypes.str_]).read(),
    )])
    check('fn(t: type) { return struct(t) }(int)', [(
        False,
        CONST,
        wtypes.type_,
        wtypes.struct([wtypes.int_]).read(),
    )])
    check('fn(t: type) { return struct(t, str) }(int)', [(
        False,
        CONST,
        wtypes.type_,
        wtypes.struct([wtypes.int_, wtypes.str_]).read(),
    )])


def test_arrays():
    check('[]', [(
        True,
        CONST,
        wtypes.array0,
        empty_array,
    )])
    check('[1,2,3]', [(
        True,
        CONST,
        wtypes.array(CONST, wtypes.int_),
        int_list(1, 2, 3),
    )])
    check('[1,2,three]', [(
        False,
        CONST,
        wtypes.array(CONST, wtypes.int_),
        int_list(1, 2, 3),
    )])
    check('[()]', [(
        True,
        CONST,
        wtypes.array(CONST, wtypes.void),
        Array.from_list([()], 0).to_welly(),
    )])
    check('[FOO: union {FOO; BAR}, BAR]', [static_const(
        wtypes.array(CONST, wtypes.union([('FOO', []), ('BAR', [])])),
        Array.from_list(
            [('FOO', []), ('BAR', [])],
            union_size,
        ).to_welly(),
    )])
    check('[FOO, BAR]', [static_const(
        wtypes.array(CONST, wtypes.union([('FOO', []), ('BAR', [])])),
        Array.from_list(
            [('FOO', []), ('BAR', [])],
            union_size,
        ).to_welly(),
    )])
    check('[FOO, FOO(3)]',
        [],
        "TypeInferenceVMError: cannot compute the array element type; "
        "UNION and UNION have no common supertype"
    )
    check('[FOO, 3]',
        [],
        "TypeInferenceVMError: cannot compute the array element type; "
        "UNION and INT have no common supertype"
    )
    check('[FOO, three]',
        [],
        "TypeInferenceVMError: cannot compute the array element type; "
        "UNION and INT have no common supertype"
    )


def test_array_types():
    check('array int', [
        (True, CONST, wtypes.type_, wtypes.array(BASE, wtypes.int_).read()),
    ])
    check('array const int', [
        (True, CONST, wtypes.type_, wtypes.array(CONST, wtypes.int_).read()),
    ])
    check('array var int', [
        (True, CONST, wtypes.type_, wtypes.array(VAR, wtypes.int_).read()),
    ])
    check('fn (t: type): type { return array t }(int)', [
        (False, CONST, wtypes.type_, wtypes.array(BASE, wtypes.int_).read()),
    ])
    check('fn (t: type): type { return array const t }(int)', [
        (False, CONST, wtypes.type_, wtypes.array(CONST, wtypes.int_).read()),
    ])
    check('fn (t: type): type { return array var t }(int)', [
        (False, CONST, wtypes.type_, wtypes.array(VAR, wtypes.int_).read()),
    ])


def test_casts():
    check('(): void', []) # Output is suppressed because its type is `void`.
    check('3: int', [const_int(3)])
    check('three: int', [const_int(3, is_static=False)])
    check('3: str', [],
        "TypeVMError: value of type INT is not of type STR"
    )
    check('3: print', [],
        "TypeVMError: the type used for the cast is not of type UNION")
    # Test casts in struct literals.
    check('((1:int), 2, 3: int)', [static_const(
        wtypes.struct([wtypes.int_, wtypes.int_, wtypes.int_]),
        (1, 2, 3)
    )])


def test_array_subtypes():
    check('[]: array var int', [static_const(
        wtypes.array(VAR, wtypes.int_),
        empty_array,
    )])
    check('[1, 2, 3]: array0',
        [],
        "TypeVMError: value of type ARRAY is not of type ARRAY0"
    )
    check('''
        module array_subtypes_local {
            const x = []
            x: array int
            x: array const int
            x: array var int
        }
    ''', [
        static_const(wtypes.array(BASE, wtypes.int_), empty_array),
        static_const(wtypes.array(CONST, wtypes.int_), empty_array),
        static_const(wtypes.array(VAR, wtypes.int_), empty_array),
    ])


def test_array_concat():
    check('[1, 2, 3] + [4, 5]', [static_const(
        wtypes.array(CONST, wtypes.int_),
        int_list(1, 2, 3, 4, 5),
    )])
    check('[1, 2, 3] + []', [static_const(
        wtypes.array(CONST, wtypes.int_),
        int_list(1, 2, 3),
    )])
    check('[] + [1, 2, 3]', [],
        'StaticMessageVMError: "x + y" is not defined when x is an array0; '
        'cast x to the desired array type'
    )
    check('[] + []', [],
        'StaticMessageVMError: "x + y" is not defined when x is an array0; '
        'cast x to the desired array type'
    )
    check('[FOO: union{FOO; BAR}] + [BAR]', [static_const(
        wtypes.array(CONST, wtypes.union([('FOO', []), ('BAR', [])])),
        Array.from_list(
            [('FOO', []), ('BAR', [])],
            union_size,
        ).to_welly(),
    )])
    check('[FOO] + [BAR: union{FOO; BAR}]',
        [],
        "TypeVMError: the element type is not of type UNION"
    )
    check('''module array_const_concat_dynamic {
        var a = [FOO]
        (a: array const union {FOO; BAR}) + [BAR]
    }''', [(
        False,
        CONST,
        wtypes.array(CONST, wtypes.union([('FOO', []), ('BAR', [])])),
        Array.from_list(
            [('FOO', []), ('BAR', [])],
            union_size,
        ).to_welly(),
    )])
    check('''module array_concat_var_dynamic {
        var a = []
        (a: array var union {FOO}) + [FOO]
    }''', [(
        False,
        CONST,
        wtypes.array(VAR, wtypes.union([('FOO', [])])),
        Array.from_list(
            [('FOO', [])],
            union_size,
        ).to_welly(),
    )])
    check('''([]: array const union {FOO}) + [FOO]''', [static_const(
        wtypes.array(CONST, wtypes.union([('FOO', [])])),
        Array.from_list(
            [('FOO', [])],
            union_size,
        ).to_welly(),
    )])
    check('''([]: array var union {FOO}) + [FOO]''', [(
        False,
        CONST,
        wtypes.array(VAR, wtypes.union([('FOO', [])])),
        Array.from_list(
            [('FOO', [])],
            union_size,
        ).to_welly(),
    )])
    check('''([]: array union {FOO}) + [FOO]''', [(
        False,
        CONST,
        wtypes.array(BASE, wtypes.union([('FOO', [])])),
        Array.from_list(
            [('FOO', [])],
            union_size,
        ).to_welly(),
    )])
    check('''module array_concat_empty {
        var x = []
        x += [6]
    }
    ''', [],
        'StaticMessageVMError: "x + y" is not defined when x is an array0; '
        'cast x to the desired array type'
    )
    check('''module array_concat_inplace1 {
        var x: array const int = []
        x += [6]
        x
    }
    ''', [(
        True,
        VAR,
        wtypes.array(CONST, wtypes.int_),
        int_list(6),
    )])
    check('''module array_concat_inplace2 {
        var x = [1]
        x += [6]
        x
    }
    ''', [(
        True,
        VAR,
        wtypes.array(CONST, wtypes.int_),
        int_list(1, 6),
    )])
    check('[1] + ["a"]',
        [],
        "TypeVMError: the element type is not of type INT"
    )
    check('''module array_concat_type {
        const a = [int, bool] + [str]
        a[0]
        a[1]
        a[2]
    }''', [
        static_const(wtypes.type_, wtypes.int_.read()),
        static_const(wtypes.type_, wtypes.bool_.read()),
        static_const(wtypes.type_, wtypes.str_.read()),
    ])
    check('''
        module array_concat_mutate {
            const a = ([]: array var int) + [1, 2, 3, 4, 5]
            a[2] = 88
            a
        }
    ''', [static_const(
        wtypes.array(VAR, wtypes.int_),
        int_list(1, 2, 88, 4, 5),
    )])
    check('''
        module array_concat_reverse {
            const a = ([]: array var int) + [1, 2, 3, 4, 5]
            var (i, j) = (0, a.length-1)
            while i < j {
              (a[i], a[j]) = (a[j], a[i])
              i += 1; j -= 1
            }
            a; i; j
        }
    ''', [
        static_const(
            wtypes.array(VAR, wtypes.int_),
            int_list(5, 4, 3, 2, 1),
        ),
        (True, VAR, wtypes.int_, (2,)),
        (True, VAR, wtypes.int_, (2,)),
    ])


def test_array_repeat():
    array_int = wtypes.array(CONST, wtypes.int_)
    array_base_int = wtypes.array(BASE, wtypes.int_)
    check('[1, 2] * 3', [static_const(
        array_int,
        int_list(1, 2, 1, 2, 1, 2),
    )])
    check('[1, 2] * -3', [],
        'ArrayRepeatVMError: invalid array repeat count -3'
    )
    check('[1, 2] * three', [
        (False, CONST, array_int, int_list(1, 2, 1, 2, 1, 2))
    ])
    check('[1, 2] * -three', [],
        'ArrayRepeatVMError: invalid array repeat count -3'
    )
    check('fn(x: array const int) { x * 3 }([1, 2])', [
        (False, CONST, array_int, int_list(1, 2, 1, 2, 1, 2))
    ])
    check('fn(x: array int) { x * 3 }([1, 2])', [
        (False, CONST, array_base_int, int_list(1, 2, 1, 2, 1, 2))
    ])
    check('fn(x: array const int) { x * -3 }([1, 2])', [],
        'ArrayRepeatVMError: invalid array repeat count -3'
    )
    check('fn(x: array const int) { x * three }([1, 2])', [
        (False, CONST, array_int, int_list(1, 2, 1, 2, 1, 2))
    ])
    check('fn(x: array int) { x * three }([1, 2])', [
        (False, CONST, array_base_int, int_list(1, 2, 1, 2, 1, 2))
    ])
    check('fn(x: array const int) { x * -three }([1, 2])', [],
        'ArrayRepeatVMError: invalid array repeat count -3'
    )
    check('3 * [1, 2]', [static_const(
        array_int,
        int_list(1, 2, 1, 2, 1, 2),
    )])
    check('-3 * [1, 2]', [],
        'ArrayRepeatVMError: invalid array repeat count -3'
    )
    check('three * [1, 2]', [
        (False, CONST, array_int, int_list(1, 2, 1, 2, 1, 2))
    ])
    check('-three * [1, 2]', [],
        'ArrayRepeatVMError: invalid array repeat count -3'
    )
    check('fn(x: array const int) { 3 * x }([1, 2])', [
        (False, CONST, array_int, int_list(1, 2, 1, 2, 1, 2))
    ])
    check('fn(x: array int) { 3 * x }([1, 2])', [
        (False, CONST, array_base_int, int_list(1, 2, 1, 2, 1, 2))
    ])
    check('fn(x: array const int) { -3 * x }([1, 2])', [],
        'ArrayRepeatVMError: invalid array repeat count -3'
    )
    check('fn(x: array const int) { three * x }([1, 2])', [
        (False, CONST, array_int, int_list(1, 2, 1, 2, 1, 2))
    ])
    check('fn(x: array int) { three * x }([1, 2])', [
        (False, CONST, array_base_int, int_list(1, 2, 1, 2, 1, 2))
    ])
    check('fn(x: array const int) { -three * x }([1, 2])', [],
        'ArrayRepeatVMError: invalid array repeat count -3'
    )
    check('''
        module array_repeat_mutate {
            const a = ([]: array var int) + [1, 2, 3, 4, 5]
            var b = a
            b[2] = 77
            b *= 1
            b[3] = 88
            a[1]
            b[1]
            a[2]
            b[2]
            a[3]
            b[3]
        }
    ''', [
        var_int(2),
        var_int(2, is_static=False),
        var_int(77),
        var_int(77, is_static=False),
        var_int(4),
        var_int(88, is_static=False),
    ])


def test_array_slice():
    array_int = wtypes.array(CONST, wtypes.int_)
    array_base_int = wtypes.array(BASE, wtypes.int_)
    a = Array.from_list([(5,), (6,), (7,), (8,)], 1)
    check('[5, 6, 7, 8].slice(1, 3)', [
        (False, CONST, array_int, a.slice(1, 3).to_welly())
    ])
    check('([5, 6, 7, 8]: array int).slice(1, 3)', [
        (False, CONST, array_base_int, a.slice(1, 3).to_welly())
    ])
    check('[5, 6, 7, 8].slice(-1, 1)', [],
        'ArraySliceVMError: array slice from -1 to 1 is out of range for an '
        'array of length 4'
    )
    check('[5, 6, 7, 8].slice(3, 5)', [],
        'ArraySliceVMError: array slice from 3 to 5 is out of range for an '
        'array of length 4'
    )
    check('[5, 6, 7, 8].slice(3, 1)', [],
        'ArraySliceVMError: array slice from 3 to 1 is out of range for an '
        'array of length 4'
    )


def test_array0_repeat():
    check('[] * 3', [static_const(
        array0,
        int_list(),
    )])
    check('[] * -3', [],
        'ArrayRepeatVMError: invalid array repeat count -3'
    )
    check('[] * three', [
        (False, CONST, array0, int_list())
    ])
    check('[] * -three', [],
        'ArrayRepeatVMError: invalid array repeat count -3'
    )
    check('fn(x: array0) { x * 3 }([])', [
        (False, CONST, array0, int_list())
    ])
    check('fn(x: array0) { x * -3 }([])', [],
        'ArrayRepeatVMError: invalid array repeat count -3'
    )
    check('fn(x: array0) { x * three }([])', [
        (False, CONST, array0, int_list())
    ])
    check('fn(x: array0) { x * -three }([])', [],
        'ArrayRepeatVMError: invalid array repeat count -3'
    )
    check('3 * []', [static_const(
        array0,
        int_list(),
    )])
    check('-3 * []', [],
        'ArrayRepeatVMError: invalid array repeat count -3'
    )
    check('three * []', [
        (False, CONST, array0, int_list())
    ])
    check('-three * []', [],
        'ArrayRepeatVMError: invalid array repeat count -3'
    )
    check('fn(x: array0) { 3 * x }([])', [
        (False, CONST, array0, int_list())
    ])
    check('fn(x: array0) { -3 * x }([])', [],
        'ArrayRepeatVMError: invalid array repeat count -3'
    )
    check('fn(x: array0) { three * x }([])', [
        (False, CONST, array0, int_list())
    ])
    check('fn(x: array0) { -three * x }([])', [],
        'ArrayRepeatVMError: invalid array repeat count -3'
    )


def test_lvalues():
    check('const () = ()', [])
    check('const () = (3,)', [],
        'NumberOfFieldsVMError: wrong number of fields - expected 0 but got 1')
    check('const (a,) = ()', [],
        'NumberOfFieldsVMError: wrong number of fields - expected 1 but got 0')
    check('var (x,) = (3,)', [])
    check('const (x,) = (3,)', [])
    check('const (x,) = (three,)', [])
    check('const (x, (y: int)) = (3, 1)', [])
    check('const (x, (y: int)) = (three, 1)', [])


def test_modules():
    check('module', [
        (True, CONST, wtypes.type_, wtypes.module.read()),
    ])
    check('module {}', [
        (False, CONST, wtypes.module, ({},)),
    ])
    check('module {1}', [
        const_int(1),
        (False, CONST, wtypes.module, ({},)),
    ])


def test_module_captures():
    check('module {int}', [
        static_const(wtypes.type_, wtypes.int_.read()),
        (False, CONST, wtypes.module, ({},)),
    ])
    check('module {int: type}', [
        static_const(wtypes.type_, wtypes.int_.read()),
        (False, CONST, wtypes.module, ({},)),
    ])


def test_module_variables():
    check('module {const x = 3}', [
        (False, CONST, wtypes.module, ({
            'x': Dynamic.new(wtypes.int_, (3,)).to_welly(),
        },)),
    ])


def test_module_variable_access():
    check('module {const x = 3; x}', [
        const_int(3),
        (False, CONST, wtypes.module, ({
            'x': Dynamic.new(wtypes.int_, (3,)).to_welly(),
        },)),
    ])
    check('module {var x = 3; x}', [
        (True, VAR, wtypes.int_, (3,)),
        (False, CONST, wtypes.module, ({
            'x': Dynamic.new(wtypes.int_, (3,), constness=VAR).to_welly(),
        },)),
    ])
    check('module {var x = 3; [x]}', [
        (False, CONST, wtypes.array(CONST, wtypes.int_), int_list(3)),
        (False, CONST, wtypes.module, ({
            'x': Dynamic.new(wtypes.int_, (3,), constness=VAR).to_welly(),
        },)),
    ])


def test_named_modules():
    check('module foo {}', [])
    check('module { module foo {} }', [
        (False, CONST, wtypes.module, ({
            'foo': Dynamic.new(wtypes.module, ({},)).to_welly(),
        },)),
    ])
    check('module { module foo {const x = 3}; module foo {} }', [
        (False, CONST, wtypes.module, ({
            'foo': Dynamic.new(wtypes.module, ({},)).to_welly(),
        },)),
    ])


def test_module_captures_nested():
    check('module { const x = 3; module foo {x} }', [
        const_int(3),
        (False, CONST, wtypes.module, ({
            'x': Dynamic.new(wtypes.int_, (3,)).to_welly(),
            'foo': Dynamic.new(wtypes.module, ({},)).to_welly(),
        },)),
    ])


def test_inplace():
    check('module { const x = 3; x -= 1 }', [],
        'ConstnessVMError: can only assign to "var" locations')
    check('module { var x = 3; x -= 1 }', [
        (False, CONST, wtypes.module, ({
            'x': Dynamic.new(wtypes.int_, (2,), constness=VAR).to_welly(),
        },)),
    ])


def test_inplace_captured():
    check('module { const x = 3; module foo { x -= 1 } }', [],
        'ConstnessVMError: can only assign to "var" locations')
    check('module { var x = 3; module foo { x -= 1 } }', [
        (False, CONST, wtypes.module, ({
            'x': Dynamic.new(wtypes.int_, (2,), constness=VAR).to_welly(),
            'foo': Dynamic.new(wtypes.module, ({},)).to_welly(),
        },)),
    ])


def test_inplace_shadowed():
    check('module { const x = 3; module foo { var x = 3; x -= 1 } }', [
        (False, CONST, wtypes.module, ({
            'x': Dynamic.new(wtypes.int_, (3,)).to_welly(),
            'foo': Dynamic.new(wtypes.module, ({
                'x': Dynamic.new(wtypes.int_, (2,), constness=VAR).to_welly(),
            },)).to_welly(),
        },)),
    ])


def test_if():
    check('if (1) { 2 } else { 3 }', [],
        'TypeVMError: the "if" condition is not of type BOOL')
    check('if (1 < 2) { 3 } else { }', [const_int(3)])
    check('if (1 < 2) { } else { 4 }', [])
    check('if (1 < 2) { 3 } else { 4 }', [const_int(3)])


def test_ternary():
    check('1 ? 2 : 3 ', [],
        'TypeVMError: the "if" condition is not of type BOOL')
    check('1 < 2 ? 3 : 4', [const_int(3)])
    check('1 > 2 ? 3 : 4', [const_int(4)])
    check('''
        module test_ternary1 {
            const b = 1 > 2
            b ? 3 : 4
        }
    ''', [const_int(4)])
    check('''
        module test_ternary2 {
            var b = 1 > 2
            b ? 3 : 4
        }
    ''', [const_int(4, is_static=False)])
    check('''
        module test_ternary3 {
            var x = 4
            1 > 2 ? 3 : x
        }
    ''', [var_int(4)])
    check('''
        module test_ternary_var1 {
            var x = FOO
            var y = BAR
            var b = 1 > 2
            (b ? x : y) = FOO
        }
    ''', [], 'ConstnessVMError: can only assign to "var" locations')
    check('''
        module test_ternary_var2 {
            var x = FOO
            var y = BAR
            var b = 1 > 2
            const a = &(b ? x : y)
            *a = FOO
        }
    ''', [], 'ConstnessVMError: can only assign to "var" locations')


def test_field():
    check(r'1 .x', [],
        "FieldAccessVMError: type INT has no field 'x'")
    check(r'fn (): int { return 1 }().x', [],
        "FieldAccessVMError: type INT has no field 'x'")
    check(r'"".x', [],
        "FieldAccessVMError: type STR has no field 'x'")
    check(r'fn (): str { return "" }().x', [],
        "FieldAccessVMError: type STR has no field 'x'")
    check(r'"".length', [
        const_int(0)
    ])
    check(r'fn (): str { return "" }().length', [
        const_int(0, is_static=False)
    ])
    check(r'"a".length', [
        const_int(1)
    ])
    check(r'fn (): str { return "a" }().length', [
        const_int(1, is_static=False)
    ])
    check(r'[].x', [],
        "FieldAccessVMError: type ARRAY0 has no field 'x'")
    check(r'fn (): array0 { return [] }().x', [],
        "FieldAccessVMError: type ARRAY0 has no field 'x'")
    check(r'[].length', [
        const_int(0)
    ])
    check(r'fn (): array0 { return [] }().length', [
        const_int(0, is_static=False)
    ])
    check(r'[3].length', [
        const_int(1)
    ])
    check(r'[three].length', [
        const_int(1, is_static=False)
    ])
    check(r'[(), ()].length', [
        const_int(2)
    ])
    check(r'fn (): (array int) { return [3] }().length', [
        const_int(1, is_static=False)
    ])
    check(r'fn (): (array void) { return [(), ()] }().length', [
        const_int(2, is_static=False)
    ])
    check(r'().x', [],
        "FieldAccessVMError: type STRUCT has no field 'x'"
    )
    check(r'fn (): void { return () }().x', [],
        "FieldAccessVMError: type STRUCT has no field 'x'"
    )
    check(r'(1,).x', [],
        "FieldAccessVMError: type STRUCT has no field 'x'"
    )
    check(r'fn (): (struct(int)) { return (1,) }().x', [],
        "FieldAccessVMError: type STRUCT has no field 'x'"
    )
    #check(r'int.x', [])
    #check(r'fn (): type { return int }().x', [])
    #check(r'module.x', [])
    #check(r'fn (): type { return module }().x', [])
    check(r'module {}.x', [],
        "StaticnessVMError: the module must be static in order to access its "
        "'x' member")
    check(r'fn (): module { return module {} }().x', [],
        "StaticnessVMError: the module must be static in order to access its "
        "'x' member")
    check(r'''module foo {
        module m {}
        m.x
    }''', [], "NameVMError: name 'x' is not defined in this module")
    check(r'''module {
        fn (): module {
            module m {};
            return m
        }().x
    }''', [],
        "StaticnessVMError: the module must be static in order to access its "
        "'x' member")
    check(r'''module foo {
        module m { const x = 1 }
        m.x
    }''', [const_int(1)])
    check(r'''module {
        fn (): module {
            module m { const x = 1 }
            return m
        }().x
    }''', [],
        "StaticnessVMError: the module must be static in order to access its "
        "'x' member")
    check(r'''module foo {
        module m { const x = 1 }
        const x = m.x
    }''', [])
    # TODO: Test struct types.
    # TODO: Test union types.


def test_subscript():
    check(r'''[1, 2, 3]["hello"]''', [],
        "TypeVMError: the index is not of type INT")
    check(r'''fn(a: array int) { a["hello"] }''', [],
        "TypeVMError: the index is not of type INT")
    check(r'''fn (s: struct(int, int, int, int)) { s[1] }((3, 4, 5, 6))''', [
        const_int(4, is_static=False)
    ])
    check(r'fn (i: int) { [3, 4, 5, 6][i] }(1)', [
        const_int(4, is_static=False)
    ])
    check(r'fn (a: array const int) { a[1] }([3, 4, 5, 6])', [
        const_int(4, is_static=False)
    ])
    check(r'fn (a: array int) { a[1] }([3, 4, 5, 6])', [
        (False, BASE, wtypes.int_, (4,)),
    ])


def test_invalid_calls():
    check(r'1()', [],
        "CallTypeVMError: cannot call values of type INT")


def test_function_calls():
    check(r'print("Hello, World!\n")', [])
    check(r'print("Hello", "World")', [],
        'NumberOfArgumentsVMError: wrong number of arguments - '
        'expected 1 but got 2')
    check(r'print()', [],
        'NumberOfArgumentsVMError: wrong number of arguments - '
        'expected 1 but got 0')
    check(r'print(3)', [],
        "TypeVMError: value of type INT is not of type STR")
    check(r'1: fn(){return int}()', [],
        'StaticnessVMError: the type used for the cast must be static')


# TODO: Test macro calls.


def test_function_types():
    check('fn(): void', [static_const(
        wtypes.type_,
        wtypes.function([], wtypes.void).read(),
    )])
    check('fn (): int', [static_const(wtypes.type_, intf.read())])
    if 0:
        # Test named function types.
        check('fn intf(): int', [])
    check('fn (str, bool): int', [
        static_const(wtypes.type_, wtypes.function(
            [wtypes.str_, wtypes.bool_],
            wtypes.int_
        ).read())
    ])
    check('fn (x: str, y: bool): int', [
        static_const(wtypes.type_, wtypes.function(
            [wtypes.str_, wtypes.bool_],
            wtypes.int_
        ).read())
    ])
    check('fn (str, bool): print', [],
        "TypeVMError: the return type is not of type UNION")
    check('fn (str, print): int', [],
        "TypeVMError: the parameter type is not of type UNION")
    check('fn (print, bool): int', [],
        "TypeVMError: the parameter type is not of type UNION")


def test_functions():
    check('const f = fn(): void {}', [])
    check('fn(): int { return 3 }()', [const_int(3, is_static=False)])
    check('fn(): print { return 3 }()', [],
        "TypeVMError: the return type is not of type UNION")
    check('fn(int): int { return 3 }(3)', [],
        'ASTError: Function parameters must use the syntax '
        '"<name>: <type>"')
    check('fn(p: print): int { return 3 }(3)', [],
        "TypeVMError: the parameter type is not of type UNION")
    check('fn(t: type): type { return fn(t): t }(int)', [
        (
            False,
            CONST,
            wtypes.type_,
            wtypes.function([wtypes.int_], wtypes.int_).read(),
        )
    ])
    check('fn f(): bool { f() }', [],
        "StaticMessageVMError: missing return value")


def test_closures():
    check('''\
    fn(x: int): (fn(): int) {
      return fn(): int { return x }
    }(3)()
    ''', [
        const_int(3, is_static=False)
    ])


def test_module_in_function():
    check('''fn(t: type) { 3: t }(int)''', [],
        'StaticnessVMError: the type used for the cast must be static')
    check('''\
    fn make(t: type): module {
        return module { 3: t }
    }
    ''', [])
    check('''\
    const m = fn (t: type): module {
        var t = t
        return module { 3: t }
    }(int)
    ''', [],
        'StaticnessVMError: the type used for the cast must be constant',
    )
    check('''\
    const m = fn (t: type): module {
        return module {
            fn id(x: t): t { return x }
        }
    }(int)
    ''', [])
    check('''\
    const m = fn (t: type): module {
        return module {
            fn id(x: t): t { return x }
        }
    }(int)
    ''', [],
        'StaticnessVMError: the value used as a type must be constant',
    )
    check('''\
    module generic_identity {
        fn make(t: type): module {
            return module {
                fn id(x: t): t { return x }
            }
        }
        const m = make(int)
        m.id(3)
    }
    ''', [
        const_int(3, is_static=False),
    ])


def test_staticness():
    check('fn (): void { 3 }()', [const_int(3)])
    check('fn (): void { -3 }()', [const_int(-3)])
    check('fn (): void { 3+3 }()', [const_int(6)])
    check('fn (): void { three }()', [var_int(3)])
    check('fn (): void { -three }()', [const_int(-3, is_static=False)])
    check('fn (): void { 3-three }()', [const_int(0, is_static=False)])


def test_function_named():
    check('''\
    module increment {
      fn increment(x: int): int { return x + 1 }
      increment(3)
    }
    ''', [
        const_int(4, is_static=False)
    ])


def test_function_subtypes():
    # As per "doc/compile-egs/function-types.txt".
    check('''\
    module function_type_check1 {
      const (subtype, midtype, supertype) =
        (array0, array const int, array int)
      fn f(x: supertype): midtype { return []: subtype }
      const g = f: (fn(midtype): supertype)
      g([]: subtype)
    }
    ''', [
        (False, CONST, wtypes.array(BASE, wtypes.int_), empty_array),
    ])
    check('''\
    module function_type_check2 {
      const (subtype, midtype, supertype) =
        (array0, array const int, array int)
      fn f(x: supertype): subtype { return []: midtype }
      const g = f: (fn(midtype): supertype)
      g([]: subtype)
    }
    ''', [],
        "TypeVMError: value of type ARRAY is not of type ARRAY0"
    )
    check('''\
    module function_type_check3 {
      const (subtype, midtype, supertype) =
        (array0, array const int, array int)
      fn f(x: supertype): midtype { return []: subtype }
      const g = f: (fn(subtype): supertype)
      g([]: midtype)
    }
    ''', [],
        "TypeVMError: value of type ARRAY is not of type ARRAY0"
    )
    check('''\
    module function_type_check4 {
      const (subtype, midtype, supertype) =
        (array0, array const int, array int)
      fn f(x: midtype): midtype { return []: subtype }
      const g = f: (fn(supertype): supertype)
      g([]: subtype)
    }
    ''', [],
        "TypeVMError: value of type FUNCTION is not of type FUNCTION"
    )
    check('''\
    module function_type_check5 {
      const (subtype, midtype, supertype) =
        (array0, array const int, array int)
      fn f(x: supertype): supertype { return []: subtype }
      const g = f: (fn(midtype): midtype)
      g([]: subtype)
    }
    ''', [],
        "TypeVMError: value of type FUNCTION is not of type FUNCTION"
    )


def test_defer():
    check('defer print', [],
        'TypeVMError: the type used for "defer" is not of type UNION')
    check('defer int', [],
        'DeferTypeVMError: cannot "defer" or "fulfil" values of type INT')
    check('const f = defer fn(): int', [])


def test_fulfil():
    check('''\
    module intf {
      const intf = fn(): int
      const k3 = defer intf
      fn k(x: int): intf {
        return fn(): int { return x }
      }
      fulfil k3 = k(3)
      k3()
    }
    ''', [const_int(3, is_static=False)])
    check('''\
    module fulfil_copy_of_defer_cell {
      const cell = defer fn(): int
      const copy = cell
      fulfil copy = fn(): int { return 3 }
      cell()
    }
    ''', [const_int(3, is_static=False)])
    check('''\
    module fulfil_forwarding_pointer {
      const x = defer fn(): int
      const y = defer fn(): int
      fulfil x = y
      fulfil y = fn(): int { return 3 }
      x()
      y()
    }
    ''', [const_int(3, is_static=False), const_int(3, is_static=False)])
    check('''\
    module self_fulfil {
      const x = defer fn(): int
      fulfil x = x
      fulfil x = fn(): int { return 3 }
      x()
    }
    ''', [const_int(3, is_static=False)])


def test_call_deferred():
    check('(defer fn(): int)()', [],
        'UnfulfilledValueVMError: the function value has not been fulfilled')
    check('''\
    module call_forwarding_pointer {
      const x = defer fn(): int
      const y = defer fn(): int
      fulfil x = y
      x()
    }
    ''',
        [],
        'UnfulfilledValueVMError: the function value has not been fulfilled'
    )


def test_double_fulfil():
    check('''\
    module double_fulfil {
      const cell = defer fn(): int
      fulfil cell = fn(): int { return 3 }
      fulfil cell = fn(): int { return 3 }
    }
    ''',
        [],
        'FulfilledValueVMError: the value has already been fulfilled'
    )
    check('''\
    module double_fulfil_copy {
      const cell = defer fn(): int
      const copy = cell
      fulfil cell = fn(): int { return 3 }
      fulfil copy = fn(): int { return 3 }
    }
    ''',
        [],
        'FulfilledValueVMError: the value has already been fulfilled'
    )
    check('''\
    module double_fulfil_forwarding_pointer {
      const x = defer fn(): int
      const y = defer fn(): int
      fulfil x = y
      fulfil y = fn(): int { return 3 }
      fulfil x = fn(): int { return 3 }
    }
    ''',
        [],
        'FulfilledValueVMError: the value has already been fulfilled'
    )


def test_cast_to_unfulfilled_type():
    check('''\
    module cast_to_unfulfilled_type {
      const t = defer type
      5: t
    }
    ''',
        [],
        'UnfulfilledValueVMError: the value has not been fulfilled'
    )


def test_cast_to_forwarded_type():
    check('''\
    module cast_to_forwarded_type {
      const t = defer type
      const u = (t,)
      fulfil t = int
      const (v,) = u
      5: v
    }
    ''', [
        const_int(5)
    ])


@slow
def test_recursion():
    check('''\
    module triangular {
      fn triangular(x: int): int {
        if x <= 0 { return 0 } else { return x + triangular(x-1) }
      }
      triangular(10)
    }
    ''', [
        const_int(55, is_static=False)
    ])
    check('''\
    module fibonacci {
      fn fibonacci(x: int): int {
        if x <= 1 { return x } else { return fibonacci(x-1) + fibonacci(x-2) }
      }
      fibonacci(10)
    }
    ''', [
        const_int(55, is_static=False)
    ])
    check('''\
    module man_or_boy {
      // `intf` is the type of functions taking nothing and returning `int`.
      // `f` turns an `int` into an `intf` with no side-effects.
      const intf = fn(): int
      fn f(x: int): intf { return fn(): int { return x } }

      fn a(k: int, x1: intf, x2: intf, x3: intf, x4: intf, x5: intf): int {
        var k = k
        fn b(): int {
          k -= 1
          return a(k, b, x1, x2, x3, x4)
        }
        if k <= 0 { return x4() + x5() } else { return b() }
      }

      a(10, f(1), f(-1), f(-1), f(1), f(0))
    }''',
        [const_int(-67, is_static=False)]
    )

def test_mutual_recursion():
    check('''\
    fn f(): void {
        fn g(): void { const x = f }
        g()
    }()
    ''', [])


def test_references():
    check('*&3', [const_int(3)])
    check('*&three', [var_int(3)])
    check('''
    module ref1 {
        const a = &5
        a
        *a
    }''', [
        (True, CONST, wtypes.ref(CONST, wtypes.int_), (Ref.new((5,)),)),
        const_int(5),
    ])
    check('''
    module ref2 {
        var a = &5
        a
        *a
    }''', [
        (True, VAR, wtypes.ref(CONST, wtypes.int_), (Ref.new((5,)),)),
        const_int(5, is_static=False),
    ])
    check('''
    module ref3 {
        const a = new 5
        a
        *a
    }''', [
        (True, CONST, wtypes.ref(VAR, wtypes.int_), (Ref.new((5,)),)),
        var_int(5),
    ])
    check('''
    module ref4 {
        var a = new 5
        a
        *a
    }''', [
        (True, VAR, wtypes.ref(VAR, wtypes.int_), (Ref.new((5,)),)),
        var_int(5, is_static=False),
    ])
    check('''
    fn ref5(x: int): void {
        const a = &x
        a
        *a
    }(5)''', [
        (False, CONST, wtypes.ref(CONST, wtypes.int_), (Ref.new((5,)),)),
        const_int(5, is_static=False),
    ])
    check('''
    fn ref5(x: int): void {
        var a = &x
        a
        *a
    }(5)''', [
        (False, VAR, wtypes.ref(CONST, wtypes.int_), (Ref.new((5,)),)),
        const_int(5, is_static=False),
    ])
    check('''
    fn ref7(x: int): void {
        var x = x
        const a = &x
        a
        *a
    }(5)''', [
        (False, CONST, wtypes.ref(VAR, wtypes.int_), (Ref.new((5,)),)),
        var_int(5, is_static=False),
    ])
    check('''
    fn ref8(x: int): void {
        var x = x
        var a = &x
        a
        *a
    }(5)''', [
        (False, VAR, wtypes.ref(VAR, wtypes.int_), (Ref.new((5,)),)),
        var_int(5, is_static=False),
    ])
    check('''
    fn ref9(x: ref int): void {
        *x
    }(&5)
    ''', [
        (False, BASE, wtypes.int_, (5,)),
    ])


def test_reference_types():
    check('ref int', [
        (True, CONST, wtypes.type_, wtypes.ref(BASE, wtypes.int_).read()),
    ])
    check('ref const int', [
        (True, CONST, wtypes.type_, wtypes.ref(CONST, wtypes.int_).read()),
    ])
    check('ref var int', [
        (True, CONST, wtypes.type_, wtypes.ref(VAR, wtypes.int_).read()),
    ])
    check('fn (t: type): type { return ref t }(int)', [
        (False, CONST, wtypes.type_, wtypes.ref(BASE, wtypes.int_).read()),
    ])
    check('fn (t: type): type { return ref const t }(int)', [
        (False, CONST, wtypes.type_, wtypes.ref(CONST, wtypes.int_).read()),
    ])
    check('fn (t: type): type { return ref var t }(int)', [
        (False, CONST, wtypes.type_, wtypes.ref(VAR, wtypes.int_).read()),
    ])


def test_union_types():
    check(
        'typeof union int_tree {LEAF(); BRANCH(int_tree, int, int_tree)}',
        [(True, CONST, wtypes.type_, wtypes.type_.read())]
    )
    check(
        'typeof union int_tree {LEAF; BRANCH(int_tree, int, int_tree)}',
        [(True, CONST, wtypes.type_, wtypes.type_.read())]
    )
    check(r'''union { FOO; FOO }''', [],
        "FlattenerError: Case 'FOO' has already been declared")


def test_union():
    check('NONE', [(
        True,
        CONST,
        wtypes.union([('NONE', [])]),
        ('NONE', []),
    )])
    check('NONE()', [(
        True,
        CONST,
        wtypes.union([('NONE', [])]),
        ('NONE', []),
    )])
    check('SOME(3)', [(
        True,
        CONST,
        wtypes.union([('SOME', [wtypes.int_])]),
        ('SOME', [3]),
    )])
    check('SOME(three)', [(
        False,
        CONST,
        wtypes.union([('SOME', [wtypes.int_])]),
        ('SOME', [3]),
    )])
    check('FOO(bool, "2", 3)', [(
        True,
        CONST,
        wtypes.union([('FOO', [wtypes.type_, wtypes.str_, wtypes.int_])]),
        pack.tag('FOO', wtypes.bool_.read(), '2', 3).read(),
    )])
    check('FOO(bool, "2", three)', [(
        False,
        CONST,
        wtypes.union([('FOO', [wtypes.type_, wtypes.str_, wtypes.int_])]),
        pack.tag('FOO', wtypes.bool_.read(), '2', 3).read(),
    )])
    # Test values with duplicate cases in their union types.
    check('FOO: UNION([("FOO", []),("FOO", [str])])', [(
        True,
        CONST,
        wtypes.union([('FOO', [])]),
        ('FOO', []),
    )])
    check('FOO(""): UNION([("FOO", []),("FOO", [str])])', [],
        "TypeVMError: value of type UNION is not of type UNION")
    check('FOO: UNION([("FOO", [str]),("FOO", [])])', [],
        "TypeVMError: value of type UNION is not of type UNION")
    check('FOO(""): UNION([("FOO", [str]),("FOO", [])])', [(
        True,
        CONST,
        wtypes.union([('FOO', [wtypes.str_])]),
        ('FOO', [""]),
    )])


def test_dynamic():
    check('box 3', [(
        True,
        CONST,
        wtypes.dynamic,
        Dynamic.new(wtypes.int_, (3,)).to_welly(),
    )])
    check('box three', [(
        True,
        CONST,
        wtypes.dynamic,
        Dynamic.new(wtypes.int_, (3,), constness=VAR).to_welly(),
    )])
    check('(box 3).value', [const_int(3)])
    check('(box three).value', [var_int(3)])
    check('(box 3).type', [
        (True, CONST, wtypes.type_, wtypes.int_.read()),
    ])
    check('(box three).type', [
        (True, CONST, wtypes.type_, wtypes.int_.read()),
    ])
    check('(box 3).constness', [
        (True, CONST, wtypes.constness, CONST.read()),
    ])
    check('(box three).constness', [
        (True, CONST, wtypes.constness, VAR.read()),
    ])
    check('fn(x: int) { box x }(3)', [(
        False,
        CONST,
        wtypes.dynamic,
        Dynamic.new(wtypes.int_, (3,)).to_welly(),
    )])
    check('fn(x: int) { var x = x; box x }(3)', [(
        False,
        CONST,
        wtypes.dynamic,
        Dynamic.new(wtypes.int_, (3,), constness=VAR).to_welly(),
    )])
    check('fn(x: int) { (box x).value }(3)', [],
        'StaticnessVMError: the dynamic value must be static in order to '
        'access its value')
    check('fn(x: int) { var x = x; (box x).value }(3)', [],
        'StaticnessVMError: the dynamic value must be static in order to '
        'access its value')
    check('fn(x: int) { (box x).type }(3)', [
        (False, CONST, wtypes.type_, wtypes.int_.read()),
    ])
    check('fn(x: int) { var x = x; (box x).type }(3)', [
        (False, CONST, wtypes.type_, wtypes.int_.read()),
    ])
    check('fn(x: int) { (box x).constness }(3)', [
        (False, CONST, wtypes.constness, CONST.read()),
    ])
    check('fn(x: int) { var x = x; (box x).constness }(3)', [
        (False, CONST, wtypes.constness, VAR.read()),
    ])


def test_switch():
    check('switch FOO(three) { case FOO(x) {1} }', [const_int(1)])
    check('switch FOO(three) { else {3} }', [const_int(3)])
    check('switch FOO(int, 2, three, 4.0) { case FOO(w, x, y, z) {x} }', [
        const_int(2, is_static=False)
    ])
    # Check that empty "else" block is treated as present.
    check('switch FOO(three) { else {} }', [])
    # Test for duplicate cases in union types.
    check(r'''
        switch (FOO: UNION([("FOO", []),("FOO", [str])])) {case FOO(){}}
    ''', [])
    check(r'''
        switch (FOO(""): UNION([("FOO", [str]),("FOO", [])])) {case FOO(){}}
    ''', [],
        "NumberOfFieldsVMError: wrong number of fields - expected 0 but got 1"
    )
    check(r'''
        switch (FOO(""): UNION([("FOO", [str]),("FOO", [])])) {case FOO(s){}}
    ''', [])


def test_static_var():
    check('module var_ref { var t = int; defer ref t; }', [],
        'StaticnessVMError: the type used for "defer" must be static')
    check(r'module array_field { var a = [3]; defer a.length }', [],
        'StaticnessVMError: the type used for "defer" must be static')
    check(r'defer fn() { return type }()', [],
        'StaticnessVMError: the type used for "defer" must be static')
    # TODO: grep "typed_code_static.py" for `is_const()` and make tests for
    # each instance of the corresponding behaviour in "typed_code.py".


def test_for():
    check('for x in 3 { x }', [
        const_int(0, is_static=False),
        const_int(1, is_static=False),
        const_int(2, is_static=False),
    ])
    check('for x in 3 { x } else {}', [
        const_int(0, is_static=False),
        const_int(1, is_static=False),
        const_int(2, is_static=False),
    ])
    check('for x in 3 { x } else { 5 }', [
        const_int(0, is_static=False),
        const_int(1, is_static=False),
        const_int(2, is_static=False),
        const_int(5, is_static=True),
    ])
    check('for x in 3 { x; if x==1 { break } } else { 5 }', [
        const_int(0, is_static=False),
        const_int(1, is_static=False),
    ])
    check('for x in 3 { if x==1 { break }; x } else { 5 }', [
        const_int(0, is_static=False),
    ])
    check('for x in 3 { x; if x==1 { continue } } else { 5 }', [
        const_int(0, is_static=False),
        const_int(1, is_static=False),
        const_int(2, is_static=False),
        const_int(5, is_static=True),
    ])
    check('for x in 3 { if x==1 { continue }; x } else { 5 }', [
        const_int(0, is_static=False),
        const_int(2, is_static=False),
        const_int(5, is_static=True),
    ])


def test_while():
    check('''
    module count {
        var a = 10
        while a > 0 {
            a -= 1
            if a < 5 { continue }
            "h"
        }
    }''', [static_const(wtypes.str_, ('h',))] * 5)


def test_assert():
    check('assert(true, "Foo")', [])
    check('assert(false, "Foo")', [], 'AssertionVMError: Foo')
