'''
Interfaces the pretty printer to the Welly data model.
'''


import re

from . import defer, const, wtypes, subtypes, pretty_type

from .errors import UnfulfilledValueVMError
from .memory import Ref, Array
from .ops import Dynamic, struct_iter


__all__ = ['wrap']


def _constant(value):
    assert type(value) is unicode
    return lambda: (value, u'', u'', [])


# To convert a string value into a Welly string literal, we have to make the
# following replacements.
_esc = {
    '\t': r'\t',
    '\n': r'\n',
    '\r': r'\r',
    '\"': r'\"',
    '\\': r'\\',
}


# Exploit the fact that Welly escapes are a subset of Python escapes.
_regexp = re.compile(r'[%s]' % ''.join(_esc.values()))


def _str(value):
    assert type(value) is str
    value = _regexp.sub(lambda m: _esc[m.group()], value)
    value = value.decode('utf8', errors='replace')
    return lambda: (u'"%s"' % value, u'', u'', [])


def _struct(value, tag=''):
    assert type(value) is Dynamic
    return lambda: (
        u'%s(' % tag,
        u', ',
        u')',
        [wrap(x) for x in struct_iter(value)]
    )


def _constness_syntax(constness):
    '''
    Returns '', 'const ' or 'var '. This is used to construct strings such as
    "ref const int" or "array var str".
    '''
    c0 = const.switch(constness)
    if c0 == 'BASE':
        return 'BASE' # Not really Welly syntax.
    elif c0 == 'CONST':
        return 'const'
    elif c0 == 'VAR':
        return 'var'
    assert False


def _binding(key, value):
    assert type(key) is str, repr(key)
    assert type(value) is Dynamic, repr(value)
    return lambda: (
        u'%s %s = ' % (_constness_syntax(value.constness), key),
        u'',
        u'',
        [wrap(value)]
    )


def wrap(x):
    '''
    Given a Welly value, returns a pretty-print-aware wrapper around it.

     - x - a Dynamic.
    '''
    tag, payload = wtypes.switch(x.type_)
    if tag in ('FLOAT', 'INT'):
        # Use a Python value with the same syntax as the Welly value.
        (value,) = x.read()
        return _constant(unicode(value))
    elif tag == 'BOOL':
        (value,) = x.read()
        return _constant(u'true' if value else u'false')
    elif tag == 'STR':
        (value,) = x.read()
        return _str(value)
    elif tag == 'REF':
        (constness, referee_type) = payload
        (value,) = x.read()
        return lambda: (u'&', u'', u'', [
            wrap(Dynamic(referee_type, value, constness=constness))
        ])
    elif tag == 'ARRAY0':
        # Use a Python value with the same syntax as the Welly value.
        return _constant(u'[]')
    elif tag == 'ARRAY':
        (constness, elem_type) = payload
        values = Array.from_welly(x.read(), wtypes.sizeof(elem_type))
        return lambda: (u'[', u', ', u']', [
            wrap(Dynamic(elem_type, ref, constness=constness))
            for ref in values
        ])
    elif tag == 'STRUCT':
        (types,) = payload
        return _struct(x)
    elif tag == 'UNION':
        (cases,) = payload
        if subtypes.eq(x.type_, wtypes.type_):
            # Use Welly syntactic sugar.
            return pretty_type.wrap(x.ref)
        try:
            tag, payload = defer.read(x.ref)
        except UnfulfilledValueVMError:
            return _constant(u'unfulfilled')
        assert type(tag) is str, x
        payload = Ref.block(payload)
        types = wtypes.union_index(cases)[tag]
        payload_t = wtypes.struct(types)
        payload_struct = Dynamic(payload_t, payload)
        return _struct(payload_struct, tag=tag)
    elif tag == 'FUNCTION':
        (param_types, return_type) = payload
        try:
            tag, payload = defer.read(x.ref)
        except UnfulfilledValueVMError:
            return _constant(u'unfulfilled')
        # We have no way of displaying any details.
        return _constant(u'<fn>')
    elif tag == 'DYNAMIC':
        value = Dynamic.from_welly(x.ref)
        return lambda: (u'box ', u'', u'', [wrap(value)])
    elif tag == 'MODULE':
        (value,) = x.read()
        return lambda: (u'module {', u', ', u'}', [
            _binding(k, Dynamic.from_welly(Ref.new(v)))
            for k, v in value.items()
        ])
        return Module(value)
    assert False, tag
